@@ -10,6 +10,9 @@ module GHC.CmmToLlvm.CodeGen ( genLlvmProc ) where
 
 import GHC.Prelude
 
+import GHC.Driver.Session
+import GHC.Driver.Ppr
+
 import GHC.Llvm
 import GHC.CmmToLlvm.Base
 import GHC.CmmToLlvm.Regs
@@ -25,11 +28,11 @@ import GHC.Cmm.Dataflow.Block
 import GHC.Cmm.Dataflow.Graph
 import GHC.Cmm.Dataflow.Collections
 
-import GHC.Driver.Session
 import GHC.Data.FastString
 import GHC.Types.ForeignCall
-import GHC.Utils.Outputable hiding (panic, pprPanic)
-import qualified GHC.Utils.Outputable as Outputable
+import GHC.Utils.Outputable
+import GHC.Utils.Panic (assertPanic)
+import qualified GHC.Utils.Panic as Panic
 import GHC.Platform
 import GHC.Data.OrdList
 import GHC.Types.Unique.Supply
@@ -185,18 +188,18 @@ barrierUnless exs = do
         else barrier
 
 -- | Foreign Calls
-genCall :: ForeignTarget -> [CmmFormal] -> [CmmActual]
-              -> LlvmM StmtData
+genCall :: ForeignTarget -> [CmmFormal] -> [CmmActual] -> LlvmM StmtData
 
 -- Barriers need to be handled specially as they are implemented as LLVM
 -- intrinsic functions.
 genCall (PrimTarget MO_ReadBarrier) _ _ =
     barrierUnless [ArchX86, ArchX86_64, ArchSPARC]
-genCall (PrimTarget MO_WriteBarrier) _ _ = do
+
+genCall (PrimTarget MO_WriteBarrier) _ _ =
     barrierUnless [ArchX86, ArchX86_64, ArchSPARC]
 
-genCall (PrimTarget MO_Touch) _ _
- = return (nilOL, [])
+genCall (PrimTarget MO_Touch) _ _ =
+    return (nilOL, [])
 
 genCall (PrimTarget (MO_UF_Conv w)) [dst] [e] = runStmtsDecls $ do
     dstV <- getCmmRegW (CmmLocal dst)
@@ -514,9 +517,8 @@ genCall target res args = do
 
     -- make the actual call
     case retTy of
-        LMVoid -> do
+        LMVoid ->
             statement $ Expr $ Call ccTy fptr argVars fnAttrs
-
         _ -> do
             v1 <- doExprW retTy $ Call ccTy fptr argVars fnAttrs
             -- get the return register
@@ -905,6 +907,37 @@ cmmPrimOpFunctions mop = do
     MO_Cmpxchg _     -> unsupported
     MO_Xchg _        -> unsupported
 
+    MO_I64_ToI       -> fsLit "hs_int64ToInt"
+    MO_I64_FromI     -> fsLit "hs_intToInt64"
+    MO_W64_ToW       -> fsLit "hs_word64ToWord"
+    MO_W64_FromW     -> fsLit "hs_wordToWord64"
+    MO_x64_Neg       -> fsLit "hs_neg64"
+    MO_x64_Add       -> fsLit "hs_add64"
+    MO_x64_Sub       -> fsLit "hs_sub64"
+    MO_x64_Mul       -> fsLit "hs_mul64"
+    MO_I64_Quot      -> fsLit "hs_quotInt64"
+    MO_I64_Rem       -> fsLit "hs_remInt64"
+    MO_W64_Quot      -> fsLit "hs_quotWord64"
+    MO_W64_Rem       -> fsLit "hs_remWord64"
+    MO_x64_And       -> fsLit "hs_and64"
+    MO_x64_Or        -> fsLit "hs_or64"
+    MO_x64_Xor       -> fsLit "hs_xor64"
+    MO_x64_Not       -> fsLit "hs_not64"
+    MO_x64_Shl       -> fsLit "hs_uncheckedShiftL64"
+    MO_I64_Shr       -> fsLit "hs_uncheckedIShiftRA64"
+    MO_W64_Shr       -> fsLit "hs_uncheckedShiftRL64"
+    MO_x64_Eq        -> fsLit "hs_eq64"
+    MO_x64_Ne        -> fsLit "hs_ne64"
+    MO_I64_Ge        -> fsLit "hs_geInt64"
+    MO_I64_Gt        -> fsLit "hs_gtInt64"
+    MO_I64_Le        -> fsLit "hs_leInt64"
+    MO_I64_Lt        -> fsLit "hs_ltInt64"
+    MO_W64_Ge        -> fsLit "hs_geWord64"
+    MO_W64_Gt        -> fsLit "hs_gtWord64"
+    MO_W64_Le        -> fsLit "hs_leWord64"
+    MO_W64_Lt        -> fsLit "hs_ltWord64"
+
+
 -- | Tail function calls
 genJump :: CmmExpr -> [GlobalReg] -> LlvmM StmtData
 
@@ -1507,9 +1540,9 @@ genMachOp_slow opt op [x, y] = case op of
     MO_And _   -> genBinMach LM_MO_And
     MO_Or  _   -> genBinMach LM_MO_Or
     MO_Xor _   -> genBinMach LM_MO_Xor
-    MO_Shl _   -> genBinMach LM_MO_Shl
-    MO_U_Shr _ -> genBinMach LM_MO_LShr
-    MO_S_Shr _ -> genBinMach LM_MO_AShr
+    MO_Shl _   -> genBinCastYMach LM_MO_Shl
+    MO_U_Shr _ -> genBinCastYMach LM_MO_LShr
+    MO_S_Shr _ -> genBinCastYMach LM_MO_AShr
 
     MO_V_Add l w   -> genCastBinMach (LMVector l (widthToLlvmInt w)) LM_MO_Add
     MO_V_Sub l w   -> genCastBinMach (LMVector l (widthToLlvmInt w)) LM_MO_Sub
@@ -1553,20 +1586,27 @@ genMachOp_slow opt op [x, y] = case op of
 #endif
 
     where
-        binLlvmOp ty binOp = do
+        binLlvmOp ty binOp allow_y_cast = do
           platform <- getPlatform
           runExprData $ do
             vx <- exprToVarW x
             vy <- exprToVarW y
-            if getVarType vx == getVarType vy
-                then do
-                    doExprW (ty vx) $ binOp vx vy
 
-                else do
+            if | getVarType vx == getVarType vy
+               -> doExprW (ty vx) $ binOp vx vy
+
+               | allow_y_cast
+               -> do
+                    vy' <- singletonPanic "binLlvmOp cast"<$>
+                            castVarsW Signed [(vy, (ty vx))]
+                    doExprW (ty vx) $ binOp vx vy'
+
+               | otherwise
+               -> do
                     -- Error. Continue anyway so we can debug the generated ll file.
                     dflags <- getDynFlags
-                    let style = mkCodeStyle CStyle
-                        toString doc = renderWithStyle (initSDocContext dflags style) doc
+                    let style = PprCode CStyle
+                        toString doc = renderWithContext (initSDocContext dflags style) doc
                         cmmToStr = (lines . toString . PprCmm.pprExpr platform)
                     statement $ Comment $ map fsLit $ cmmToStr x
                     statement $ Comment $ map fsLit $ cmmToStr y
@@ -1584,7 +1624,7 @@ genMachOp_slow opt op [x, y] = case op of
         -- comparisons while LLVM return i1. Need to extend to llvmWord type
         -- if expected. See Note [Literals and branch conditions].
         genBinComp opt cmp = do
-            ed@(v1, stmts, top) <- binLlvmOp (\_ -> i1) (Compare cmp)
+            ed@(v1, stmts, top) <- binLlvmOp (\_ -> i1) (Compare cmp) False
             dflags <- getDynFlags
             platform <- getPlatform
             if getVarType v1 == i1
@@ -1598,7 +1638,9 @@ genMachOp_slow opt op [x, y] = case op of
                     panic $ "genBinComp: Compare returned type other then i1! "
                         ++ (showSDoc dflags $ ppr $ getVarType v1)
 
-        genBinMach op = binLlvmOp getVarType (LlvmOp op)
+        genBinMach op = binLlvmOp getVarType (LlvmOp op) False
+
+        genBinCastYMach op = binLlvmOp getVarType (LlvmOp op) True
 
         genCastBinMach ty op = binCastLlvmOp ty (LlvmOp op)
 
@@ -1717,19 +1759,19 @@ genLoad_slow atomic e ty meta = do
   runExprData $ do
     iptr <- exprToVarW e
     case getVarType iptr of
-         LMPointer _ -> do
+        LMPointer _ ->
                     doExprW (cmmToLlvmType ty) (MExpr meta $ loadInstr iptr)
 
-         i@(LMInt _) | i == llvmWord platform -> do
+        i@(LMInt _) | i == llvmWord platform -> do
                     let pty = LMPointer $ cmmToLlvmType ty
                     ptr <- doExprW pty $ Cast LM_Inttoptr iptr pty
                     doExprW (cmmToLlvmType ty) (MExpr meta $ loadInstr ptr)
 
-         other -> do pprPanic "exprToVar: CmmLoad expression is not right type!"
-                        (PprCmm.pprExpr platform e <+> text (
-                            "Size of Ptr: " ++ show (llvmPtrBits platform) ++
-                            ", Size of var: " ++ show (llvmWidthInBits platform other) ++
-                            ", Var: " ++ showSDoc dflags (ppVar opts iptr)))
+        other -> pprPanic "exprToVar: CmmLoad expression is not right type!"
+                     (PprCmm.pprExpr platform e <+> text (
+                         "Size of Ptr: " ++ show (llvmPtrBits platform) ++
+                         ", Size of var: " ++ show (llvmWidthInBits platform other) ++
+                         ", Var: " ++ showSDoc dflags (ppVar opts iptr)))
   where
     loadInstr ptr | atomic    = ALoad SyncSeqCst False ptr
                   | otherwise = Load ptr
@@ -2004,10 +2046,10 @@ toIWord platform = mkIntLit (llvmWord platform)
 
 -- | Error functions
 panic :: HasCallStack => String -> a
-panic s = Outputable.panic $ "GHC.CmmToLlvm.CodeGen." ++ s
+panic s = Panic.panic $ "GHC.CmmToLlvm.CodeGen." ++ s
 
 pprPanic :: HasCallStack => String -> SDoc -> a
-pprPanic s d = Outputable.pprPanic ("GHC.CmmToLlvm.CodeGen." ++ s) d
+pprPanic s d = Panic.pprPanic ("GHC.CmmToLlvm.CodeGen." ++ s) d
 
 
 -- | Returns TBAA meta data by unique
