@@ -23,20 +23,25 @@ module GHC.Core.Opt.FloatIn ( floatInwards ) where
 import GHC.Prelude
 import GHC.Platform
 
+import GHC.Driver.Session
+
 import GHC.Core
 import GHC.Core.Make hiding ( wrapFloats )
-import GHC.Driver.Types     ( ModGuts(..) )
 import GHC.Core.Utils
 import GHC.Core.FVs
-import GHC.Core.Opt.Monad    ( CoreM )
-import GHC.Types.Id         ( isOneShotBndr, idType, isJoinId, isJoinId_maybe )
-import GHC.Types.Var
+import GHC.Core.Opt.Monad   ( CoreM )
 import GHC.Core.Type
-import GHC.Types.Var.Set
-import GHC.Utils.Misc
-import GHC.Driver.Session
-import GHC.Utils.Outputable
+
 import GHC.Types.Basic      ( RecFlag(..), isRec )
+import GHC.Types.Id         ( isOneShotBndr, idType, isJoinId, isJoinId_maybe )
+import GHC.Types.Tickish
+import GHC.Types.Var
+import GHC.Types.Var.Set
+
+import GHC.Unit.Module.ModGuts
+
+import GHC.Utils.Misc
+import GHC.Utils.Panic
 
 {-
 Top-level interface function, @floatInwards@.  Note that we do not
@@ -450,7 +455,7 @@ bindings are:
 
 -}
 
-fiExpr platform to_drop (_, AnnCase scrut case_bndr _ [(con,alt_bndrs,rhs)])
+fiExpr platform to_drop (_, AnnCase scrut case_bndr _ [AnnAlt con alt_bndrs rhs])
   | isUnliftedType (idType case_bndr)
   , exprOkForSideEffects (deAnnotate scrut)
       -- See Note [Floating primops]
@@ -489,12 +494,12 @@ fiExpr platform to_drop (_, AnnCase scrut case_bndr ty alts)
     scrut_fvs    = freeVarsOf scrut
     alts_fvs     = map alt_fvs alts
     all_alts_fvs = unionDVarSets alts_fvs
-    alt_fvs (_con, args, rhs)
+    alt_fvs (AnnAlt _con args rhs)
       = foldl' delDVarSet (freeVarsOf rhs) (case_bndr:args)
            -- Delete case_bndr and args from free vars of rhs
            -- to get free vars of alt
 
-    fi_alt to_drop (con, args, rhs) = (con, args, fiExpr platform to_drop rhs)
+    fi_alt to_drop (AnnAlt con args rhs) = Alt con args (fiExpr platform to_drop rhs)
 
 ------------------
 fiBind :: Platform
