@@ -1,3 +1,8 @@
+{-# LANGUAGE CPP          #-}
+{-# LANGUAGE BangPatterns #-}
+
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
+
 {-
 (c) The University of Glasgow, 1994-2006
 
@@ -5,10 +10,6 @@
 Core pass to saturate constructors and PrimOps
 -}
 
-{-# LANGUAGE BangPatterns, CPP, MultiWayIf #-}
-
-{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
-
 module GHC.CoreToStg.Prep
    ( corePrepPgm
    , corePrepExpr
@@ -19,15 +20,23 @@ where
 #include "HsVersions.h"
 
 import GHC.Prelude
+
 import GHC.Platform
+import GHC.Platform.Ways
 
-import GHC.Core.Opt.OccurAnal
+import GHC.Driver.Session
+import GHC.Driver.Env
+import GHC.Driver.Ppr
 
+import GHC.Tc.Utils.Env
+import GHC.Unit
+
+import GHC.Builtin.Names
 import GHC.Builtin.PrimOps
+import GHC.Builtin.Types
 import GHC.Builtin.Types.Prim ( realWorldStatePrimTy )
 import GHC.Types.Id.Make ( realWorldPrimId, mkPrimOpId )
-import GHC.Driver.Types
-import GHC.Builtin.Names
+
 import GHC.Core.Utils
 import GHC.Core.Opt.Arity
 import GHC.Core.FVs
@@ -36,39 +45,43 @@ import GHC.Core.Lint    ( endPassIO )
 import GHC.Core
 import GHC.Core.Make hiding( FloatBind(..) )   -- We use our own FloatBind here
 import GHC.Core.Type
-import GHC.Types.Literal
 import GHC.Core.Coercion
-import GHC.Tc.Utils.Env
 import GHC.Core.TyCon
+import GHC.Core.DataCon
+import GHC.Core.Opt.OccurAnal
 import GHC.Core.TyCo.Rep( UnivCoProvenance(..) )
+
+
+import GHC.Data.Maybe
+import GHC.Data.OrdList
+import GHC.Data.FastString
+
+import GHC.Utils.Error
+import GHC.Utils.Misc
+import GHC.Utils.Panic
+import GHC.Utils.Outputable
+import GHC.Utils.Monad  ( mapAccumLM )
+import GHC.Utils.Logger
+
 import GHC.Types.Demand
 import GHC.Types.Var
 import GHC.Types.Var.Set
 import GHC.Types.Var.Env
 import GHC.Types.Id
 import GHC.Types.Id.Info
-import GHC.Builtin.Types
-import GHC.Core.DataCon
 import GHC.Types.Basic
-import GHC.Unit.Module
-import GHC.Types.Unique.Supply
-import GHC.Data.Maybe
-import GHC.Data.OrdList
-import GHC.Utils.Error
-import GHC.Driver.Session
-import GHC.Driver.Ways
-import GHC.Utils.Misc
-import GHC.Utils.Outputable
-import GHC.Data.FastString
 import GHC.Types.Name   ( NamedThing(..), nameSrcSpan, isInternalName )
 import GHC.Types.SrcLoc ( SrcSpan(..), realSrcLocSpan, mkRealSrcLoc )
+import GHC.Types.Literal
+import GHC.Types.Tickish
+import GHC.Types.TyThing
+import GHC.Types.CostCentre ( CostCentre, ccFromThisModule )
+import GHC.Types.Unique.Supply
+
 import GHC.Data.Pair
-import Data.Bits
-import GHC.Utils.Monad  ( mapAccumLM )
 import Data.List        ( unfoldr )
 import Data.Functor.Identity
 import Control.Monad
-import GHC.Types.CostCentre ( CostCentre, ccFromThisModule )
 import qualified Data.Set as S
 
 {-
@@ -228,9 +241,9 @@ type CpeRhs  = CoreExpr    -- Non-terminal 'rhs'
 corePrepPgm :: HscEnv -> Module -> ModLocation -> CoreProgram -> [TyCon]
             -> IO (CoreProgram, S.Set CostCentre)
 corePrepPgm hsc_env this_mod mod_loc binds data_tycons =
-    withTiming dflags
+    withTiming logger dflags
                (text "CorePrep"<+>brackets (ppr this_mod))
-               (const ()) $ do
+               (\(a,b) -> a `seqList` b `seq` ()) $ do
     us <- mkSplitUniqSupply 's'
     initialCorePrepEnv <- mkInitialCorePrepEnv hsc_env
 
@@ -253,15 +266,17 @@ corePrepPgm hsc_env this_mod mod_loc binds data_tycons =
     return (binds_out, cost_centres)
   where
     dflags = hsc_dflags hsc_env
+    logger = hsc_logger hsc_env
 
 corePrepExpr :: HscEnv -> CoreExpr -> IO CoreExpr
 corePrepExpr hsc_env expr = do
     let dflags = hsc_dflags hsc_env
-    withTiming dflags (text "CorePrep [expr]") (const ()) $ do
+    let logger = hsc_logger hsc_env
+    withTiming logger dflags (text "CorePrep [expr]") (\e -> e `seq` ()) $ do
       us <- mkSplitUniqSupply 's'
       initialCorePrepEnv <- mkInitialCorePrepEnv hsc_env
       let new_expr = initUs_ us (cpeBodyNF initialCorePrepEnv expr)
-      dumpIfSet_dyn dflags Opt_D_dump_prep "CorePrep" FormatCore (ppr new_expr)
+      dumpIfSet_dyn logger dflags Opt_D_dump_prep "CorePrep" FormatCore (ppr new_expr)
       return new_expr
 
 corePrepTopBinds :: CorePrepEnv -> [CoreBind] -> UniqSM Floats
@@ -792,9 +807,9 @@ cpeRhsE env (Tick tickish expr)
   = do { body <- cpeBodyNF env expr
        ; return (emptyFloats, mkTick tickish' body) }
   where
-    tickish' | Breakpoint n fvs <- tickish
+    tickish' | Breakpoint ext n fvs <- tickish
              -- See also 'substTickish'
-             = Breakpoint n (map (getIdFromTrivialExpr . lookupCorePrepEnv env) fvs)
+             = Breakpoint ext n (map (getIdFromTrivialExpr . lookupCorePrepEnv env) fvs)
              | otherwise
              = tickish
 
@@ -825,7 +840,7 @@ cpeRhsE env (Case scrut bndr _ alts)
   | isUnsafeEqualityProof scrut
   , isDeadBinder bndr -- We can only discard the case if the case-binder
                       -- is dead.  It usually is, but see #18227
-  , [(_, [co_var], rhs)] <- alts
+  , [Alt _ [co_var] rhs] <- alts
   , let Pair ty1 ty2 = coVarTypes co_var
         the_co = mkUnsafeCo Nominal (cpSubstTy env ty1) (cpSubstTy env ty2)
         env'   = extendCoVarEnv env co_var the_co
@@ -851,10 +866,10 @@ cpeRhsE env (Case scrut bndr ty alts)
 
        ; return (floats, Case scrut' bndr2 ty alts'') }
   where
-    sat_alt env (con, bs, rhs)
+    sat_alt env (Alt con bs rhs)
        = do { (env2, bs') <- cpCloneBndrs env bs
             ; rhs' <- cpeBodyNF env2 rhs
-            ; return (con, bs', rhs') }
+            ; return (Alt con bs' rhs') }
 
 -- ---------------------------------------------------------------------------
 --              CpeBody: produces a result satisfying CpeBody
@@ -924,7 +939,7 @@ rhsToBody expr = return (emptyFloats, expr)
 
 data ArgInfo = CpeApp  CoreArg
              | CpeCast Coercion
-             | CpeTick (Tickish Id)
+             | CpeTick CoreTickish
 
 instance Outputable ArgInfo where
   ppr (CpeApp arg) = text "app" <+> ppr arg
@@ -967,7 +982,10 @@ cpeApp top_env expr
             -> UniqSM (Floats, CpeRhs)
     cpe_app env (Var f) (CpeApp Type{} : CpeApp arg : args) depth
         | f `hasKey` lazyIdKey          -- Replace (lazy a) with a, and
+            -- See Note [lazyId magic] in GHC.Types.Id.Make
        || f `hasKey` noinlineIdKey      -- Replace (noinline a) with a
+            -- See Note [noinlineId magic] in GHC.Types.Id.Make
+
         -- Consider the code:
         --
         --      lazy (f x) y
@@ -1005,9 +1023,9 @@ cpeApp top_env expr
                   Lam s body -> cpe_app (extendCorePrepEnvExpr env s s0) body rest (n-2)
                   _          -> cpe_app env k (CpeApp s0 : rest) (n-1)
              ; let touchId = mkPrimOpId TouchOp
-                   expr = Case k' y result_ty [(DEFAULT, [], rhs)]
+                   expr = Case k' y result_ty [Alt DEFAULT [] rhs]
                    rhs = let scrut = mkApps (Var touchId) [Type arg_rep, Type arg_ty, arg, Var realWorldPrimId]
-                         in Case scrut s2 result_ty [(DEFAULT, [], Var y)]
+                         in Case scrut s2 result_ty [Alt DEFAULT [] (Var y)]
              ; (floats', expr') <- cpeBody env expr
              ; return (floats `appendFloats` floats', expr')
              }
@@ -1159,7 +1177,7 @@ no further floating will occur. This allows us to safely inline things like
    GHC.Magic. This definition is used in cases where runRW is curried.
 
  * In addition to its normal Haskell definition in GHC.Magic, we give it
-   a special late inlining here in CorePrep and GHC.CoreToByteCode, avoiding
+   a special late inlining here in CorePrep and GHC.StgToByteCode, avoiding
    the incorrect sharing due to float-out noted above.
 
  * It is levity-polymorphic:
@@ -1260,7 +1278,7 @@ Performing the transform described above would result in:
 
 If runRW# were a "normal" function this call to join point j would not be
 allowed in its continuation argument. However, since runRW# is inlined (as
-described in Note [runRW magic] above), such join point occurences are
+described in Note [runRW magic] above), such join point occurrences are
 completely fine. Both occurrence analysis (see the runRW guard in occAnalApp)
 and Core Lint (see the App case of lintCoreExpr) have special treatment for
 runRW# applications. See Note [Linting of runRW#] for details on the latter.
@@ -1533,6 +1551,22 @@ tryEtaReducePrep _ _ = Nothing
 Note [Pin demand info on floats]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 We pin demand info on floated lets, so that we can see the one-shot thunks.
+
+Note [Speculative evaluation]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Since call-by-value is much cheaper than call-by-need, we case-bind arguments
+that are either
+
+  1. Strictly evaluated anyway, according to the DmdSig of the callee, or
+  2. ok-for-spec, according to 'exprOkForSpeculation'
+
+While (1) is a no-brainer and always beneficial, (2) is a bit
+more subtle, as the careful haddock for 'exprOkForSpeculation'
+points out. Still, by case-binding the argument we don't need
+to allocate a thunk for it, whose closure must be retained as
+long as the callee might evaluate it. And if it is evaluated on
+most code paths anyway, we get to turn the unknown eval in the
+callee into a known call at the call site.
 -}
 
 data FloatingBind
@@ -1548,7 +1582,7 @@ data FloatingBind
                       -- but lifted binding
 
  -- | See Note [Floating Ticks in CorePrep]
- | FloatTick (Tickish Id)
+ | FloatTick CoreTickish
 
 data Floats = Floats OkToSpec (OrdList FloatingBind)
 
@@ -1581,19 +1615,33 @@ data OkToSpec
 
 mkFloat :: Demand -> Bool -> Id -> CpeRhs -> FloatingBind
 mkFloat dmd is_unlifted bndr rhs
-  | is_strict
-  , not is_hnf  = FloatCase rhs bndr DEFAULT [] (exprOkForSpeculation rhs)
+  | is_strict || ok_for_spec -- See Note [Speculative evaluation]
+  , not is_hnf  = FloatCase rhs bndr DEFAULT [] ok_for_spec
     -- Don't make a case for a HNF binding, even if it's strict
     -- Otherwise we get  case (\x -> e) of ...!
 
-  | is_unlifted = ASSERT2( exprOkForSpeculation rhs, ppr rhs )
-                  FloatCase rhs bndr DEFAULT [] True
-  | is_hnf    = FloatLet (NonRec bndr                       rhs)
-  | otherwise = FloatLet (NonRec (setIdDemandInfo bndr dmd) rhs)
+  | is_unlifted = FloatCase rhs bndr DEFAULT [] True
+      -- we used to ASSERT2(ok_for_spec, ppr rhs) here, but it is now disabled
+      -- because exprOkForSpeculation isn't stable under ANF-ing. See for
+      -- example #19489 where the following unlifted expression:
+      --
+      --    GHC.Prim.(#|_#) @LiftedRep @LiftedRep @[a_ax0] @[a_ax0]
+      --                    (GHC.Types.: @a_ax0 a2_agq a3_agl)
+      --
+      -- is ok-for-spec but is ANF-ised into:
+      --
+      --    let sat = GHC.Types.: @a_ax0 a2_agq a3_agl
+      --    in GHC.Prim.(#|_#) @LiftedRep @LiftedRep @[a_ax0] @[a_ax0] sat
+      --
+      -- which isn't ok-for-spec because of the let-expression.
+
+  | is_hnf      = FloatLet (NonRec bndr                       rhs)
+  | otherwise   = FloatLet (NonRec (setIdDemandInfo bndr dmd) rhs)
                    -- See Note [Pin demand info on floats]
   where
-    is_hnf    = exprIsHNF rhs
-    is_strict = isStrictDmd dmd
+    is_hnf      = exprIsHNF rhs
+    is_strict   = isStrUsedDmd dmd
+    ok_for_spec = exprOkForSpeculation rhs
 
 emptyFloats :: Floats
 emptyFloats = Floats OkToSpec nilOL
@@ -1605,7 +1653,7 @@ wrapBinds :: Floats -> CpeBody -> CpeBody
 wrapBinds (Floats _ binds) body
   = foldrOL mk_bind body binds
   where
-    mk_bind (FloatCase rhs bndr con bs _) body = Case rhs bndr (exprType body) [(con,bs,body)]
+    mk_bind (FloatCase rhs bndr con bs _) body = Case rhs bndr (exprType body) [Alt con bs body]
     mk_bind (FloatLet bind)               body = Let bind body
     mk_bind (FloatTick tickish)           body = mkTick tickish body
 
@@ -1680,7 +1728,7 @@ canFloat (Floats ok_to_spec fs) rhs
 wantFloatNested :: RecFlag -> Demand -> Bool -> Floats -> CpeRhs -> Bool
 wantFloatNested is_rec dmd is_unlifted floats rhs
   =  isEmptyFloats floats
-  || isStrictDmd dmd
+  || isStrUsedDmd dmd
   || is_unlifted
   || (allLazyNested is_rec floats && exprIsHNF rhs)
         -- Why the test for allLazyNested?
@@ -1797,9 +1845,9 @@ mkInitialCorePrepEnv :: HscEnv -> IO CorePrepEnv
 mkInitialCorePrepEnv hsc_env = do
    convertNumLit <- mkConvertNumLiteral hsc_env
    return $ CPE
-      { cpe_dynFlags = hsc_dflags hsc_env
-      , cpe_env = emptyVarEnv
-      , cpe_tyco_env = Nothing
+      { cpe_dynFlags      = hsc_dflags hsc_env
+      , cpe_env           = emptyVarEnv
+      , cpe_tyco_env      = Nothing
       , cpe_convertNumLit = convertNumLit
       }
 
@@ -2091,7 +2139,7 @@ collectCostCentres mod_name
       Type{} -> cs
       Coercion{} -> cs
 
-    go_alts = foldl' (\cs (_con, _bndrs, e) -> go cs e)
+    go_alts = foldl' (\cs (Alt _con _bndrs e) -> go cs e)
 
     go_bind :: S.Set CostCentre -> CoreBind -> S.Set CostCentre
     go_bind cs (NonRec b e) =
@@ -2116,10 +2164,11 @@ mkConvertNumLiteral hsc_env = do
    let
       dflags   = hsc_dflags hsc_env
       platform = targetPlatform dflags
+      home_unit = hsc_home_unit hsc_env
       guardBignum act
-         | homeUnitId dflags == primUnitId
+         | isHomeUnitInstanceOf home_unit primUnitId
          = return $ panic "Bignum literals are not supported in ghc-prim"
-         | homeUnitId dflags == bignumUnitId
+         | isHomeUnitInstanceOf home_unit bignumUnitId
          = return $ panic "Bignum literals are not supported in ghc-bignum"
          | otherwise = act
 
@@ -2192,4 +2241,3 @@ mkConvertNumLiteral hsc_env = do
 
 
    return convertNumLit
-
