@@ -9,20 +9,25 @@
 module GHC.StgToCmm.Foreign (
   cgForeignCall,
   emitPrimCall, emitCCall,
-  emitForeignCall,     -- For CmmParse
+  emitForeignCall,
   emitSaveThreadState,
   saveThreadState,
   emitLoadThreadState,
   emitSaveRegs,
   emitRestoreRegs,
+  emitPushTupleRegs,
+  emitPopTupleRegs,
   loadThreadState,
   emitOpenNursery,
   emitCloseNursery,
  ) where
 
-import GhcPrelude hiding( succ, (<*>) )
+import GHC.Prelude hiding( succ, (<*>) )
 
-import StgSyn
+import GHC.Platform
+import GHC.Platform.Profile
+
+import GHC.Stg.Syntax
 import GHC.StgToCmm.Prof (storeCurCCS, ccsType)
 import GHC.StgToCmm.Env
 import GHC.StgToCmm.Monad
@@ -30,25 +35,25 @@ import GHC.StgToCmm.Utils
 import GHC.StgToCmm.Closure
 import GHC.StgToCmm.Layout
 
-import BlockId (newBlockId)
-import Cmm
-import CmmUtils
-import CmmCallConv
-import MkGraph
-import Type
-import RepType
-import CLabel
-import SMRep
-import ForeignCall
-import DynFlags
-import Maybes
-import Outputable
-import UniqSupply
-import BasicTypes
+import GHC.Cmm.BlockId (newBlockId)
+import GHC.Cmm
+import GHC.Cmm.Utils
+import GHC.Cmm.Graph
+import GHC.Cmm.CallConv
+import GHC.Core.Type
+import GHC.Types.RepType
+import GHC.Cmm.CLabel
+import GHC.Runtime.Heap.Layout
+import GHC.Types.ForeignCall
+import GHC.Data.Maybe
+import GHC.Utils.Panic
+import GHC.Types.Unique.Supply
+import GHC.Types.Basic
+import GHC.Unit.Types
 
-import TyCoRep
-import TysPrim
-import Util (zipEqual)
+import GHC.Core.TyCo.Rep
+import GHC.Builtin.Types.Prim
+import GHC.Utils.Misc (zipEqual)
 
 import Control.Monad
 
@@ -65,8 +70,8 @@ cgForeignCall :: ForeignCall            -- the op
               -> Type                   -- result type
               -> FCode ReturnKind
 
-cgForeignCall (CCall (CCallSpec target cconv safety ret_rep arg_reps)) typ stg_args res_ty
-  = do  { dflags <- getDynFlags
+cgForeignCall (CCall (CCallSpec target cconv safety)) typ stg_args res_ty
+  = do  { platform <- getPlatform
         ; let -- in the stdcall calling convention, the symbol needs @size appended
               -- to it, where size is the total number of bytes of arguments.  We
               -- attach this info to the CLabel here, and the CLabel pretty printer
@@ -76,9 +81,11 @@ cgForeignCall (CCall (CCallSpec target cconv safety ret_rep arg_reps)) typ stg_a
               | otherwise            = Nothing
 
               -- ToDo: this might not be correct for 64-bit API
-            arg_size (arg, _) = max (widthInBytes $ typeWidth $ cmmExprType dflags arg)
-                                     (wORD_SIZE dflags)
+              -- This is correct for the PowerPC ELF ABI version 1 and 2.
+            arg_size (arg, _) = max (widthInBytes $ typeWidth $ cmmExprType platform arg)
+                                     (platformWordSizeInBytes platform)
         ; cmm_args <- getFCallArgs stg_args typ
+        -- ; traceM $ show cmm_args
         ; (res_regs, res_hints) <- newUnboxedTupleRegs res_ty
         ; let ((call_args, arg_hints), cmm_target)
                 = case target of
@@ -88,7 +95,7 @@ cgForeignCall (CCall (CCallSpec target cconv safety ret_rep arg_reps)) typ stg_a
                      -> let labelSource
                                 = case mPkgId of
                                         Nothing         -> ForeignLabelInThisPackage
-                                        Just pkgId      -> ForeignLabelInPackage pkgId
+                                        Just pkgId      -> ForeignLabelInPackage (toUnitId pkgId)
                             size = call_size cmm_args
                         in  ( unzip cmm_args
                             , CmmLit (CmmLabel
@@ -97,7 +104,7 @@ cgForeignCall (CCall (CCallSpec target cconv safety ret_rep arg_reps)) typ stg_a
                    DynamicTarget    ->  case cmm_args of
                                            (fn,_):rest -> (unzip rest, fn)
                                            [] -> panic "cgForeignCall []"
-              fc = ForeignConvention cconv arg_hints res_hints CmmMayReturn ret_rep arg_reps
+              fc = ForeignConvention cconv arg_hints res_hints CmmMayReturn
               call_target = ForeignTarget cmm_target fc
 
         -- we want to emit code for the call, and then emitReturn.
@@ -188,29 +195,24 @@ continuation, resulting in just one proc point instead of two. Yay!
 -}
 
 
-emitCCall :: [(CmmFormal, PrimRep, ForeignHint)]
+emitCCall :: [(CmmFormal,ForeignHint)]
           -> CmmExpr
-          -> [(CmmActual, PrimRep, ForeignHint)]
+          -> [(CmmActual,ForeignHint)]
           -> FCode ()
 emitCCall hinted_results fn hinted_args
   = void $ emitForeignCall PlayRisky results target args
   where
-    (args, arg_reps, arg_hints) = unzip3 hinted_args
-    (results, result_reps, result_hints) = unzip3 hinted_results
-    -- extract result, we can only deal with 0 or 1 result types.
-    res_rep = case result_reps of
-      []  -> VoidRep
-      [r] -> r
-      _   -> error "can not deal with multiple return values in emitCCall"
+    (args, arg_hints) = unzip hinted_args
+    (results, result_hints) = unzip hinted_results
     target = ForeignTarget fn fc
-    fc = ForeignConvention CCallConv arg_hints result_hints CmmMayReturn res_rep arg_reps
+    fc = ForeignConvention CCallConv arg_hints result_hints CmmMayReturn
 
 
 emitPrimCall :: [CmmFormal] -> CallishMachOp -> [CmmActual] -> FCode ()
 emitPrimCall res op args
   = void $ emitForeignCall PlayRisky res (PrimTarget op) args
 
--- alternative entry point, used by CmmParse
+-- alternative entry point, used by GHC.Cmm.Parser
 emitForeignCall
         :: Safety
         -> [CmmFormal]          -- where to put the results
@@ -219,8 +221,8 @@ emitForeignCall
         -> FCode ReturnKind
 emitForeignCall safety results target args
   | not (playSafe safety) = do
-    dflags <- getDynFlags
-    let (caller_save, caller_load) = callerSaveVolatileRegs dflags
+    platform <- getPlatform
+    let (caller_save, caller_load) = callerSaveVolatileRegs platform
     emit caller_save
     target' <- load_target_into_temp target
     args' <- mapM maybe_assign_temp args
@@ -229,16 +231,17 @@ emitForeignCall safety results target args
     return AssignedDirectly
 
   | otherwise = do
-    dflags <- getDynFlags
+    profile <- getProfile
+    platform <- getPlatform
     updfr_off <- getUpdFrameOff
     target' <- load_target_into_temp target
     args' <- mapM maybe_assign_temp args
     k <- newBlockId
-    let (off, _, copyout) = copyInOflow dflags NativeReturn (Young k) results []
+    let (off, _, copyout) = copyInOflow profile NativeReturn (Young k) results []
        -- see Note [safe foreign call convention]
     tscope <- getTickScope
     emit $
-           (    mkStore (CmmStackSlot (Young k) (widthInBytes (wordWidth dflags)))
+           (    mkStore (CmmStackSlot (Young k) (widthInBytes (wordWidth platform)))
                         (CmmLit (CmmBlock k))
             <*> mkLast (CmmForeignCall { tgt  = target'
                                        , res  = results
@@ -262,18 +265,18 @@ load_target_into_temp other_target@(PrimTarget _) =
 -- What we want to do here is create a new temporary for the foreign
 -- call argument if it is not safe to use the expression directly,
 -- because the expression mentions caller-saves GlobalRegs (see
--- Note [Register Parameter Passing]).
+-- Note [Register parameter passing]).
 --
 -- However, we can't pattern-match on the expression here, because
--- this is used in a loop by CmmParse, and testing the expression
+-- this is used in a loop by GHC.Cmm.Parser, and testing the expression
 -- results in a black hole.  So we always create a temporary, and rely
--- on CmmSink to clean it up later.  (Yuck, ToDo).  The generated code
+-- on GHC.Cmm.Sink to clean it up later.  (Yuck, ToDo).  The generated code
 -- ends up being the same, at least for the RTS .cmm code.
 --
 maybe_assign_temp :: CmmExpr -> FCode CmmExpr
 maybe_assign_temp e = do
-  dflags <- getDynFlags
-  reg <- newTemp (cmmExprType dflags e)
+  platform <- getPlatform
+  reg <- newTemp (cmmExprType platform e)
   emitAssign (CmmLocal reg) e
   return (CmmReg (CmmLocal reg))
 
@@ -285,31 +288,35 @@ maybe_assign_temp e = do
 
 emitSaveThreadState :: FCode ()
 emitSaveThreadState = do
-  dflags <- getDynFlags
-  code <- saveThreadState dflags
+  profile <- getProfile
+  code <- saveThreadState profile
   emit code
 
 -- | Produce code to save the current thread state to @CurrentTSO@
-saveThreadState :: MonadUnique m => DynFlags -> m CmmAGraph
-saveThreadState dflags = do
-  tso <- newTemp (gcWord dflags)
-  close_nursery <- closeNursery dflags tso
-  pure $ catAGraphs [
-    -- tso = CurrentTSO;
-    mkAssign (CmmLocal tso) currentTSOExpr,
-    -- tso->stackobj->sp = Sp;
-    mkStore (cmmOffset dflags
-                       (CmmLoad (cmmOffset dflags
-                                           (CmmReg (CmmLocal tso))
-                                           (tso_stackobj dflags))
-                                (bWord dflags))
-                       (stack_SP dflags))
-            spExpr,
-    close_nursery,
-    -- and save the current cost centre stack in the TSO when profiling:
-    if gopt Opt_SccProfilingOn dflags then
-        mkStore (cmmOffset dflags (CmmReg (CmmLocal tso)) (tso_CCCS dflags)) cccsExpr
-      else mkNop
+saveThreadState :: MonadUnique m => Profile -> m CmmAGraph
+saveThreadState profile = do
+  let platform = profilePlatform profile
+  tso <- newTemp (gcWord platform)
+  close_nursery <- closeNursery profile tso
+  pure $ catAGraphs
+   [ -- tso = CurrentTSO;
+     mkAssign (CmmLocal tso) currentTSOExpr
+
+   , -- tso->stackobj->sp = Sp;
+     mkStore (cmmOffset platform
+                        (CmmLoad (cmmOffset platform
+                                            (CmmReg (CmmLocal tso))
+                                            (tso_stackobj profile))
+                                 (bWord platform))
+                        (stack_SP profile))
+             spExpr
+
+    , close_nursery
+
+    , -- and save the current cost centre stack in the TSO when profiling:
+      if profileIsProfiling profile
+         then mkStore (cmmOffset platform (CmmReg (CmmLocal tso)) (tso_CCCS profile)) cccsExpr
+         else mkNop
     ]
 
 
@@ -324,25 +331,83 @@ saveThreadState dflags = do
 -- are live, we might have to save them all.
 emitSaveRegs :: FCode ()
 emitSaveRegs = do
-   dflags <- getDynFlags
-   let regs = realArgRegsCover dflags
-       save = catAGraphs (map (callerSaveGlobalReg dflags) regs)
+   platform <- getPlatform
+   let regs = realArgRegsCover platform
+       save = catAGraphs (map (callerSaveGlobalReg platform) regs)
    emit save
 
 -- | Restore STG registers (see 'emitSaveRegs')
 emitRestoreRegs :: FCode ()
 emitRestoreRegs = do
-   dflags <- getDynFlags
-   let regs = realArgRegsCover dflags
-       save = catAGraphs (map (callerRestoreGlobalReg dflags) regs)
-   emit save
+   platform <- getPlatform
+   let regs    = realArgRegsCover platform
+       restore = catAGraphs (map (callerRestoreGlobalReg platform) regs)
+   emit restore
+
+-- | Push a subset of STG registers onto the stack, specified by the bitmap
+--
+-- Sometimes, a "live" subset of the STG registers needs to be saved on the
+-- stack, for example when storing an unboxed tuple to be used in the GHCi
+-- bytecode interpreter.
+--
+-- The "live registers" bitmap corresponds to the list of registers given by
+-- 'tupleRegsCover', with the least significant bit indicating liveness of
+-- the first register in the list.
+--
+-- Each register is saved to a stack slot of one or more machine words, even
+-- if the register size itself is smaller.
+--
+-- The resulting Cmm code looks like this, with a line for each real or
+-- virtual register used for returning tuples:
+--
+--    ...
+--    if((mask & 2) != 0) { Sp_adj(-1); Sp(0) = R2; }
+--    if((mask & 1) != 0) { Sp_adj(-1); Sp(0) = R1; }
+--
+-- See Note [GHCi tuple layout]
+
+emitPushTupleRegs :: CmmExpr -> FCode ()
+emitPushTupleRegs regs_live = do
+  platform <- getPlatform
+  let regs = zip (tupleRegsCover platform) [0..]
+      save_arg (reg, n) =
+        let mask     = CmmLit (CmmInt (1 `shiftL` n) (wordWidth platform))
+            live     = cmmAndWord platform regs_live mask
+            cond     = cmmNeWord platform live (zeroExpr platform)
+            reg_ty   = cmmRegType platform (CmmGlobal reg)
+            width    = roundUpToWords platform
+                                      (widthInBytes $ typeWidth reg_ty)
+            adj_sp   = mkAssign spReg
+                                (cmmOffset platform spExpr (negate width))
+            save_reg = mkStore spExpr (CmmReg $ CmmGlobal reg)
+        in mkCmmIfThen cond $ catAGraphs [adj_sp, save_reg]
+  emit . catAGraphs =<< mapM save_arg (reverse regs)
+
+-- | Pop a subset of STG registers from the stack (see 'emitPushTupleRegs')
+emitPopTupleRegs :: CmmExpr -> FCode ()
+emitPopTupleRegs regs_live = do
+  platform <- getPlatform
+  let regs = zip (tupleRegsCover platform) [0..]
+      save_arg (reg, n) =
+        let mask     = CmmLit (CmmInt (1 `shiftL` n) (wordWidth platform))
+            live     = cmmAndWord platform regs_live mask
+            cond     = cmmNeWord platform live (zeroExpr platform)
+            reg_ty   = cmmRegType platform (CmmGlobal reg)
+            width    = roundUpToWords platform
+                                      (widthInBytes $ typeWidth reg_ty)
+            adj_sp   = mkAssign spReg
+                                (cmmOffset platform spExpr width)
+            restore_reg = mkAssign (CmmGlobal reg) (CmmLoad spExpr reg_ty)
+        in mkCmmIfThen cond $ catAGraphs [restore_reg, adj_sp]
+  emit . catAGraphs =<< mapM save_arg regs
 
 
 emitCloseNursery :: FCode ()
 emitCloseNursery = do
-  dflags <- getDynFlags
-  tso <- newTemp (bWord dflags)
-  code <- closeNursery dflags tso
+  profile <- getProfile
+  let platform = profilePlatform profile
+  tso <- newTemp (bWord platform)
+  code <- closeNursery profile tso
   emit $ mkAssign (CmmLocal tso) currentTSOExpr <*> code
 
 {- |
@@ -366,76 +431,79 @@ Closing the nursery corresponds to the following code:
   cn->free = Hp + WDS(1);
 @
 -}
-closeNursery :: MonadUnique m => DynFlags -> LocalReg -> m CmmAGraph
-closeNursery df tso = do
-  let tsoreg  = CmmLocal tso
-  cnreg      <- CmmLocal <$> newTemp (bWord df)
+closeNursery :: MonadUnique m => Profile -> LocalReg -> m CmmAGraph
+closeNursery profile tso = do
+  let tsoreg   = CmmLocal tso
+      platform = profilePlatform profile
+  cnreg      <- CmmLocal <$> newTemp (bWord platform)
   pure $ catAGraphs [
     mkAssign cnreg currentNurseryExpr,
 
     -- CurrentNursery->free = Hp+1;
-    mkStore (nursery_bdescr_free df cnreg) (cmmOffsetW df hpExpr 1),
+    mkStore (nursery_bdescr_free platform cnreg) (cmmOffsetW platform hpExpr 1),
 
     let alloc =
-           CmmMachOp (mo_wordSub df)
-              [ cmmOffsetW df hpExpr 1
-              , CmmLoad (nursery_bdescr_start df cnreg) (bWord df)
+           CmmMachOp (mo_wordSub platform)
+              [ cmmOffsetW platform hpExpr 1
+              , CmmLoad (nursery_bdescr_start platform cnreg) (bWord platform)
               ]
 
-        alloc_limit = cmmOffset df (CmmReg tsoreg) (tso_alloc_limit df)
+        alloc_limit = cmmOffset platform (CmmReg tsoreg) (tso_alloc_limit profile)
     in
 
     -- tso->alloc_limit += alloc
     mkStore alloc_limit (CmmMachOp (MO_Sub W64)
                                [ CmmLoad alloc_limit b64
-                               , CmmMachOp (mo_WordTo64 df) [alloc] ])
+                               , CmmMachOp (mo_WordTo64 platform) [alloc] ])
    ]
 
 emitLoadThreadState :: FCode ()
 emitLoadThreadState = do
-  dflags <- getDynFlags
-  code <- loadThreadState dflags
+  profile <- getProfile
+  code <- loadThreadState profile
   emit code
 
 -- | Produce code to load the current thread state from @CurrentTSO@
-loadThreadState :: MonadUnique m => DynFlags -> m CmmAGraph
-loadThreadState dflags = do
-  tso <- newTemp (gcWord dflags)
-  stack <- newTemp (gcWord dflags)
-  open_nursery <- openNursery dflags tso
+loadThreadState :: MonadUnique m => Profile -> m CmmAGraph
+loadThreadState profile = do
+  let platform = profilePlatform profile
+  tso <- newTemp (gcWord platform)
+  stack <- newTemp (gcWord platform)
+  open_nursery <- openNursery profile tso
   pure $ catAGraphs [
     -- tso = CurrentTSO;
     mkAssign (CmmLocal tso) currentTSOExpr,
     -- stack = tso->stackobj;
-    mkAssign (CmmLocal stack) (CmmLoad (cmmOffset dflags (CmmReg (CmmLocal tso)) (tso_stackobj dflags)) (bWord dflags)),
+    mkAssign (CmmLocal stack) (CmmLoad (cmmOffset platform (CmmReg (CmmLocal tso)) (tso_stackobj profile)) (bWord platform)),
     -- Sp = stack->sp;
-    mkAssign spReg (CmmLoad (cmmOffset dflags (CmmReg (CmmLocal stack)) (stack_SP dflags)) (bWord dflags)),
+    mkAssign spReg (CmmLoad (cmmOffset platform (CmmReg (CmmLocal stack)) (stack_SP profile)) (bWord platform)),
     -- SpLim = stack->stack + RESERVED_STACK_WORDS;
-    mkAssign spLimReg (cmmOffsetW dflags (cmmOffset dflags (CmmReg (CmmLocal stack)) (stack_STACK dflags))
-                                (rESERVED_STACK_WORDS dflags)),
+    mkAssign spLimReg (cmmOffsetW platform (cmmOffset platform (CmmReg (CmmLocal stack)) (stack_STACK profile))
+                                (pc_RESERVED_STACK_WORDS (platformConstants platform))),
     -- HpAlloc = 0;
     --   HpAlloc is assumed to be set to non-zero only by a failed
     --   a heap check, see HeapStackCheck.cmm:GC_GENERIC
-    mkAssign hpAllocReg (zeroExpr dflags),
+    mkAssign hpAllocReg (zeroExpr platform),
     open_nursery,
     -- and load the current cost centre stack from the TSO when profiling:
-    if gopt Opt_SccProfilingOn dflags
+    if profileIsProfiling profile
        then storeCurCCS
-              (CmmLoad (cmmOffset dflags (CmmReg (CmmLocal tso))
-                 (tso_CCCS dflags)) (ccsType dflags))
+              (CmmLoad (cmmOffset platform (CmmReg (CmmLocal tso))
+                 (tso_CCCS profile)) (ccsType platform))
        else mkNop
    ]
 
 
 emitOpenNursery :: FCode ()
 emitOpenNursery = do
-  dflags <- getDynFlags
-  tso <- newTemp (bWord dflags)
-  code <- openNursery dflags tso
+  profile <- getProfile
+  let platform = profilePlatform profile
+  tso <- newTemp (bWord platform)
+  code <- openNursery profile tso
   emit $ mkAssign (CmmLocal tso) currentTSOExpr <*> code
 
 {- |
-@openNursery dflags tso@ produces code to open the nursery. A local register
+@openNursery profile tso@ produces code to open the nursery. A local register
 holding the value of @CurrentTSO@ is expected for efficiency.
 
 Opening the nursery corresponds to the following code:
@@ -452,7 +520,7 @@ Opening the nursery corresponds to the following code:
    tso->alloc_limit += bdfree - bdstart;
 
    // Set Hp to the last occupied word of the heap block.  Why not the
-   // next unocupied word?  Doing it this way means that we get to use
+   // next unoccupied word?  Doing it this way means that we get to use
    // an offset of zero more often, which might lead to slightly smaller
    // code on some architectures.
    Hp = bdfree - WDS(1);
@@ -462,12 +530,13 @@ Opening the nursery corresponds to the following code:
    HpLim = bdstart + CurrentNursery->blocks*BLOCK_SIZE_W - 1;
 @
 -}
-openNursery :: MonadUnique m => DynFlags -> LocalReg -> m CmmAGraph
-openNursery df tso = do
-  let tsoreg =  CmmLocal tso
-  cnreg      <- CmmLocal <$> newTemp (bWord df)
-  bdfreereg  <- CmmLocal <$> newTemp (bWord df)
-  bdstartreg <- CmmLocal <$> newTemp (bWord df)
+openNursery :: MonadUnique m => Profile -> LocalReg -> m CmmAGraph
+openNursery profile tso = do
+  let tsoreg   = CmmLocal tso
+      platform = profilePlatform profile
+  cnreg      <- CmmLocal <$> newTemp (bWord platform)
+  bdfreereg  <- CmmLocal <$> newTemp (bWord platform)
+  bdstartreg <- CmmLocal <$> newTemp (bWord platform)
 
   -- These assignments are carefully ordered to reduce register
   -- pressure and generate not completely awful code on x86.  To see
@@ -475,23 +544,23 @@ openNursery df tso = do
   -- stg_returnToStackTop in rts/StgStartup.cmm.
   pure $ catAGraphs [
      mkAssign cnreg currentNurseryExpr,
-     mkAssign bdfreereg  (CmmLoad (nursery_bdescr_free df cnreg)  (bWord df)),
+     mkAssign bdfreereg  (CmmLoad (nursery_bdescr_free platform cnreg)  (bWord platform)),
 
      -- Hp = CurrentNursery->free - 1;
-     mkAssign hpReg (cmmOffsetW df (CmmReg bdfreereg) (-1)),
+     mkAssign hpReg (cmmOffsetW platform (CmmReg bdfreereg) (-1)),
 
-     mkAssign bdstartreg (CmmLoad (nursery_bdescr_start df cnreg) (bWord df)),
+     mkAssign bdstartreg (CmmLoad (nursery_bdescr_start platform cnreg) (bWord platform)),
 
      -- HpLim = CurrentNursery->start +
      --              CurrentNursery->blocks*BLOCK_SIZE_W - 1;
      mkAssign hpLimReg
-         (cmmOffsetExpr df
+         (cmmOffsetExpr platform
              (CmmReg bdstartreg)
-             (cmmOffset df
-               (CmmMachOp (mo_wordMul df) [
-                 CmmMachOp (MO_SS_Conv W32 (wordWidth df))
-                   [CmmLoad (nursery_bdescr_blocks df cnreg) b32],
-                 mkIntExpr df (bLOCK_SIZE df)
+             (cmmOffset platform
+               (CmmMachOp (mo_wordMul platform) [
+                 CmmMachOp (MO_SS_Conv W32 (wordWidth platform))
+                   [CmmLoad (nursery_bdescr_blocks platform cnreg) b32],
+                 mkIntExpr platform (pc_BLOCK_SIZE (platformConstants platform))
                 ])
                (-1)
              )
@@ -499,37 +568,37 @@ openNursery df tso = do
 
      -- alloc = bd->free - bd->start
      let alloc =
-           CmmMachOp (mo_wordSub df) [CmmReg bdfreereg, CmmReg bdstartreg]
+           CmmMachOp (mo_wordSub platform) [CmmReg bdfreereg, CmmReg bdstartreg]
 
-         alloc_limit = cmmOffset df (CmmReg tsoreg) (tso_alloc_limit df)
+         alloc_limit = cmmOffset platform (CmmReg tsoreg) (tso_alloc_limit profile)
      in
 
      -- tso->alloc_limit += alloc
      mkStore alloc_limit (CmmMachOp (MO_Add W64)
                                [ CmmLoad alloc_limit b64
-                               , CmmMachOp (mo_WordTo64 df) [alloc] ])
+                               , CmmMachOp (mo_WordTo64 platform) [alloc] ])
 
    ]
 
 nursery_bdescr_free, nursery_bdescr_start, nursery_bdescr_blocks
-  :: DynFlags -> CmmReg -> CmmExpr
-nursery_bdescr_free   dflags cn =
-  cmmOffset dflags (CmmReg cn) (oFFSET_bdescr_free dflags)
-nursery_bdescr_start  dflags cn =
-  cmmOffset dflags (CmmReg cn) (oFFSET_bdescr_start dflags)
-nursery_bdescr_blocks dflags cn =
-  cmmOffset dflags (CmmReg cn) (oFFSET_bdescr_blocks dflags)
+  :: Platform -> CmmReg -> CmmExpr
+nursery_bdescr_free   platform cn =
+  cmmOffset platform (CmmReg cn) (pc_OFFSET_bdescr_free (platformConstants platform))
+nursery_bdescr_start  platform cn =
+  cmmOffset platform (CmmReg cn) (pc_OFFSET_bdescr_start (platformConstants platform))
+nursery_bdescr_blocks platform cn =
+  cmmOffset platform (CmmReg cn) (pc_OFFSET_bdescr_blocks (platformConstants platform))
 
-tso_stackobj, tso_CCCS, tso_alloc_limit, stack_STACK, stack_SP :: DynFlags -> ByteOff
-tso_stackobj dflags = closureField dflags (oFFSET_StgTSO_stackobj dflags)
-tso_alloc_limit dflags = closureField dflags (oFFSET_StgTSO_alloc_limit dflags)
-tso_CCCS     dflags = closureField dflags (oFFSET_StgTSO_cccs dflags)
-stack_STACK  dflags = closureField dflags (oFFSET_StgStack_stack dflags)
-stack_SP     dflags = closureField dflags (oFFSET_StgStack_sp dflags)
+tso_stackobj, tso_CCCS, tso_alloc_limit, stack_STACK, stack_SP :: Profile -> ByteOff
+tso_stackobj    profile = closureField profile (pc_OFFSET_StgTSO_stackobj    (profileConstants profile))
+tso_alloc_limit profile = closureField profile (pc_OFFSET_StgTSO_alloc_limit (profileConstants profile))
+tso_CCCS        profile = closureField profile (pc_OFFSET_StgTSO_cccs        (profileConstants profile))
+stack_STACK     profile = closureField profile (pc_OFFSET_StgStack_stack     (profileConstants profile))
+stack_SP        profile = closureField profile (pc_OFFSET_StgStack_sp        (profileConstants profile))
 
 
-closureField :: DynFlags -> ByteOff -> ByteOff
-closureField dflags off = off + fixedHdrSize dflags
+closureField :: Profile -> ByteOff -> ByteOff
+closureField profile off = off + fixedHdrSize profile
 
 -- Note [Unlifted boxed arguments to foreign calls]
 -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -597,8 +666,8 @@ getFCallArgs args typ
       = return Nothing
       | otherwise
       = do { cmm <- getArgAmode (NonVoid arg)
-           ; dflags <- getDynFlags
-           ; return (Just (add_shim dflags typ cmm, hint)) }
+           ; profile <- getProfile
+           ; return (Just (add_shim profile typ cmm, hint)) }
       where
         arg_ty   = stgArgType arg
         arg_reps = typePrimRep arg_ty
@@ -614,12 +683,14 @@ data StgFArgType
   | StgByteArrayType
 
 -- See Note [Unlifted boxed arguments to foreign calls]
-add_shim :: DynFlags -> StgFArgType -> CmmExpr -> CmmExpr
-add_shim dflags ty expr = case ty of
-  StgPlainType -> expr
-  StgArrayType -> cmmOffsetB dflags expr (arrPtrsHdrSize dflags)
-  StgSmallArrayType -> cmmOffsetB dflags expr (smallArrPtrsHdrSize dflags)
-  StgByteArrayType -> cmmOffsetB dflags expr (arrWordsHdrSize dflags)
+add_shim :: Profile -> StgFArgType -> CmmExpr -> CmmExpr
+add_shim profile ty expr = case ty of
+  StgPlainType      -> expr
+  StgArrayType      -> cmmOffsetB platform expr (arrPtrsHdrSize profile)
+  StgSmallArrayType -> cmmOffsetB platform expr (smallArrPtrsHdrSize profile)
+  StgByteArrayType  -> cmmOffsetB platform expr (arrWordsHdrSize profile)
+  where
+    platform = profilePlatform profile
 
 -- From a function, extract information needed to determine
 -- the offset of each argument when used as a C FFI argument.
