@@ -6,6 +6,8 @@
 --
 -----------------------------------------------------------------------------
 
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
+
 module GHC.StgToCmm.Heap (
         getVirtHp, setVirtHp, setRealHp,
         getHpRelOffset,
@@ -20,10 +22,10 @@ module GHC.StgToCmm.Heap (
         emitSetDynHdr
     ) where
 
-import GhcPrelude hiding ((<*>))
+import GHC.Prelude hiding ((<*>))
 
-import StgSyn
-import CLabel
+import GHC.Stg.Syntax
+import GHC.Cmm.CLabel
 import GHC.StgToCmm.Layout
 import GHC.StgToCmm.Utils
 import GHC.StgToCmm.Monad
@@ -32,20 +34,22 @@ import GHC.StgToCmm.Ticky
 import GHC.StgToCmm.Closure
 import GHC.StgToCmm.Env
 
-import MkGraph
+import GHC.Cmm.Graph
 
-import Hoopl.Label
-import SMRep
-import BlockId
-import Cmm
-import CmmUtils
-import CostCentre
-import IdInfo( CafInfo(..), mayHaveCafRefs )
-import Id ( Id )
-import Module
-import DynFlags
-import FastString( mkFastString, fsLit )
-import Panic( sorry )
+import GHC.Cmm.Dataflow.Label
+import GHC.Runtime.Heap.Layout
+import GHC.Cmm.BlockId
+import GHC.Cmm
+import GHC.Cmm.Utils
+import GHC.Types.CostCentre
+import GHC.Types.Id.Info( CafInfo(..), mayHaveCafRefs )
+import GHC.Types.Id ( Id )
+import GHC.Unit
+import GHC.Driver.Session
+import GHC.Platform
+import GHC.Platform.Profile
+import GHC.Data.FastString( mkFastString, fsLit )
+import GHC.Utils.Panic( sorry )
 
 import Control.Monad (when)
 import Data.Maybe (isJust)
@@ -132,28 +136,28 @@ allocHeapClosure rep info_ptr use_cc payload = do
   hpStore base payload
 
   -- Bump the virtual heap pointer
-  dflags <- getDynFlags
-  setVirtHp (virt_hp + heapClosureSizeW dflags rep)
+  profile <- getProfile
+  setVirtHp (virt_hp + heapClosureSizeW profile rep)
 
   return base
 
 
 emitSetDynHdr :: CmmExpr -> CmmExpr -> CmmExpr -> FCode ()
 emitSetDynHdr base info_ptr ccs
-  = do dflags <- getDynFlags
-       hpStore base (zip (header dflags) [0, wORD_SIZE dflags ..])
+  = do profile <- getProfile
+       hpStore base (zip (header profile) [0, profileWordSizeInBytes profile ..])
   where
-    header :: DynFlags -> [CmmExpr]
-    header dflags = [info_ptr] ++ dynProfHdr dflags ccs
+    header :: Profile -> [CmmExpr]
+    header profile = [info_ptr] ++ dynProfHdr profile ccs
         -- ToDo: Parallel stuff
         -- No ticky header
 
 -- Store the item (expr,off) in base[off]
 hpStore :: CmmExpr -> [(CmmExpr, ByteOff)] -> FCode ()
 hpStore base vals = do
-  dflags <- getDynFlags
+  platform <- getPlatform
   sequence_ $
-    [ emitStore (cmmOffsetB dflags base off) val | (val,off) <- vals ]
+    [ emitStore (cmmOffsetB platform base off) val | (val,off) <- vals ]
 
 -----------------------------------------------------------
 --              Layout of static closures
@@ -163,16 +167,17 @@ hpStore base vals = do
 -- and adding a static link field if necessary.
 
 mkStaticClosureFields
-        :: DynFlags
+        :: Profile
         -> CmmInfoTable
         -> CostCentreStack
         -> CafInfo
         -> [CmmLit]             -- Payload
         -> [CmmLit]             -- The full closure
-mkStaticClosureFields dflags info_tbl ccs caf_refs payload
-  = mkStaticClosure dflags info_lbl ccs payload padding
+mkStaticClosureFields profile info_tbl ccs caf_refs payload
+  = mkStaticClosure profile info_lbl ccs payload padding
         static_link_field saved_info_field
   where
+    platform = profilePlatform profile
     info_lbl = cit_lbl info_tbl
 
     -- CAFs must have consistent layout, regardless of whether they
@@ -190,33 +195,35 @@ mkStaticClosureFields dflags info_tbl ccs caf_refs payload
     is_caf = isThunkRep (cit_rep info_tbl)
 
     padding
-        | is_caf && null payload = [mkIntCLit dflags 0]
+        | is_caf && null payload = [mkIntCLit platform 0]
         | otherwise = []
 
     static_link_field
-        | is_caf || staticClosureNeedsLink (mayHaveCafRefs caf_refs) info_tbl
+        | is_caf
+        = [mkIntCLit platform 0]
+        | staticClosureNeedsLink (mayHaveCafRefs caf_refs) info_tbl
         = [static_link_value]
         | otherwise
         = []
 
     saved_info_field
-        | is_caf     = [mkIntCLit dflags 0]
+        | is_caf     = [mkIntCLit platform 0]
         | otherwise  = []
 
         -- For a static constructor which has NoCafRefs, we set the
         -- static link field to a non-zero value so the garbage
         -- collector will ignore it.
     static_link_value
-        | mayHaveCafRefs caf_refs  = mkIntCLit dflags 0
-        | otherwise                = mkIntCLit dflags 3  -- No CAF refs
+        | mayHaveCafRefs caf_refs  = mkIntCLit platform 0
+        | otherwise                = mkIntCLit platform 3  -- No CAF refs
                                       -- See Note [STATIC_LINK fields]
                                       -- in rts/sm/Storage.h
 
-mkStaticClosure :: DynFlags -> CLabel -> CostCentreStack -> [CmmLit]
+mkStaticClosure :: Profile -> CLabel -> CostCentreStack -> [CmmLit]
   -> [CmmLit] -> [CmmLit] -> [CmmLit] -> [CmmLit]
-mkStaticClosure dflags info_lbl ccs payload padding static_link_field saved_info_field
+mkStaticClosure profile info_lbl ccs payload padding static_link_field saved_info_field
   =  [CmmLabel info_lbl]
-  ++ staticProfHdr dflags ccs
+  ++ staticProfHdr profile ccs
   ++ payload
   ++ padding
   ++ static_link_field
@@ -326,18 +333,20 @@ entryHeapCheck :: ClosureInfo
                -> FCode ()
                -> FCode ()
 
-entryHeapCheck cl_info nodeSet arity args code
-  = entryHeapCheck' is_fastf node arity args code
-  where
+entryHeapCheck cl_info nodeSet arity args code = do
+  platform <- getPlatform
+  let
     node = case nodeSet of
               Just r  -> CmmReg (CmmLocal r)
-              Nothing -> CmmLit (CmmLabel $ staticClosureLabel cl_info)
+              Nothing -> CmmLit (CmmLabel $ staticClosureLabel platform cl_info)
 
     is_fastf = case closureFunInfo cl_info of
                  Just (_, ArgGen _) -> False
                  _otherwise         -> True
 
--- | lower-level version for CmmParse
+  entryHeapCheck' is_fastf node arity args code
+
+-- | lower-level version for "GHC.Cmm.Parser"
 entryHeapCheck' :: Bool           -- is a known function pattern
                 -> CmmExpr        -- expression for the closure pointer
                 -> Int            -- Arity -- not same as len args b/c of voids
@@ -345,7 +354,7 @@ entryHeapCheck' :: Bool           -- is a known function pattern
                 -> FCode ()
                 -> FCode ()
 entryHeapCheck' is_fastf node arity args code
-  = do dflags <- getDynFlags
+  = do profile <- getProfile
        let is_thunk = arity == 0
 
            args' = map (CmmReg . CmmLocal) args
@@ -360,13 +369,13 @@ entryHeapCheck' is_fastf node arity args code
            -}
            gc_call upd
                | is_thunk
-                 = mkJump dflags NativeNodeCall stg_gc_enter1 [node] upd
+                 = mkJump profile NativeNodeCall stg_gc_enter1 [node] upd
 
                | is_fastf
-                 = mkJump dflags NativeNodeCall stg_gc_fun (node : args') upd
+                 = mkJump profile NativeNodeCall stg_gc_fun (node : args') upd
 
                | otherwise
-                 = mkJump dflags Slow stg_gc_fun (node : args') upd
+                 = mkJump profile Slow stg_gc_fun (node : args') upd
 
        updfr_sz <- getUpdFrameOff
 
@@ -397,12 +406,13 @@ altHeapCheck regs code = altOrNoEscapeHeapCheck False regs code
 
 altOrNoEscapeHeapCheck :: Bool -> [LocalReg] -> FCode a -> FCode a
 altOrNoEscapeHeapCheck checkYield regs code = do
-    dflags <- getDynFlags
-    case cannedGCEntryPoint dflags regs of
+    profile <- getProfile
+    platform <- getPlatform
+    case cannedGCEntryPoint platform regs of
       Nothing -> genericGC checkYield code
       Just gc -> do
         lret <- newBlockId
-        let (off, _, copyin) = copyInOflow dflags NativeReturn (Young lret) regs []
+        let (off, _, copyin) = copyInOflow profile NativeReturn (Young lret) regs []
         lcont <- newBlockId
         tscope <- getTickScope
         emitOutOfLine lret (copyin <*> mkBranch lcont, tscope)
@@ -411,8 +421,8 @@ altOrNoEscapeHeapCheck checkYield regs code = do
 
 altHeapCheckReturnsTo :: [LocalReg] -> Label -> ByteOff -> FCode a -> FCode a
 altHeapCheckReturnsTo regs lret off code
-  = do dflags <- getDynFlags
-       case cannedGCEntryPoint dflags regs of
+  = do platform <- getPlatform
+       case cannedGCEntryPoint platform regs of
            Nothing -> genericGC False code
            Just gc -> cannedGCReturnsTo False True gc regs lret off code
 
@@ -426,9 +436,9 @@ cannedGCReturnsTo :: Bool -> Bool -> CmmExpr -> [LocalReg] -> Label -> ByteOff
                   -> FCode a
                   -> FCode a
 cannedGCReturnsTo checkYield cont_on_stack gc regs lret off code
-  = do dflags <- getDynFlags
+  = do profile <- getProfile
        updfr_sz <- getUpdFrameOff
-       heapCheck False checkYield (gc_call dflags gc updfr_sz) code
+       heapCheck False checkYield (gc_call profile gc updfr_sz) code
   where
     reg_exprs = map (CmmReg . CmmLocal) regs
       -- Note [stg_gc arguments]
@@ -437,11 +447,11 @@ cannedGCReturnsTo checkYield cont_on_stack gc regs lret off code
       -- to the canned heap-check routines, because we are in a case
       -- alternative and hence the [LocalReg] was passed to us in the
       -- NativeReturn convention.
-    gc_call dflags label sp
+    gc_call profile label sp
       | cont_on_stack
-      = mkJumpReturnsTo dflags label NativeReturn reg_exprs lret off sp
+      = mkJumpReturnsTo profile label NativeReturn reg_exprs lret off sp
       | otherwise
-      = mkCallReturnsTo dflags label NativeReturn reg_exprs lret off sp []
+      = mkCallReturnsTo profile label NativeReturn reg_exprs lret off sp []
 
 genericGC :: Bool -> FCode a -> FCode a
 genericGC checkYield code
@@ -451,8 +461,8 @@ genericGC checkYield code
        call <- mkCall generic_gc (GC, GC) [] [] updfr_sz []
        heapCheck False checkYield (call <*> mkBranch lretry) code
 
-cannedGCEntryPoint :: DynFlags -> [LocalReg] -> Maybe CmmExpr
-cannedGCEntryPoint dflags regs
+cannedGCEntryPoint :: Platform -> [LocalReg] -> Maybe CmmExpr
+cannedGCEntryPoint platform regs
   = case map localRegType regs of
       []  -> Just (mkGcLabel "stg_gc_noregs")
       [ty]
@@ -462,9 +472,9 @@ cannedGCEntryPoint dflags regs
                                   W64       -> Just (mkGcLabel "stg_gc_d1")
                                   _         -> Nothing
 
-          | width == wordWidth dflags -> Just (mkGcLabel "stg_gc_unbx_r1")
-          | width == W64              -> Just (mkGcLabel "stg_gc_l1")
-          | otherwise                 -> Nothing
+          | width == wordWidth platform -> Just (mkGcLabel "stg_gc_unbx_r1")
+          | width == W64                -> Just (mkGcLabel "stg_gc_l1")
+          | otherwise                   -> Nothing
           where
               width = typeWidth ty
       [ty1,ty2]
@@ -513,7 +523,7 @@ heapCheck checkStack checkYield do_gc code
   = getHeapUsage $ \ hpHw ->
     -- Emit heap checks, but be sure to do it lazily so
     -- that the conditionals on hpHw don't cause a black hole
-    do  { dflags <- getDynFlags
+    do  { platform <- getPlatform
         ; let mb_alloc_bytes
                  | hpHw > mBLOCK_SIZE = sorry $ unlines
                     [" Trying to allocate more than "++show mBLOCK_SIZE++" bytes.",
@@ -522,9 +532,12 @@ heapCheck checkStack checkYield do_gc code
                      "See https://gitlab.haskell.org/ghc/ghc/issues/4505 for details.",
                      "Suggestion: read data from a file instead of having large static data",
                      "structures in code."]
-                 | hpHw > 0  = Just (mkIntExpr dflags (hpHw * (wORD_SIZE dflags)))
+                 | hpHw > 0  = Just (mkIntExpr platform (hpHw * (platformWordSizeInBytes platform)))
                  | otherwise = Nothing
-                 where mBLOCK_SIZE = bLOCKS_PER_MBLOCK dflags * bLOCK_SIZE_W dflags
+                 where
+                  constants = platformConstants platform
+                  bLOCK_SIZE_W = pc_BLOCK_SIZE (platformConstants platform) `quot` platformWordSizeInBytes platform
+                  mBLOCK_SIZE = pc_BLOCKS_PER_MBLOCK constants * bLOCK_SIZE_W
               stk_hwm | checkStack = Just (CmmLit CmmHighStackMark)
                       | otherwise  = Nothing
         ; codeOnly $ do_checks stk_hwm checkYield mb_alloc_bytes do_gc
@@ -594,26 +607,27 @@ do_checks :: Maybe CmmExpr    -- Should we check the stack?
           -> FCode ()
 do_checks mb_stk_hwm checkYield mb_alloc_lit do_gc = do
   dflags <- getDynFlags
+  platform <- getPlatform
   gc_id <- newBlockId
 
   let
     Just alloc_lit = mb_alloc_lit
 
-    bump_hp   = cmmOffsetExprB dflags hpExpr alloc_lit
+    bump_hp   = cmmOffsetExprB platform hpExpr alloc_lit
 
     -- Sp overflow if ((old + 0) - CmmHighStack < SpLim)
     -- At the beginning of a function old + 0 = Sp
     -- See Note [Single stack check]
     sp_oflo sp_hwm =
-         CmmMachOp (mo_wordULt dflags)
-                  [CmmMachOp (MO_Sub (typeWidth (cmmRegType dflags spReg)))
+         CmmMachOp (mo_wordULt platform)
+                  [CmmMachOp (MO_Sub (typeWidth (cmmRegType platform spReg)))
                              [CmmStackSlot Old 0, sp_hwm],
                    CmmReg spLimReg]
 
     -- Hp overflow if (Hp > HpLim)
     -- (Hp has been incremented by now)
     -- HpLim points to the LAST WORD of valid allocation space.
-    hp_oflo = CmmMachOp (mo_wordUGt dflags) [hpExpr, hpLimExpr]
+    hp_oflo = CmmMachOp (mo_wordUGt platform) [hpExpr, hpLimExpr]
 
     alloc_n = mkAssign hpAllocReg alloc_lit
 
@@ -636,12 +650,12 @@ do_checks mb_stk_hwm checkYield mb_alloc_lit do_gc = do
      tickyHeapCheck
      emitAssign hpReg bump_hp
      emit =<< mkCmmIfThen' hp_oflo (alloc_n <*> mkBranch gc_id) (Just False)
-    else do
+    else
       when (checkYield && not (gopt Opt_OmitYields dflags)) $ do
          -- Yielding if HpLim == 0
-         let yielding = CmmMachOp (mo_wordEq dflags)
+         let yielding = CmmMachOp (mo_wordEq platform)
                                   [CmmReg hpLimReg,
-                                   CmmLit (zeroCLit dflags)]
+                                   CmmLit (zeroCLit platform)]
          emit =<< mkCmmIfGoto' yielding gc_id (Just False)
 
   tscope <- getTickScope
