@@ -9,18 +9,20 @@
 module GHC.StgToCmm.Hpc ( initHpc, mkTickBox ) where
 
 import GHC.Prelude
+import GHC.Platform
+
+import GHC.Driver.Session
 
 import GHC.StgToCmm.Monad
+import GHC.StgToCmm.Utils
 
-import GHC.Platform
 import GHC.Cmm.Graph
 import GHC.Cmm.Expr
 import GHC.Cmm.CLabel
-import GHC.Unit.Module
 import GHC.Cmm.Utils
-import GHC.StgToCmm.Utils
-import GHC.Driver.Types
-import GHC.Driver.Session
+
+import GHC.Unit.Module
+import GHC.Types.HpcInfo
 
 import Control.Monad
 
