@@ -32,7 +32,10 @@ module GHC.StgToCmm.Layout (
 
 #include "HsVersions.h"
 
-import GhcPrelude hiding ((<*>))
+import GHC.Prelude hiding ((<*>))
+
+import GHC.Driver.Session
+import GHC.Driver.Ppr
 
 import GHC.StgToCmm.Closure
 import GHC.StgToCmm.Env
@@ -41,24 +44,26 @@ import GHC.StgToCmm.Ticky
 import GHC.StgToCmm.Monad
 import GHC.StgToCmm.Utils
 
-import MkGraph
-import SMRep
-import BlockId
-import Cmm
-import CmmUtils
-import CmmInfo
-import CLabel
-import StgSyn
-import Id
-import TyCon             ( PrimRep(..), primRepSizeB )
-import BasicTypes        ( RepArity )
-import DynFlags
-import Module
+import GHC.Cmm.Graph
+import GHC.Runtime.Heap.Layout
+import GHC.Cmm.BlockId
+import GHC.Cmm
+import GHC.Cmm.Utils
+import GHC.Cmm.Info
+import GHC.Cmm.CLabel
+import GHC.Stg.Syntax
+import GHC.Types.Id
+import GHC.Core.TyCon    ( PrimRep(..), primRepSizeB )
+import GHC.Types.Basic   ( RepArity )
+import GHC.Platform
+import GHC.Platform.Profile
+import GHC.Unit
 
-import Util
-import Data.List
-import Outputable
-import FastString
+import GHC.Utils.Misc
+import Data.List (mapAccumL, partition)
+import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Data.FastString
 import Control.Monad
 
 ------------------------------------------------------------------------
@@ -77,14 +82,15 @@ import Control.Monad
 --
 emitReturn :: [CmmExpr] -> FCode ReturnKind
 emitReturn results
-  = do { dflags    <- getDynFlags
+  = do { profile   <- getProfile
+       ; platform  <- getPlatform
        ; sequel    <- getSequel
        ; updfr_off <- getUpdFrameOff
        ; case sequel of
            Return ->
              do { adjustHpBackwards
-                ; let e = CmmLoad (CmmStackSlot Old updfr_off) (gcWord dflags)
-                ; emit (mkReturn dflags (entryCode dflags e) results updfr_off)
+                ; let e = CmmLoad (CmmStackSlot Old updfr_off) (gcWord platform)
+                ; emit (mkReturn profile (entryCode platform e) results updfr_off)
                 }
            AssignTo regs adjust ->
              do { when adjust adjustHpBackwards
@@ -111,19 +117,19 @@ emitCallWithExtraStack
    :: (Convention, Convention) -> CmmExpr -> [CmmExpr]
    -> [CmmExpr] -> FCode ReturnKind
 emitCallWithExtraStack (callConv, retConv) fun args extra_stack
-  = do  { dflags <- getDynFlags
+  = do  { profile <- getProfile
         ; adjustHpBackwards
         ; sequel <- getSequel
         ; updfr_off <- getUpdFrameOff
         ; case sequel of
             Return -> do
-              emit $ mkJumpExtra dflags callConv fun args updfr_off extra_stack
+              emit $ mkJumpExtra profile callConv fun args updfr_off extra_stack
               return AssignedDirectly
             AssignTo res_regs _ -> do
               k <- newBlockId
               let area = Young k
-                  (off, _, copyin) = copyInOflow dflags retConv area res_regs []
-                  copyout = mkCallReturnsTo dflags fun callConv args k off updfr_off
+                  (off, _, copyin) = copyInOflow profile retConv area res_regs []
+                  copyout = mkCallReturnsTo profile fun callConv args k off updfr_off
                                    extra_stack
               tscope <- getTickScope
               emit (copyout <*> mkLabel k tscope <*> copyin)
@@ -189,6 +195,8 @@ slowCall :: CmmExpr -> [StgArg] -> FCode ReturnKind
 -- (slowCall fun args) applies fun to args, returning the results to Sequel
 slowCall fun stg_args
   = do  dflags <- getDynFlags
+        profile <- getProfile
+        let platform = profilePlatform profile
         argsreps <- getArgRepsAmodes stg_args
         let (rts_fun, arity) = slowCallPattern (map fst argsreps)
 
@@ -196,7 +204,7 @@ slowCall fun stg_args
            r <- direct_call "slow_call" NativeNodeCall
                  (mkRtsApFastLabel rts_fun) arity ((P,Just fun):argsreps)
            emitComment $ mkFastString ("slow_call for " ++
-                                      showSDoc dflags (ppr fun) ++
+                                      showSDoc dflags (pdoc platform fun) ++
                                       " with pat " ++ unpackFS rts_fun)
            return r
 
@@ -204,9 +212,10 @@ slowCall fun stg_args
         let n_args = length stg_args
         if n_args > arity && optLevel dflags >= 2
            then do
+             ptr_opts <- getPtrOpts
              funv <- (CmmReg . CmmLocal) `fmap` assignTemp fun
              fun_iptr <- (CmmReg . CmmLocal) `fmap`
-                    assignTemp (closureInfoPtr dflags (cmmUntag dflags funv))
+                    assignTemp (closureInfoPtr ptr_opts (cmmUntag platform funv))
 
              -- ToDo: we could do slightly better here by reusing the
              -- continuation from the slow call, which we have in r.
@@ -219,7 +228,7 @@ slowCall fun stg_args
 
              fast_code <- getCode $
                 emitCall (NativeNodeCall, NativeReturn)
-                  (entryCode dflags fun_iptr)
+                  (entryCode platform fun_iptr)
                   (nonVArgs ((P,Just funv):argsreps))
 
              slow_lbl <- newBlockId
@@ -227,11 +236,11 @@ slowCall fun stg_args
              is_tagged_lbl <- newBlockId
              end_lbl <- newBlockId
 
-             let correct_arity = cmmEqWord dflags (funInfoArity dflags fun_iptr)
-                                                  (mkIntExpr dflags n_args)
+             let correct_arity = cmmEqWord platform (funInfoArity profile fun_iptr)
+                                                    (mkIntExpr platform n_args)
 
              tscope <- getTickScope
-             emit (mkCbranch (cmmIsTagged dflags funv)
+             emit (mkCbranch (cmmIsTagged platform funv)
                              is_tagged_lbl slow_lbl (Just True)
                    <*> mkLabel is_tagged_lbl tscope
                    <*> mkCbranch correct_arity fast_lbl slow_lbl (Just True)
@@ -282,10 +291,11 @@ direct_call :: String
 direct_call caller call_conv lbl arity args
   | debugIsOn && args `lengthLessThan` real_arity  -- Too few args
   = do -- Caller should ensure that there enough args!
+       platform <- getPlatform
        pprPanic "direct_call" $
             text caller <+> ppr arity <+>
-            ppr lbl <+> ppr (length args) <+>
-            ppr (map snd args) <+> ppr (map fst args)
+            pdoc platform lbl <+> ppr (length args) <+>
+            pdoc platform (map snd args) <+> ppr (map fst args)
 
   | null rest_args  -- Precisely the right number of arguments
   = emitCall (call_conv, NativeReturn) target (nonVArgs args)
@@ -311,12 +321,14 @@ direct_call caller call_conv lbl arity args
 -- using zeroCLit or even undefined would work, but would be ugly).
 --
 getArgRepsAmodes :: [StgArg] -> FCode [(ArgRep, Maybe CmmExpr)]
-getArgRepsAmodes = mapM getArgRepAmode
-  where getArgRepAmode arg
+getArgRepsAmodes args = do
+   platform <- profilePlatform <$> getProfile
+   mapM (getArgRepAmode platform) args
+  where getArgRepAmode platform arg
            | V <- rep  = return (V, Nothing)
            | otherwise = do expr <- getArgAmode (NonVoid arg)
                             return (rep, Just expr)
-           where rep = toArgRep (argPrimRep arg)
+           where rep = toArgRep platform (argPrimRep arg)
 
 nonVArgs :: [(ArgRep, Maybe CmmExpr)] -> [CmmExpr]
 nonVArgs [] = []
@@ -364,7 +376,7 @@ just more arguments that we are passing on the stack (cml_args).
 slowArgs :: DynFlags -> [(ArgRep, Maybe CmmExpr)] -> [(ArgRep, Maybe CmmExpr)]
 slowArgs _ [] = []
 slowArgs dflags args -- careful: reps contains voids (V), but args does not
-  | gopt Opt_SccProfilingOn dflags
+  | sccProfilingEnabled dflags
               = save_cccs ++ this_pat ++ slowArgs dflags rest_args
   | otherwise =              this_pat ++ slowArgs dflags rest_args
   where
@@ -389,9 +401,9 @@ hpRel hp off = off - hp
 getHpRelOffset :: VirtualHpOffset -> FCode CmmExpr
 -- See Note [Virtual and real heap pointers] in GHC.StgToCmm.Monad
 getHpRelOffset virtual_offset
-  = do dflags <- getDynFlags
+  = do platform <- getPlatform
        hp_usg <- getHpUsage
-       return (cmmRegOffW dflags hpReg (hpRel (realHp hp_usg) virtual_offset))
+       return (cmmRegOffW platform hpReg (hpRel (realHp hp_usg) virtual_offset))
 
 data FieldOffOrPadding a
     = FieldOff (NonVoid a) -- Something that needs an offset.
@@ -408,7 +420,7 @@ data ClosureHeader
   | ThunkHeader
 
 mkVirtHeapOffsetsWithPadding
-  :: DynFlags
+  :: Profile
   -> ClosureHeader            -- What kind of header to account for
   -> [NonVoid (PrimRep, a)]   -- Things to make offsets for
   -> ( WordOff                -- Total number of words allocated
@@ -423,18 +435,19 @@ mkVirtHeapOffsetsWithPadding
 -- mkVirtHeapOffsetsWithPadding always returns boxed things with smaller offsets
 -- than the unboxed things
 
-mkVirtHeapOffsetsWithPadding dflags header things =
+mkVirtHeapOffsetsWithPadding profile header things =
     ASSERT(not (any (isVoidRep . fst . fromNonVoid) things))
     ( tot_wds
-    , bytesToWordsRoundUp dflags bytes_of_ptrs
+    , bytesToWordsRoundUp platform bytes_of_ptrs
     , concat (ptrs_w_offsets ++ non_ptrs_w_offsets) ++ final_pad
     )
   where
+    platform = profilePlatform profile
     hdr_words = case header of
       NoHeader -> 0
-      StdHeader -> fixedHdrSizeW dflags
-      ThunkHeader -> thunkHdrSize dflags
-    hdr_bytes = wordsToBytes dflags hdr_words
+      StdHeader -> fixedHdrSizeW profile
+      ThunkHeader -> thunkHdrSize profile
+    hdr_bytes = wordsToBytes platform hdr_words
 
     (ptrs, non_ptrs) = partition (isGcPtrRep . fst . fromNonVoid) things
 
@@ -443,7 +456,7 @@ mkVirtHeapOffsetsWithPadding dflags header things =
     (tot_bytes, non_ptrs_w_offsets) =
        mapAccumL computeOffset bytes_of_ptrs non_ptrs
 
-    tot_wds = bytesToWordsRoundUp dflags tot_bytes
+    tot_wds = bytesToWordsRoundUp platform tot_bytes
 
     final_pad_size = tot_wds * word_size - tot_bytes
     final_pad
@@ -451,7 +464,7 @@ mkVirtHeapOffsetsWithPadding dflags header things =
                                          (hdr_bytes + tot_bytes))]
         | otherwise          = []
 
-    word_size = wORD_SIZE dflags
+    word_size = platformWordSizeInBytes platform
 
     computeOffset bytes_so_far nv_thing =
         (new_bytes_so_far, with_padding field_off)
@@ -459,7 +472,7 @@ mkVirtHeapOffsetsWithPadding dflags header things =
         (rep, thing) = fromNonVoid nv_thing
 
         -- Size of the field in bytes.
-        !sizeB = primRepSizeB dflags rep
+        !sizeB = primRepSizeB platform rep
 
         -- Align the start offset (eg, 2-byte value should be 2-byte aligned).
         -- But not more than to a word.
@@ -481,36 +494,36 @@ mkVirtHeapOffsetsWithPadding dflags header things =
 
 
 mkVirtHeapOffsets
-  :: DynFlags
+  :: Profile
   -> ClosureHeader            -- What kind of header to account for
   -> [NonVoid (PrimRep,a)]    -- Things to make offsets for
   -> (WordOff,                -- _Total_ number of words allocated
       WordOff,                -- Number of words allocated for *pointers*
       [(NonVoid a, ByteOff)])
-mkVirtHeapOffsets dflags header things =
+mkVirtHeapOffsets profile header things =
     ( tot_wds
     , ptr_wds
     , [ (field, offset) | (FieldOff field offset) <- things_offsets ]
     )
   where
    (tot_wds, ptr_wds, things_offsets) =
-       mkVirtHeapOffsetsWithPadding dflags header things
+       mkVirtHeapOffsetsWithPadding profile header things
 
 -- | Just like mkVirtHeapOffsets, but for constructors
 mkVirtConstrOffsets
-  :: DynFlags -> [NonVoid (PrimRep, a)]
+  :: Profile -> [NonVoid (PrimRep, a)]
   -> (WordOff, WordOff, [(NonVoid a, ByteOff)])
-mkVirtConstrOffsets dflags = mkVirtHeapOffsets dflags StdHeader
+mkVirtConstrOffsets profile = mkVirtHeapOffsets profile StdHeader
 
 -- | Just like mkVirtConstrOffsets, but used when we don't have the actual
 -- arguments. Useful when e.g. generating info tables; we just need to know
 -- sizes of pointer and non-pointer fields.
-mkVirtConstrSizes :: DynFlags -> [NonVoid PrimRep] -> (WordOff, WordOff)
-mkVirtConstrSizes dflags field_reps
+mkVirtConstrSizes :: Profile -> [NonVoid PrimRep] -> (WordOff, WordOff)
+mkVirtConstrSizes profile field_reps
   = (tot_wds, ptr_wds)
   where
     (tot_wds, ptr_wds, _) =
-       mkVirtConstrOffsets dflags
+       mkVirtConstrOffsets profile
          (map (\nv_rep -> NonVoid (fromNonVoid nv_rep, ())) field_reps)
 
 -------------------------------------------------------------------------
@@ -528,20 +541,20 @@ mkVirtConstrSizes dflags field_reps
 -- bring in ARG_P, ARG_N, etc.
 #include "../includes/rts/storage/FunTypes.h"
 
-mkArgDescr :: DynFlags -> [Id] -> ArgDescr
-mkArgDescr dflags args
-  = let arg_bits = argBits dflags arg_reps
-        arg_reps = filter isNonV (map idArgRep args)
+mkArgDescr :: Platform -> [Id] -> ArgDescr
+mkArgDescr platform args
+  = let arg_bits = argBits platform arg_reps
+        arg_reps = filter isNonV (map (idArgRep platform) args)
            -- Getting rid of voids eases matching of standard patterns
     in case stdPattern arg_reps of
          Just spec_id -> ArgSpec spec_id
          Nothing      -> ArgGen  arg_bits
 
-argBits :: DynFlags -> [ArgRep] -> [Bool]        -- True for non-ptr, False for ptr
-argBits _      []           = []
-argBits dflags (P   : args) = False : argBits dflags args
-argBits dflags (arg : args) = take (argRepSizeW dflags arg) (repeat True)
-                    ++ argBits dflags args
+argBits :: Platform -> [ArgRep] -> [Bool]        -- True for non-ptr, False for ptr
+argBits _         []           = []
+argBits platform (P   : args) = False : argBits platform args
+argBits platform (arg : args) = take (argRepSizeW platform arg) (repeat True)
+                                 ++ argBits platform args
 
 ----------------------
 stdPattern :: [ArgRep] -> Maybe Int
@@ -597,27 +610,28 @@ emitClosureProcAndInfoTable :: Bool                    -- top-level?
                             -> ((Int, LocalReg, [LocalReg]) -> FCode ()) -- function body
                             -> FCode ()
 emitClosureProcAndInfoTable top_lvl bndr lf_info info_tbl args body
- = do   { dflags <- getDynFlags
+ = do   { profile <- getProfile
+        ; platform <- getPlatform
         -- Bind the binder itself, but only if it's not a top-level
         -- binding. We need non-top let-bindings to refer to the
         -- top-level binding, which this binding would incorrectly shadow.
-        ; node <- if top_lvl then return $ idToReg dflags (NonVoid bndr)
+        ; node <- if top_lvl then return $ idToReg platform (NonVoid bndr)
                   else bindToReg (NonVoid bndr) lf_info
-        ; let node_points = nodeMustPointToIt dflags lf_info
+        ; let node_points = nodeMustPointToIt profile lf_info
         ; arg_regs <- bindArgsToRegs args
         ; let args' = if node_points then (node : arg_regs) else arg_regs
-              conv  = if nodeMustPointToIt dflags lf_info then NativeNodeCall
+              conv  = if nodeMustPointToIt profile lf_info then NativeNodeCall
                                                           else NativeDirectCall
-              (offset, _, _) = mkCallEntry dflags conv args' []
-        ; emitClosureAndInfoTable info_tbl conv args' $ body (offset, node, arg_regs)
+              (offset, _, _) = mkCallEntry profile conv args' []
+        ; emitClosureAndInfoTable (profilePlatform profile) info_tbl conv args' $ body (offset, node, arg_regs)
         }
 
 -- Data constructors need closures, but not with all the argument handling
 -- needed for functions. The shared part goes here.
-emitClosureAndInfoTable ::
-  CmmInfoTable -> Convention -> [LocalReg] -> FCode () -> FCode ()
-emitClosureAndInfoTable info_tbl conv args body
+emitClosureAndInfoTable
+   :: Platform -> CmmInfoTable -> Convention -> [LocalReg] -> FCode () -> FCode ()
+emitClosureAndInfoTable platform info_tbl conv args body
   = do { (_, blks) <- getCodeScoped body
-       ; let entry_lbl = toEntryLbl (cit_lbl info_tbl)
+       ; let entry_lbl = toEntryLbl platform (cit_lbl info_tbl)
        ; emitProcWithConvention conv (Just info_tbl) entry_lbl args blks
        }
