@@ -1,5 +1,9 @@
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE GADTs #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternSynonyms #-}
+
 
 -----------------------------------------------------------------------------
 --
@@ -22,8 +26,9 @@ module GHC.StgToCmm.Monad (
         emitOutOfLine, emitAssign, emitStore,
         emitComment, emitTick, emitUnwind,
 
-        getCmm, aGraphToGraph,
+        getCmm, aGraphToGraph, getPlatform, getProfile,
         getCodeR, getCode, getCodeScoped, getHeapUsage,
+        getCallOpts, getPtrOpts,
 
         mkCmmIfThenElse, mkCmmIfThen, mkCmmIfGoto,
         mkCmmIfThenElse', mkCmmIfThen', mkCmmIfGoto',
@@ -49,40 +54,43 @@ module GHC.StgToCmm.Monad (
         getModuleName,
 
         -- ideally we wouldn't export these, but some other modules access internal state
-        getState, setState, getSelfLoop, withSelfLoop, getInfoDown, getDynFlags, getThisPackage,
+        getState, setState, getSelfLoop, withSelfLoop, getInfoDown, getDynFlags,
 
         -- more localised access to monad state
         CgIdInfo(..),
         getBinds, setBinds,
-
         -- out of general friendliness, we also export ...
-        CgInfoDownwards(..), CgState(..)        -- non-abstract
+        CgInfoDownwards(..), CgState(..) -- non-abstract
     ) where
 
-import GhcPrelude hiding( sequence, succ )
+import GHC.Prelude hiding( sequence, succ )
 
-import Cmm
+import GHC.Platform
+import GHC.Platform.Profile
+import GHC.Cmm
 import GHC.StgToCmm.Closure
-import DynFlags
-import Hoopl.Collections
-import MkGraph
-import BlockId
-import CLabel
-import SMRep
-import Module
-import Id
-import VarEnv
-import OrdList
-import BasicTypes( ConTagZ )
-import Unique
-import UniqSupply
-import FastString
-import Outputable
-import Util
+import GHC.Driver.Session
+import GHC.Cmm.Dataflow.Collections
+import GHC.Cmm.Graph as CmmGraph
+import GHC.Cmm.BlockId
+import GHC.Cmm.CLabel
+import GHC.Cmm.Info
+import GHC.Runtime.Heap.Layout
+import GHC.Unit
+import GHC.Types.Id
+import GHC.Types.Var.Env
+import GHC.Data.OrdList
+import GHC.Types.Basic( ConTagZ )
+import GHC.Types.Unique
+import GHC.Types.Unique.Supply
+import GHC.Data.FastString
+import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Utils.Misc
+import GHC.Exts (oneShot)
 
 import Control.Monad
-import Data.List
-
+import Data.List (mapAccumL)
 
 
 --------------------------------------------------------
@@ -111,8 +119,26 @@ import Data.List
 
 --------------------------------------------------------
 
-newtype FCode a = FCode { doFCode :: CgInfoDownwards -> CgState -> (a, CgState) }
-    deriving (Functor)
+newtype FCode a = FCode' { doFCode :: CgInfoDownwards -> CgState -> (a, CgState) }
+
+-- Not derived because of #18202.
+-- See Note [The one-shot state monad trick] in GHC.Utils.Monad
+instance Functor FCode where
+  fmap f (FCode m) =
+    FCode $ \info_down state ->
+      case m info_down state of
+        (x, state') -> (f x, state')
+
+-- This pattern synonym makes the simplifier monad eta-expand,
+-- which as a very beneficial effect on compiler performance
+-- See #18202.
+-- See Note [The one-shot state monad trick] in GHC.Utils.Monad
+{-# COMPLETE FCode #-}
+pattern FCode :: (CgInfoDownwards -> CgState -> (a, CgState))
+              -> FCode a
+pattern FCode m <- FCode' m
+  where
+    FCode m = FCode' $ oneShot (\cgInfoDown -> oneShot (\state ->m cgInfoDown state))
 
 instance Applicative FCode where
     pure val = FCode (\_info_down state -> (val, state))
@@ -177,9 +203,9 @@ data CgIdInfo
         , cg_loc :: CgLoc                     -- CmmExpr for the *tagged* value
         }
 
-instance Outputable CgIdInfo where
-  ppr (CgIdInfo { cg_id = id, cg_loc = loc })
-    = ppr id <+> text "-->" <+> ppr loc
+instance OutputableP Platform CgIdInfo where
+  pdoc env (CgIdInfo { cg_id = id, cg_loc = loc })
+    = ppr id <+> text "-->" <+> pdoc env loc
 
 -- Sequel tells what to do with the result of this expression
 data Sequel
@@ -276,7 +302,7 @@ initCgInfoDown :: DynFlags -> Module -> CgInfoDownwards
 initCgInfoDown dflags mod
   = MkCgInfoDown { cgd_dflags    = dflags
                  , cgd_mod       = mod
-                 , cgd_updfr_off = initUpdFrameOff dflags
+                 , cgd_updfr_off = initUpdFrameOff (targetPlatform dflags)
                  , cgd_ticky     = mkTopTickyCtrLabel
                  , cgd_sequel    = initSequel
                  , cgd_self_loop = Nothing
@@ -285,8 +311,8 @@ initCgInfoDown dflags mod
 initSequel :: Sequel
 initSequel = Return
 
-initUpdFrameOff :: DynFlags -> UpdFrameOffset
-initUpdFrameOff dflags = widthInBytes (wordWidth dflags) -- space for the RA
+initUpdFrameOff :: Platform -> UpdFrameOffset
+initUpdFrameOff platform = platformWordSizeInBytes platform -- space for the RA
 
 
 --------------------------------------------------------
@@ -307,6 +333,9 @@ data CgState
      cgs_hp_usg  :: HeapUsage,
 
      cgs_uniqs :: UniqSupply }
+-- If you are wondering why you have to be careful forcing CgState then
+-- the reason is the knot-tying in 'getHeapUsage'. This problem is tracked
+-- in #19245
 
 data HeapUsage   -- See Note [Virtual and real heap pointers]
   = HeapUsage {
@@ -369,10 +398,9 @@ addCodeBlocksFrom :: CgState -> CgState -> CgState
 -- Add code blocks from the latter to the former
 -- (The cgs_stmts will often be empty, but not always; see codeOnly)
 s1 `addCodeBlocksFrom` s2
-  = s1 { cgs_stmts = cgs_stmts s1 MkGraph.<*> cgs_stmts s2,
+  = s1 { cgs_stmts = cgs_stmts s1 CmmGraph.<*> cgs_stmts s2,
          cgs_tops  = cgs_tops  s1 `appOL` cgs_tops  s2 }
 
-
 -- The heap high water mark is the larger of virtHp and hwHp.  The latter is
 -- only records the high water marks of forked-off branches, so to find the
 -- heap high water mark you have to take the max of virtHp and hwHp.  Remember,
@@ -470,8 +498,31 @@ withSelfLoop self_loop code = do
 instance HasDynFlags FCode where
     getDynFlags = liftM cgd_dflags getInfoDown
 
-getThisPackage :: FCode UnitId
-getThisPackage = liftM thisPackage getDynFlags
+getProfile :: FCode Profile
+getProfile = targetProfile <$> getDynFlags
+
+getPlatform :: FCode Platform
+getPlatform = profilePlatform <$> getProfile
+
+getCallOpts :: FCode CallOpts
+getCallOpts = do
+   dflags <- getDynFlags
+   profile <- getProfile
+   pure $ CallOpts
+    { co_profile       = profile
+    , co_loopification = gopt Opt_Loopification dflags
+    , co_ticky         = gopt Opt_Ticky dflags
+    }
+
+getPtrOpts :: FCode PtrOpts
+getPtrOpts = do
+   dflags <- getDynFlags
+   profile <- getProfile
+   pure $ PtrOpts
+      { po_profile     = profile
+      , po_align_check = gopt Opt_AlignmentSanitisation dflags
+      }
+
 
 withInfoDown :: FCode a -> CgInfoDownwards -> FCode a
 withInfoDown (FCode fcode) info_down = FCode $ \_ state -> fcode info_down state
@@ -562,12 +613,12 @@ forkClosureBody :: FCode () -> FCode ()
 --     re-bind the free variables to a field of the closure.
 
 forkClosureBody body_code
-  = do  { dflags <- getDynFlags
+  = do  { platform <- getPlatform
         ; info   <- getInfoDown
         ; us     <- newUniqSupply
         ; state  <- getState
         ; let body_info_down = info { cgd_sequel    = initSequel
-                                    , cgd_updfr_off = initUpdFrameOff dflags
+                                    , cgd_updfr_off = initUpdFrameOff platform
                                     , cgd_self_loop = Nothing }
               fork_state_in = (initCgState us) { cgs_binds = cgs_binds state }
               ((),fork_state_out) = doFCode body_code body_info_down fork_state_in
@@ -703,7 +754,7 @@ emitTick = emitCgStmt . CgStmt . CmmTick
 emitUnwind :: [(GlobalReg, Maybe CmmExpr)] -> FCode ()
 emitUnwind regs = do
   dflags <- getDynFlags
-  when (debugLevel dflags > 0) $ do
+  when (debugLevel dflags > 0) $
      emitCgStmt $ CgStmt $ CmmUnwind regs
 
 emitAssign :: CmmReg  -> CmmExpr -> FCode ()
@@ -715,7 +766,7 @@ emitStore l r = emitCgStmt (CgStmt (CmmStore l r))
 emit :: CmmAGraph -> FCode ()
 emit ag
   = do  { state <- getState
-        ; setState $ state { cgs_stmts = cgs_stmts state MkGraph.<*> ag } }
+        ; setState $ state { cgs_stmts = cgs_stmts state CmmGraph.<*> ag } }
 
 emitDecl :: CmmDecl -> FCode ()
 emitDecl decl
@@ -736,14 +787,14 @@ emitProcWithStackFrame
    -> FCode ()
 
 emitProcWithStackFrame _conv mb_info lbl _stk_args [] blocks False
-  = do  { dflags <- getDynFlags
-        ; emitProc mb_info lbl [] blocks (widthInBytes (wordWidth dflags)) False
+  = do  { platform <- getPlatform
+        ; emitProc mb_info lbl [] blocks (widthInBytes (wordWidth platform)) False
         }
 emitProcWithStackFrame conv mb_info lbl stk_args args (graph, tscope) True
         -- do layout
-  = do  { dflags <- getDynFlags
-        ; let (offset, live, entry) = mkCallEntry dflags conv args stk_args
-              graph' = entry MkGraph.<*> graph
+  = do  { profile <- getProfile
+        ; let (offset, live, entry) = mkCallEntry profile conv args stk_args
+              graph' = entry CmmGraph.<*> graph
         ; emitProc mb_info lbl live (graph', tscope) offset True
         }
 emitProcWithStackFrame _ _ _ _ _ _ _ = panic "emitProcWithStackFrame"
@@ -758,8 +809,7 @@ emitProcWithConvention conv mb_info lbl args blocks
 emitProc :: Maybe CmmInfoTable -> CLabel -> [GlobalReg] -> CmmAGraphScoped
          -> Int -> Bool -> FCode ()
 emitProc mb_info lbl live blocks offset do_layout
-  = do  { dflags <- getDynFlags
-        ; l <- newBlockId
+  = do  { l <- newBlockId
         ; let
               blks :: CmmGraph
               blks = labelAGraph l blocks
@@ -768,7 +818,6 @@ emitProc mb_info lbl live blocks offset do_layout
                     | otherwise            = mapEmpty
 
               sinfo = StackInfo { arg_space = offset
-                                , updfr_space = Just (initUpdFrameOff dflags)
                                 , do_layout = do_layout }
 
               tinfo = TopInfo { info_tbls = infos
@@ -779,15 +828,15 @@ emitProc mb_info lbl live blocks offset do_layout
         ; state <- getState
         ; setState $ state { cgs_tops = cgs_tops state `snocOL` proc_block } }
 
-getCmm :: FCode () -> FCode CmmGroup
+getCmm :: FCode a -> FCode (a, CmmGroup)
 -- Get all the CmmTops (there should be no stmts)
 -- Return a single Cmm which may be split from other Cmms by
 -- object splitting (at a later stage)
 getCmm code
   = do  { state1 <- getState
-        ; ((), state2) <- withState code (state1 { cgs_tops  = nilOL })
+        ; (a, state2) <- withState code (state1 { cgs_tops  = nilOL })
         ; setState $ state2 { cgs_tops = cgs_tops state1 }
-        ; return (fromOL (cgs_tops state2)) }
+        ; return (a, fromOL (cgs_tops state2)) }
 
 
 mkCmmIfThenElse :: CmmExpr -> CmmAGraph -> CmmAGraph -> FCode CmmAGraph
@@ -838,12 +887,12 @@ mkCmmIfThen' e tbranch l = do
 mkCall :: CmmExpr -> (Convention, Convention) -> [CmmFormal] -> [CmmExpr]
        -> UpdFrameOffset -> [CmmExpr] -> FCode CmmAGraph
 mkCall f (callConv, retConv) results actuals updfr_off extra_stack = do
-  dflags <- getDynFlags
-  k      <- newBlockId
-  tscp   <- getTickScope
+  profile <- getProfile
+  k       <- newBlockId
+  tscp    <- getTickScope
   let area = Young k
-      (off, _, copyin) = copyInOflow dflags retConv area results []
-      copyout = mkCallReturnsTo dflags f callConv actuals k off updfr_off extra_stack
+      (off, _, copyin) = copyInOflow profile retConv area results []
+      copyout = mkCallReturnsTo profile f callConv actuals k off updfr_off extra_stack
   return $ catAGraphs [copyout, mkLabel k tscp, copyin]
 
 mkCmmCall :: CmmExpr -> [CmmFormal] -> [CmmExpr] -> UpdFrameOffset
