@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE LambdaCase #-}
 
 -----------------------------------------------------------------------------
 --
@@ -10,8 +11,8 @@
 
 module GHC.StgToCmm.Utils (
         cgLit, mkSimpleLit,
-        emitDataLits, mkDataLits,
-        emitRODataLits, mkRODataLits,
+        emitDataLits, emitRODataLits,
+        emitDataCon,
         emitRtsCall, emitRtsCallWithResult, emitRtsCallGen,
         assignTemp, newTemp,
 
@@ -37,54 +38,65 @@ module GHC.StgToCmm.Utils (
         cmmUntag, cmmIsTagged,
 
         addToMem, addToMemE, addToMemLblE, addToMemLbl,
-        mkWordCLit,
         newStringCLit, newByteStringCLit,
-        blankWord,
 
         -- * Update remembered set operations
         whenUpdRemSetEnabled,
         emitUpdRemSetPush,
         emitUpdRemSetPushThunk,
+
+        convertInfoProvMap, cmmInfoTableToInfoProvEnt
   ) where
 
 #include "HsVersions.h"
 
-import GhcPrelude
+import GHC.Prelude
 
+import GHC.Platform
 import GHC.StgToCmm.Monad
 import GHC.StgToCmm.Closure
-import Cmm
-import BlockId
-import MkGraph
+import GHC.Cmm
+import GHC.Cmm.BlockId
+import GHC.Cmm.Graph as CmmGraph
 import GHC.Platform.Regs
-import CLabel
-import CmmUtils
-import CmmSwitch
+import GHC.Cmm.CLabel
+import GHC.Cmm.Utils
+import GHC.Cmm.Switch
 import GHC.StgToCmm.CgUtils
 
-import ForeignCall
-import IdInfo
-import Type
-import TyCon
-import SMRep
-import Module
-import Literal
-import Digraph
-import Util
-import Unique
-import UniqSupply (MonadUnique(..))
-import DynFlags
-import FastString
-import Outputable
-import RepType
+import GHC.Types.ForeignCall
+import GHC.Types.Id.Info
+import GHC.Core.Type
+import GHC.Core.TyCon
+import GHC.Runtime.Heap.Layout
+import GHC.Unit
+import GHC.Types.Literal
+import GHC.Data.Graph.Directed
+import GHC.Utils.Misc
+import GHC.Types.Unique
+import GHC.Types.Unique.Supply (MonadUnique(..))
+import GHC.Driver.Session
+import GHC.Data.FastString
+import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Types.RepType
+import GHC.Types.CostCentre
+import GHC.Types.IPE
 
 import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as BS8
 import qualified Data.Map as M
 import Data.Char
-import Data.List
+import Data.List (sortBy)
 import Data.Ord
-
+import GHC.Types.Unique.Map
+import Data.Maybe
+import GHC.Driver.Ppr
+import qualified Data.List.NonEmpty as NE
+import GHC.Core.DataCon
+import GHC.Types.Unique.FM
+import GHC.Data.Maybe
+import Control.Monad
 
 -------------------------------------------------------------------------
 --
@@ -95,25 +107,32 @@ import Data.Ord
 cgLit :: Literal -> FCode CmmLit
 cgLit (LitString s) = newByteStringCLit s
  -- not unpackFS; we want the UTF-8 byte stream.
-cgLit other_lit     = do dflags <- getDynFlags
-                         return (mkSimpleLit dflags other_lit)
+cgLit other_lit     = do platform <- getPlatform
+                         return (mkSimpleLit platform other_lit)
 
-mkSimpleLit :: DynFlags -> Literal -> CmmLit
-mkSimpleLit dflags (LitChar   c)                = CmmInt (fromIntegral (ord c))
-                                                         (wordWidth dflags)
-mkSimpleLit dflags LitNullAddr                  = zeroCLit dflags
-mkSimpleLit dflags (LitNumber LitNumInt i _)    = CmmInt i (wordWidth dflags)
-mkSimpleLit _      (LitNumber LitNumInt64 i _)  = CmmInt i W64
-mkSimpleLit dflags (LitNumber LitNumWord i _)   = CmmInt i (wordWidth dflags)
-mkSimpleLit _      (LitNumber LitNumWord64 i _) = CmmInt i W64
-mkSimpleLit _      (LitFloat r)                 = CmmFloat r W32
-mkSimpleLit _      (LitDouble r)                = CmmFloat r W64
-mkSimpleLit _      (LitLabel fs ms fod)
-  = let -- TODO: Literal labels might not actually be in the current package...
-        labelSrc = ForeignLabelInThisPackage
-    in CmmLabel (mkForeignLabel fs ms labelSrc fod)
--- NB: LitRubbish should have been lowered in "CoreToStg"
-mkSimpleLit _      other = pprPanic "mkSimpleLit" (ppr other)
+mkSimpleLit :: Platform -> Literal -> CmmLit
+mkSimpleLit platform = \case
+   (LitChar   c)                -> CmmInt (fromIntegral (ord c))
+                                          (wordWidth platform)
+   LitNullAddr                  -> zeroCLit platform
+   (LitNumber LitNumInt i)      -> CmmInt i (wordWidth platform)
+   (LitNumber LitNumInt8 i)     -> CmmInt i W8
+   (LitNumber LitNumInt16 i)    -> CmmInt i W16
+   (LitNumber LitNumInt32 i)    -> CmmInt i W32
+   (LitNumber LitNumInt64 i)    -> CmmInt i W64
+   (LitNumber LitNumWord i)     -> CmmInt i (wordWidth platform)
+   (LitNumber LitNumWord8 i)    -> CmmInt i W8
+   (LitNumber LitNumWord16 i)   -> CmmInt i W16
+   (LitNumber LitNumWord32 i)   -> CmmInt i W32
+   (LitNumber LitNumWord64 i)   -> CmmInt i W64
+   (LitFloat r)                 -> CmmFloat r W32
+   (LitDouble r)                -> CmmFloat r W64
+   (LitLabel fs ms fod)
+     -> let -- TODO: Literal labels might not actually be in the current package...
+            labelSrc = ForeignLabelInThisPackage
+        in CmmLabel (mkForeignLabel fs ms labelSrc fod)
+   -- NB: LitRubbish should have been lowered in "CoreToStg"
+   other -> pprPanic "mkSimpleLit" (ppr other)
 
 --------------------------------------------------------------------------
 --
@@ -149,13 +168,13 @@ addToMemE rep ptr n
 -------------------------------------------------------------------------
 
 mkTaggedObjectLoad
-  :: DynFlags -> LocalReg -> LocalReg -> ByteOff -> DynTag -> CmmAGraph
+  :: Platform -> LocalReg -> LocalReg -> ByteOff -> DynTag -> CmmAGraph
 -- (loadTaggedObjectField reg base off tag) generates assignment
 --      reg = bitsK[ base + off - tag ]
 -- where K is fixed by 'reg'
-mkTaggedObjectLoad dflags reg base offset tag
+mkTaggedObjectLoad platform reg base offset tag
   = mkAssign (CmmLocal reg)
-             (CmmLoad (cmmOffsetB dflags
+             (CmmLoad (cmmOffsetB platform
                                   (CmmReg (CmmLocal base))
                                   (offset - tag))
                       (localRegType reg))
@@ -167,9 +186,9 @@ mkTaggedObjectLoad dflags reg base offset tag
 --
 -------------------------------------------------------------------------
 
-tagToClosure :: DynFlags -> TyCon -> CmmExpr -> CmmExpr
-tagToClosure dflags tycon tag
-  = CmmLoad (cmmOffsetExprW dflags closure_tbl tag) (bWord dflags)
+tagToClosure :: Platform -> TyCon -> CmmExpr -> CmmExpr
+tagToClosure platform tycon tag
+  = CmmLoad (cmmOffsetExprW platform closure_tbl tag) (bWord platform)
   where closure_tbl = CmmLit (CmmLabel lbl)
         lbl = mkClosureTableLabel (tyConName tycon) NoCafRefs
 
@@ -179,25 +198,25 @@ tagToClosure dflags tycon tag
 --
 -------------------------------------------------------------------------
 
-emitRtsCall :: UnitId -> FastString -> [(CmmExpr, PrimRep, ForeignHint)] -> Bool -> FCode ()
+emitRtsCall :: UnitId -> FastString -> [(CmmExpr,ForeignHint)] -> Bool -> FCode ()
 emitRtsCall pkg fun args safe = emitRtsCallGen [] (mkCmmCodeLabel pkg fun) args safe
 
-emitRtsCallWithResult :: LocalReg -> PrimRep -> ForeignHint -> UnitId -> FastString
-        -> [(CmmExpr, PrimRep, ForeignHint)] -> Bool -> FCode ()
-emitRtsCallWithResult res rep hint pkg fun args safe
-   = emitRtsCallGen [(res, rep, hint)] (mkCmmCodeLabel pkg fun) args safe
+emitRtsCallWithResult :: LocalReg -> ForeignHint -> UnitId -> FastString
+        -> [(CmmExpr,ForeignHint)] -> Bool -> FCode ()
+emitRtsCallWithResult res hint pkg fun args safe
+   = emitRtsCallGen [(res,hint)] (mkCmmCodeLabel pkg fun) args safe
 
 -- Make a call to an RTS C procedure
 emitRtsCallGen
-   :: [(LocalReg, PrimRep, ForeignHint)]
+   :: [(LocalReg,ForeignHint)]
    -> CLabel
-   -> [(CmmExpr, PrimRep, ForeignHint)]
+   -> [(CmmExpr,ForeignHint)]
    -> Bool -- True <=> CmmSafe call
    -> FCode ()
 emitRtsCallGen res lbl args safe
-  = do { dflags <- getDynFlags
+  = do { platform <- targetPlatform <$> getDynFlags
        ; updfr_off <- getUpdFrameOff
-       ; let (caller_save, caller_load) = callerSaveVolatileRegs dflags
+       ; let (caller_save, caller_load) = callerSaveVolatileRegs platform
        ; emit caller_save
        ; call updfr_off
        ; emit caller_load }
@@ -206,14 +225,10 @@ emitRtsCallGen res lbl args safe
       if safe then
         emit =<< mkCmmCall fun_expr res' args' updfr_off
       else do
-        let conv = ForeignConvention CCallConv arg_hints res_hints CmmMayReturn res_rep arg_reps
+        let conv = ForeignConvention CCallConv arg_hints res_hints CmmMayReturn
         emit $ mkUnsafeCall (ForeignTarget fun_expr conv) res' args'
-    (args', arg_reps, arg_hints) = unzip3 args
-    (res',  res_reps, res_hints) = unzip3 res
-    res_rep = case res_reps of
-      []  -> VoidRep
-      [r] -> r
-      _   -> error "can not deal with multiple return values"
+    (args', arg_hints) = unzip args
+    (res',  res_hints) = unzip res
     fun_expr = mkLblExpr lbl
 
 
@@ -244,16 +259,14 @@ emitRtsCallGen res lbl args safe
 -- here, as we don't have liveness information.  And really, we
 -- shouldn't be doing the workaround at this point in the pipeline, see
 -- Note [Register parameter passing] and the ToDo on CmmCall in
--- cmm/CmmNode.hs.  Right now the workaround is to avoid inlining across
--- unsafe foreign calls in rewriteAssignments, but this is strictly
+-- "GHC.Cmm.Node".  Right now the workaround is to avoid inlining across
+-- unsafe foreign calls in GHC.Cmm.Sink, but this is strictly
 -- temporary.
-callerSaveVolatileRegs :: DynFlags -> (CmmAGraph, CmmAGraph)
-callerSaveVolatileRegs dflags = (caller_save, caller_load)
+callerSaveVolatileRegs :: Platform -> (CmmAGraph, CmmAGraph)
+callerSaveVolatileRegs platform = (caller_save, caller_load)
   where
-    platform = targetPlatform dflags
-
-    caller_save = catAGraphs (map (callerSaveGlobalReg    dflags) regs_to_save)
-    caller_load = catAGraphs (map (callerRestoreGlobalReg dflags) regs_to_save)
+    caller_save = catAGraphs (map (callerSaveGlobalReg    platform) regs_to_save)
+    caller_load = catAGraphs (map (callerRestoreGlobalReg platform) regs_to_save)
 
     system_regs = [ Sp,SpLim,Hp,HpLim,CCCS,CurrentTSO,CurrentNursery
                     {- ,SparkHd,SparkTl,SparkBase,SparkLim -}
@@ -261,14 +274,14 @@ callerSaveVolatileRegs dflags = (caller_save, caller_load)
 
     regs_to_save = filter (callerSaves platform) system_regs
 
-callerSaveGlobalReg :: DynFlags -> GlobalReg -> CmmAGraph
-callerSaveGlobalReg dflags reg
-    = mkStore (get_GlobalReg_addr dflags reg) (CmmReg (CmmGlobal reg))
+callerSaveGlobalReg :: Platform -> GlobalReg -> CmmAGraph
+callerSaveGlobalReg platform reg
+    = mkStore (get_GlobalReg_addr platform reg) (CmmReg (CmmGlobal reg))
 
-callerRestoreGlobalReg :: DynFlags -> GlobalReg -> CmmAGraph
-callerRestoreGlobalReg dflags reg
+callerRestoreGlobalReg :: Platform -> GlobalReg -> CmmAGraph
+callerRestoreGlobalReg platform reg
     = mkAssign (CmmGlobal reg)
-               (CmmLoad (get_GlobalReg_addr dflags reg) (globalRegType dflags reg))
+               (CmmLoad (get_GlobalReg_addr platform reg) (globalRegType platform reg))
 
 
 -------------------------------------------------------------------------
@@ -277,14 +290,18 @@ callerRestoreGlobalReg dflags reg
 --
 -------------------------------------------------------------------------
 
+-- | Emit a data-segment data block
 emitDataLits :: CLabel -> [CmmLit] -> FCode ()
--- Emit a data-segment data block
 emitDataLits lbl lits = emitDecl (mkDataLits (Section Data lbl) lbl lits)
 
+-- | Emit a read-only data block
 emitRODataLits :: CLabel -> [CmmLit] -> FCode ()
--- Emit a read-only data block
 emitRODataLits lbl lits = emitDecl (mkRODataLits lbl lits)
 
+emitDataCon :: CLabel -> CmmInfoTable -> CostCentreStack -> [CmmLit] -> FCode ()
+emitDataCon lbl itbl ccs payload =
+  emitDecl (CmmData (Section Data lbl) (CmmStatics lbl itbl ccs payload))
+
 newStringCLit :: String -> FCode CmmLit
 -- Make a global definition for the string,
 -- and return its label
@@ -312,9 +329,9 @@ assignTemp :: CmmExpr -> FCode LocalReg
 -- due to them being trashed on foreign calls--though it means
 -- the optimization pass doesn't have to do as much work)
 assignTemp (CmmReg (CmmLocal reg)) = return reg
-assignTemp e = do { dflags <- getDynFlags
+assignTemp e = do { platform <- getPlatform
                   ; uniq <- newUnique
-                  ; let reg = LocalReg uniq (cmmExprType dflags e)
+                  ; let reg = LocalReg uniq (cmmExprType platform e)
                   ; emitAssign (CmmLocal reg) e
                   ; return reg }
 
@@ -329,15 +346,15 @@ newUnboxedTupleRegs :: Type -> FCode ([LocalReg], [ForeignHint])
 -- regs it wants will save later assignments.
 newUnboxedTupleRegs res_ty
   = ASSERT( isUnboxedTupleType res_ty )
-    do  { dflags <- getDynFlags
+    do  { platform <- getPlatform
         ; sequel <- getSequel
-        ; regs <- choose_regs dflags sequel
+        ; regs <- choose_regs platform sequel
         ; ASSERT( regs `equalLength` reps )
           return (regs, map primRepForeignHint reps) }
   where
     reps = typePrimRep res_ty
     choose_regs _ (AssignTo regs _) = return regs
-    choose_regs dflags _            = mapM (newTemp . primRepCmmType dflags) reps
+    choose_regs platform _          = mapM (newTemp . primRepCmmType platform) reps
 
 
 
@@ -363,12 +380,12 @@ type Stmt = (LocalReg, CmmExpr) -- r := e
 emitMultiAssign []    []    = return ()
 emitMultiAssign [reg] [rhs] = emitAssign (CmmLocal reg) rhs
 emitMultiAssign regs rhss   = do
-  dflags <- getDynFlags
-  ASSERT2( equalLength regs rhss, ppr regs $$ ppr rhss )
-    unscramble dflags ([1..] `zip` (regs `zip` rhss))
+  platform <- getPlatform
+  ASSERT2( equalLength regs rhss, ppr regs $$ pdoc platform rhss )
+    unscramble platform ([1..] `zip` (regs `zip` rhss))
 
-unscramble :: DynFlags -> [Vrtx] -> FCode ()
-unscramble dflags vertices = mapM_ do_component components
+unscramble :: Platform -> [Vrtx] -> FCode ()
+unscramble platform vertices = mapM_ do_component components
   where
         edges :: [ Node Key Vrtx ]
         edges = [ DigraphNode vertex key1 (edges_from stmt1)
@@ -391,25 +408,24 @@ unscramble dflags vertices = mapM_ do_component components
                 -- Cyclic?  Then go via temporaries.  Pick one to
                 -- break the loop and try again with the rest.
         do_component (CyclicSCC ((_,first_stmt) : rest)) = do
-            dflags <- getDynFlags
             u <- newUnique
-            let (to_tmp, from_tmp) = split dflags u first_stmt
+            let (to_tmp, from_tmp) = split u first_stmt
             mk_graph to_tmp
-            unscramble dflags rest
+            unscramble platform rest
             mk_graph from_tmp
 
-        split :: DynFlags -> Unique -> Stmt -> (Stmt, Stmt)
-        split dflags uniq (reg, rhs)
+        split :: Unique -> Stmt -> (Stmt, Stmt)
+        split uniq (reg, rhs)
           = ((tmp, rhs), (reg, CmmReg (CmmLocal tmp)))
           where
-            rep = cmmExprType dflags rhs
+            rep = cmmExprType platform rhs
             tmp = LocalReg uniq rep
 
         mk_graph :: Stmt -> FCode ()
         mk_graph (reg, rhs) = emitAssign (CmmLocal reg) rhs
 
         mustFollow :: Stmt -> Stmt -> Bool
-        (reg, _) `mustFollow` (_, rhs) = regUsedIn dflags (CmmLocal reg) rhs
+        (reg, _) `mustFollow` (_, rhs) = regUsedIn platform (CmmLocal reg) rhs
 
 -------------------------------------------------------------------------
 --      mkSwitch
@@ -465,8 +481,8 @@ mk_discrete_switch _ _tag_expr [(_tag,lbl)] Nothing _
         -- In that situation we can be sure the (:) case
         -- can't happen, so no need to test
 
--- SOMETHING MORE COMPLICATED: defer to CmmImplementSwitchPlans
--- See Note [Cmm Switches, the general plan] in CmmSwitch
+-- SOMETHING MORE COMPLICATED: defer to GHC.Cmm.Switch.Implement
+-- See Note [Cmm Switches, the general plan] in GHC.Cmm.Switch
 mk_discrete_switch signed tag_expr branches mb_deflt range
   = mkSwitch tag_expr $ mkSwitchTargets signed range mb_deflt (M.fromList branches)
 
@@ -492,17 +508,17 @@ emitCmmLitSwitch scrut  branches deflt = do
     deflt_lbl <- label_code join_lbl deflt
     branches_lbls <- label_branches join_lbl branches
 
-    dflags <- getDynFlags
-    let cmm_ty = cmmExprType dflags scrut
+    platform <- getPlatform
+    let cmm_ty = cmmExprType platform scrut
         rep = typeWidth cmm_ty
 
     -- We find the necessary type information in the literals in the branches
     let signed = case head branches of
-                    (LitNumber nt _ _, _) -> litNumIsSigned nt
+                    (LitNumber nt _, _) -> litNumIsSigned nt
                     _ -> False
 
-    let range | signed    = (tARGET_MIN_INT dflags, tARGET_MAX_INT dflags)
-              | otherwise = (0, tARGET_MAX_WORD dflags)
+    let range | signed    = (platformMinInt platform, platformMaxInt platform)
+              | otherwise = (0, platformMaxWord platform)
 
     if isFloatType cmm_ty
     then emit =<< mk_float_switch rep scrut' deflt_lbl noBound branches_lbls
@@ -525,28 +541,28 @@ mk_float_switch :: Width -> CmmExpr -> BlockId
               -> [(Literal,BlockId)]
               -> FCode CmmAGraph
 mk_float_switch rep scrut deflt _bounds [(lit,blk)]
-  = do dflags <- getDynFlags
-       return $ mkCbranch (cond dflags) deflt blk Nothing
+  = do platform <- getPlatform
+       return $ mkCbranch (cond platform) deflt blk Nothing
   where
-    cond dflags = CmmMachOp ne [scrut, CmmLit cmm_lit]
+    cond platform = CmmMachOp ne [scrut, CmmLit cmm_lit]
       where
-        cmm_lit = mkSimpleLit dflags lit
+        cmm_lit = mkSimpleLit platform lit
         ne      = MO_F_Ne rep
 
 mk_float_switch rep scrut deflt_blk_id (lo_bound, hi_bound) branches
-  = do dflags <- getDynFlags
+  = do platform <- getPlatform
        lo_blk <- mk_float_switch rep scrut deflt_blk_id bounds_lo lo_branches
        hi_blk <- mk_float_switch rep scrut deflt_blk_id bounds_hi hi_branches
-       mkCmmIfThenElse (cond dflags) lo_blk hi_blk
+       mkCmmIfThenElse (cond platform) lo_blk hi_blk
   where
     (lo_branches, mid_lit, hi_branches) = divideBranches branches
 
     bounds_lo = (lo_bound, Just mid_lit)
     bounds_hi = (Just mid_lit, hi_bound)
 
-    cond dflags = CmmMachOp lt [scrut, CmmLit cmm_lit]
+    cond platform = CmmMachOp lt [scrut, CmmLit cmm_lit]
       where
-        cmm_lit = mkSimpleLit dflags mid_lit
+        cmm_lit = mkSimpleLit platform mid_lit
         lt      = MO_F_Lt rep
 
 
@@ -575,7 +591,7 @@ label_code :: BlockId -> CmmAGraphScoped -> FCode BlockId
 -- and returns L
 label_code join_lbl (code,tsc) = do
     lbl <- newBlockId
-    emitOutOfLine lbl (code MkGraph.<*> mkBranch join_lbl, tsc)
+    emitOutOfLine lbl (code CmmGraph.<*> mkBranch join_lbl, tsc)
     return lbl
 
 --------------
@@ -583,8 +599,8 @@ assignTemp' :: CmmExpr -> FCode CmmExpr
 assignTemp' e
   | isTrivialCmmExpr e = return e
   | otherwise = do
-       dflags <- getDynFlags
-       lreg <- newTemp (cmmExprType dflags e)
+       platform <- getPlatform
+       lreg <- newTemp (cmmExprType platform e)
        let reg = CmmLocal lreg
        emitAssign reg e
        return (CmmReg reg)
@@ -594,34 +610,71 @@ assignTemp' e
 -- Pushing to the update remembered set
 ---------------------------------------------------------------------------
 
-whenUpdRemSetEnabled :: DynFlags -> FCode a -> FCode ()
-whenUpdRemSetEnabled dflags code = do
+whenUpdRemSetEnabled :: FCode a -> FCode ()
+whenUpdRemSetEnabled code = do
+    platform <- getPlatform
     do_it <- getCode code
+    let
+      enabled = CmmLoad (CmmLit $ CmmLabel mkNonmovingWriteBarrierEnabledLabel) (bWord platform)
+      zero = zeroExpr platform
+      is_enabled = cmmNeWord platform enabled zero
     the_if <- mkCmmIfThenElse' is_enabled do_it mkNop (Just False)
     emit the_if
-  where
-    enabled = CmmLoad (CmmLit $ CmmLabel mkNonmovingWriteBarrierEnabledLabel) (bWord dflags)
-    zero = zeroExpr dflags
-    is_enabled = cmmNeWord dflags enabled zero
 
 -- | Emit code to add an entry to a now-overwritten pointer to the update
 -- remembered set.
 emitUpdRemSetPush :: CmmExpr   -- ^ value of pointer which was overwritten
                   -> FCode ()
-emitUpdRemSetPush ptr = do
+emitUpdRemSetPush ptr =
     emitRtsCall
       rtsUnitId
       (fsLit "updateRemembSetPushClosure_")
-      [(CmmReg (CmmGlobal BaseReg), AddrRep, AddrHint),
-       (ptr, AddrRep, AddrHint)]
+      [(CmmReg (CmmGlobal BaseReg), AddrHint),
+       (ptr, AddrHint)]
       False
 
 emitUpdRemSetPushThunk :: CmmExpr -- ^ the thunk
                        -> FCode ()
-emitUpdRemSetPushThunk ptr = do
+emitUpdRemSetPushThunk ptr =
     emitRtsCall
       rtsUnitId
       (fsLit "updateRemembSetPushThunk_")
-      [(CmmReg (CmmGlobal BaseReg), AddrRep, AddrHint),
-       (ptr, AddrRep, AddrHint)]
+      [(CmmReg (CmmGlobal BaseReg), AddrHint),
+       (ptr, AddrHint)]
       False
+
+-- | A bare bones InfoProvEnt for things which don't have a good source location
+cmmInfoTableToInfoProvEnt :: Module -> CmmInfoTable -> InfoProvEnt
+cmmInfoTableToInfoProvEnt this_mod cmit =
+    let cl = cit_lbl cmit
+        cn  = rtsClosureType (cit_rep cmit)
+    in InfoProvEnt cl cn "" this_mod Nothing
+
+-- | Convert source information collected about identifiers in 'GHC.STG.Debug'
+-- to entries suitable for placing into the info table provenenance table.
+convertInfoProvMap :: DynFlags -> [CmmInfoTable] -> Module -> InfoTableProvMap -> [InfoProvEnt]
+convertInfoProvMap dflags defns this_mod (InfoTableProvMap (UniqMap dcenv) denv) =
+  map (\cmit ->
+    let cl = cit_lbl cmit
+        cn  = rtsClosureType (cit_rep cmit)
+
+        tyString :: Outputable a => a -> String
+        tyString t = showPpr dflags t
+
+        lookupClosureMap :: Maybe InfoProvEnt
+        lookupClosureMap = case hasHaskellName cl >>= lookupUniqMap denv of
+                                Just (ty, mbspan) -> Just (InfoProvEnt cl cn (tyString ty) this_mod mbspan)
+                                Nothing -> Nothing
+
+        lookupDataConMap = do
+            UsageSite _ n <- hasIdLabelInfo cl >>= getConInfoTableLocation
+            -- This is a bit grimy, relies on the DataCon and Name having the same Unique, which they do
+            (dc, ns) <- (hasHaskellName cl >>= lookupUFM_Directly dcenv . getUnique)
+            -- Lookup is linear but lists will be small (< 100)
+            return $ InfoProvEnt cl cn (tyString (dataConTyCon dc)) this_mod (join $ lookup n (NE.toList ns))
+
+        -- This catches things like prim closure types and anything else which doesn't have a
+        -- source location
+        simpleFallback = cmmInfoTableToInfoProvEnt this_mod cmit
+
+    in fromMaybe simpleFallback (lookupDataConMap `firstJust` lookupClosureMap)) defns
