@@ -37,10 +37,10 @@ import qualified Distribution.InstalledPackageInfo             as Installed
 import qualified Distribution.Simple.PackageIndex              as C
 import qualified Distribution.Text                             as C
 import qualified Distribution.Types.LocalBuildInfo             as C
-import qualified Distribution.Types.CondTree                   as C
 import qualified Distribution.Types.MungedPackageId            as C
-#if MIN_VERSION_Cabal(3,2,0)
 import qualified Distribution.Utils.ShortText                  as C
+#if !MIN_VERSION_Cabal(3,4,0)
+import qualified Distribution.Types.CondTree                   as C
 #endif
 import qualified Distribution.Verbosity                        as C
 import Hadrian.Expression
@@ -74,8 +74,8 @@ parsePackageData pkg = do
         deps    = nubOrd sorted \\ [name]
         depPkgs = catMaybes $ map findPackageByName deps
     return $ PackageData name version
-                         (shortTextToString (C.synopsis pd))
-                         (shortTextToString (C.description pd))
+                         (C.fromShortText (C.synopsis pd))
+                         (C.fromShortText (C.description pd))
                          depPkgs gpd
   where
     -- Collect an overapproximation of dependencies by ignoring conditionals
@@ -85,14 +85,6 @@ parsePackageData pkg = do
       where
         f (C.CondBranch _ t mt) = collectDeps (Just t) ++ collectDeps mt
 
-#if MIN_VERSION_Cabal(3,2,0)
-    shortTextToString :: C.ShortText -> String
-    shortTextToString = C.fromShortText
-#else
-    shortTextToString :: String -> String
-    shortTextToString = id
-#endif
-
 -- | Parse the package identifier from a Cabal file.
 parseCabalPkgId :: FilePath -> IO String
 parseCabalPkgId file = C.display . C.package . C.packageDescription <$> C.readGenericPackageDescription C.silent file
@@ -160,12 +152,12 @@ configurePackage context@Context {..} = do
     trackArgsHash (target context (Cabal Flags stage) [] [])
     trackArgsHash (target context (Cabal Setup stage) [] [])
     verbosity   <- getVerbosity
-    let v = if verbosity >= Loud then "-v3" else "-v0"
-        argList' = argList ++ ["--flags=" ++ unwords flagList, v]
     when (verbosity >= Loud) $
-        putProgressInfo $ "| Package " ++ quote (pkgName package) ++ " configuration flags: " ++ unwords argList'
+        putProgressInfo $ "| Package " ++ quote (pkgName package) ++ " configuration flags: " ++ unwords argList
+    let v = if verbosity >= Loud then "-v3" else "-v0"
     traced "cabal-configure" $
-        C.defaultMainWithHooksNoReadArgs hooks gpd argList'
+        C.defaultMainWithHooksNoReadArgs hooks gpd
+            (argList ++ ["--flags=" ++ unwords flagList, v])
 
     dir <- Context.buildPath context
     files <- liftIO $ getDirectoryFilesIO "." [ dir -/- "include" -/- "**"
@@ -291,6 +283,7 @@ resolveContextData context@Context {..} = do
           , extraLibDirs    = C.extraLibDirs    buildInfo
           , asmSrcs         = C.asmSources      buildInfo
           , cSrcs           = C.cSources        buildInfo
+          , cxxSrcs         = C.cxxSources      buildInfo
           , cmmSrcs         = C.cmmSources      buildInfo
           , hcOpts          = C.programDefaultArgs ghcProg
               ++ C.hcOptions C.GHC buildInfo
@@ -299,6 +292,7 @@ resolveContextData context@Context {..} = do
               ++ C.programOverrideArgs ghcProg
           , asmOpts            = C.asmOptions buildInfo
           , ccOpts             = C.ccOptions  buildInfo
+          , cxxOpts            = C.cxxOptions buildInfo
           , cmmOpts            = C.cmmOptions buildInfo
           , cppOpts            = C.cppOptions buildInfo
           , ldOpts             = C.ldOptions  buildInfo
