@@ -4,6 +4,7 @@
 module Main (main) where
 
 import Data.List          (isPrefixOf)
+import Data.Maybe         (fromMaybe)
 import GHC.Generics       (Generic)
 import System.Environment (getArgs)
 import System.Exit        (exitFailure)
@@ -21,16 +22,17 @@ main = do
             -- this shouldn't fail (run-time errors are due bugs in zinza)
             w <- run Z
                 { zJobs =
-                    [ GhcJob "8.10.1" False ""                          False ["8.8.3"] defSteps
-                    , GhcJob "8.8.3"  False "--solver-benchmarks"       False []        defSteps
-                    , GhcJob "8.6.5"  False "--complete-hackage-tests"  False ["8.8.3"] defSteps
-                    , GhcJob "8.4.4"  False ""                          False ["8.8.3"] defSteps
-                    , GhcJob "8.2.2"  False ""                          False ["8.8.3"] defSteps
-                    , GhcJob "8.0.2"  False ""                          False ["8.8.3"] defSteps
-                    , GhcJob "7.10.3" False ""                          False ["8.8.3"] defSteps
-                    , GhcJob "7.8.4"  False "--lib-only"                False ["8.8.3"] libSteps
-                    , GhcJob "7.6.3"  True  "--lib-only"                False ["8.8.3"] libSteps
-                    , GhcJob "8.8.3"  True  "--lib-only"                True  ["8.8.3"] $
+                    [ GhcJob "9.0.1"  False "--lib-only"                False ["8.8.4"] libSteps
+                    , GhcJob "8.10.1" False ""                          False ["8.8.4"] defSteps
+                    , GhcJob "8.8.4"  False "--solver-benchmarks"       False []        defSteps
+                    , GhcJob "8.6.5"  False "--complete-hackage-tests"  False ["8.8.4"] defSteps
+                    , GhcJob "8.4.4"  False ""                          False ["8.8.4"] defSteps
+                    , GhcJob "8.2.2"  False ""                          False ["8.8.4"] defSteps
+                    , GhcJob "8.0.2"  False ""                          False ["8.8.4"] defSteps
+                    , GhcJob "7.10.3" False ""                          False ["8.8.4"] defSteps
+                    , GhcJob "7.8.4"  False "--lib-only"                False ["8.8.4"] libSteps
+                    , GhcJob "7.6.3"  True  "--lib-only"                False ["8.8.4"] libSteps
+                    , GhcJob "8.8.4"  True  "--lib-only"                True  ["8.8.4"] $
                         libSteps ++
                         [ "lib-suite-extras --extra-hc /opt/ghc/7.0.4/bin/ghc-7.0.4"
                         , "lib-suite-extras --extra-hc /opt/ghc/7.2.2/bin/ghc-7.2.2"
@@ -38,18 +40,12 @@ main = do
                         ]
                     ]
                 , zMacosJobs =
-                    [ mkMacGhcJob "8.8.3" "https://downloads.haskell.org/~ghc/8.8.3/ghc-8.8.3-x86_64-apple-darwin.tar.xz"
+                    [ mkMacGhcJob "8.8.4" "https://downloads.haskell.org/~ghc/8.8.4/ghc-8.8.4-x86_64-apple-darwin.tar.xz"
                     , mkMacGhcJob "8.6.5" "https://downloads.haskell.org/~ghc/8.6.5/ghc-8.6.5-x86_64-apple-darwin.tar.xz"
                     ]
                 , zWinJobs =
-                    -- 8.8.1 fails atm,
-                    -- Shutting down GHCi sessions (please be patient)...
-                    -- Unexpected failure on GHCi exit: fd:10: hClose: resource vanished (Broken pipe)
-                    -- cabal-tests: fd:10: hClose: resource vanished (Broken pipe)
-                    -- [ WinGhcJob "8.8.1" ["8.6.5"]
-                    [ WinGhcJob "8.6.5" []
-                    , WinGhcJob "8.8.3"  []
-                    , WinGhcJob "8.10.1" ["8.8.3"]
+                    [ mkWinGhcJob "8.6.5"  Nothing           []
+                    , mkWinGhcJob "8.10.4" (Just "8.10.4")   []
                     ]
                 , zMangleVersion = map mangleChar
                 , zOr            = (||)
@@ -123,21 +119,25 @@ data MacGhcJob = MacGhcJob
     }
   deriving (Generic)
 
-data WinGhcJob = WinGhcJob
-    { wgjVersion :: String
-    , wgjNeeds   :: [String]
-    }
-  deriving (Generic)
-
 mkMacGhcJob :: String -> String -> MacGhcJob
 mkMacGhcJob v u = MacGhcJob
     { mgjVersion = v
     , mgjGhcUrl  = u
     , mgjFlags   = ""
-    , mgjNeeds   = ["8.8.3" | not $ "8.8" `isPrefixOf` v ]
+    , mgjNeeds   = ["8.8.4" | not $ "8.8" `isPrefixOf` v ]
     , mgjSteps   = defSteps
     }
 
+data WinGhcJob = WinGhcJob
+    { wgjVersion      :: String
+    , wgjChocoVersion :: String
+    , wgjNeeds        :: [String]
+    }
+  deriving (Generic)
+
+mkWinGhcJob :: String -> Maybe String -> [String] -> WinGhcJob
+mkWinGhcJob v mv = WinGhcJob v (fromMaybe v mv)
+
 instance Zinza Z where
     toType    = genericToTypeSFP
     toValue   = genericToValueSFP
