@@ -1,5 +1,6 @@
 {-# LANGUAGE Safe #-}
 {-# LANGUAGE GADTs #-}
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -99,6 +100,9 @@ import Numeric
 import Numeric.Natural
 import System.IO
 
+-- $setup
+-- >>> import Prelude
+
 -------------------
 
 -- | Format a variable number of arguments with the C-style formatting string.
@@ -291,7 +295,7 @@ instance (a ~ ()) => PrintfType (IO a) where
 
 -- | @since 4.7.0.0
 instance (a ~ ()) => HPrintfType (IO a) where
-    hspr hdl fmts args = do
+    hspr hdl fmts args =
         hPutStr hdl (uprintf fmts (reverse args))
 
 -- | @since 2.01
