@@ -102,6 +102,7 @@ module Data.ByteString.Lazy (
         all,                    -- :: (Word8 -> Bool) -> ByteString -> Bool
         maximum,                -- :: ByteString -> Word8
         minimum,                -- :: ByteString -> Word8
+        compareLength,          -- :: ByteString -> Int64 -> Ordering
 
         -- * Building ByteStrings
         -- ** Scans
@@ -151,8 +152,6 @@ module Data.ByteString.Lazy (
 
         -- ** Search for arbitrary substrings
 --        isSubstringOf,          -- :: ByteString -> ByteString -> Bool
---        findSubstring,          -- :: ByteString -> ByteString -> Maybe Int
---        findSubstrings,         -- :: ByteString -> ByteString -> [Int]
 
         -- * Searching ByteStrings
 
@@ -167,6 +166,8 @@ module Data.ByteString.Lazy (
 
         -- * Indexing ByteStrings
         index,                  -- :: ByteString -> Int64 -> Word8
+        indexMaybe,             -- :: ByteString -> Int64 -> Maybe Word8
+        (!?),                   -- :: ByteString -> Int64 -> Maybe Word8
         elemIndex,              -- :: Word8 -> ByteString -> Maybe Int64
         elemIndexEnd,           -- :: Word8 -> ByteString -> Maybe Int64
         elemIndices,            -- :: Word8 -> ByteString -> [Int64]
@@ -178,6 +179,7 @@ module Data.ByteString.Lazy (
         -- * Zipping and unzipping ByteStrings
         zip,                    -- :: ByteString -> ByteString -> [(Word8,Word8)]
         zipWith,                -- :: (Word8 -> Word8 -> c) -> ByteString -> ByteString -> [c]
+        packZipWith,            -- :: (Word8 -> Word8 -> Word8) -> ByteString -> ByteString -> ByteString
         unzip,                  -- :: [(Word8,Word8)] -> (ByteString,ByteString)
 
         -- * Ordered ByteStrings
@@ -189,11 +191,11 @@ module Data.ByteString.Lazy (
 --        defrag,                -- :: ByteString -> ByteString
 
         -- * I\/O with 'ByteString's
+        -- $IOChunk
 
         -- ** Standard input and output
         getContents,            -- :: IO ByteString
         putStr,                 -- :: ByteString -> IO ()
-        putStrLn,               -- :: ByteString -> IO ()
         interact,               -- :: (ByteString -> ByteString) -> IO ()
 
         -- ** Files
@@ -237,7 +239,6 @@ import System.IO                (Handle,openBinaryFile,stdin,stdout,withBinaryFi
 import System.IO.Error          (mkIOError, illegalOperationErrorType)
 import System.IO.Unsafe
 
-import Foreign.ForeignPtr       (withForeignPtr)
 import Foreign.Ptr
 import Foreign.Storable
 
@@ -265,58 +266,11 @@ unpack = unpackBytes
 
 -- | /O(c)/ Convert a list of strict 'ByteString' into a lazy 'ByteString'
 fromChunks :: [P.ByteString] -> ByteString
-fromChunks cs = L.foldr chunk Empty cs
+fromChunks = L.foldr chunk Empty
 
 -- | /O(c)/ Convert a lazy 'ByteString' into a list of strict 'ByteString'
 toChunks :: ByteString -> [P.ByteString]
-toChunks cs = foldrChunks (:) [] cs
-
--- |/O(1)/ Convert a strict 'ByteString' into a lazy 'ByteString'.
-fromStrict :: P.ByteString -> ByteString
-fromStrict bs | S.null bs = Empty
-              | otherwise = Chunk bs Empty
-
--- |/O(n)/ Convert a lazy 'ByteString' into a strict 'ByteString'.
---
--- Note that this is an /expensive/ operation that forces the whole lazy
--- ByteString into memory and then copies all the data. If possible, try to
--- avoid converting back and forth between strict and lazy bytestrings.
---
-toStrict :: ByteString -> S.ByteString
-toStrict = \cs -> goLen0 cs cs
-    -- We pass the original [ByteString] (bss0) through as an argument through
-    -- goLen0, goLen1, and goLen since we will need it again in goCopy. Passing
-    -- it as an explicit argument avoids capturing it in these functions'
-    -- closures which would result in unnecessary closure allocation.
-  where
-    -- It's still possible that the result is empty
-    goLen0 _   Empty                   = S.empty
-    goLen0 cs0 (Chunk c cs) | S.null c = goLen0 cs0 cs
-    goLen0 cs0 (Chunk c cs)            = goLen1 cs0 c cs
-
-    -- It's still possible that the result is a single chunk
-    goLen1 _   bs Empty                = bs
-    goLen1 cs0 bs (Chunk c cs)
-      | S.null c                   = goLen1 cs0 bs cs
-      | otherwise                  =
-        goLen cs0 (S.checkedAdd "Lazy.concat" (S.length bs) (S.length c)) cs
-
-    -- General case, just find the total length we'll need
-    goLen cs0 !total (Chunk c cs)      = goLen cs0 total' cs
-      where
-        total' = S.checkedAdd "Lazy.concat" total (S.length c)
-    goLen cs0 total Empty =
-      S.unsafeCreate total $ \ptr -> goCopy cs0 ptr
-
-    -- Copy the data
-    goCopy Empty                        !_   = return ()
-    goCopy (Chunk (S.PS _  _   0  ) cs) !ptr = goCopy cs ptr
-    goCopy (Chunk (S.PS fp off len) cs) !ptr = do
-      withForeignPtr fp $ \p -> do
-        S.memcpy ptr (p `plusPtr` off) len
-        goCopy cs (ptr `plusPtr` len)
--- See the comment on Data.ByteString.Internal.concat for some background on
--- this implementation.
+toChunks = foldrChunks (:) []
 
 ------------------------------------------------------------------------
 
@@ -346,8 +300,8 @@ null _     = False
 
 -- | /O(c)/ 'length' returns the length of a ByteString as an 'Int64'
 length :: ByteString -> Int64
-length cs = foldlChunks (\n c -> n + fromIntegral (S.length c)) 0 cs
-{-# INLINE length #-}
+length = foldlChunks (\n c -> n + fromIntegral (S.length c)) 0
+{-# INLINE [1] length #-}
 
 infixr 5 `cons`, `cons'` --same as list (:)
 infixl 5 `snoc`
@@ -355,7 +309,7 @@ infixl 5 `snoc`
 -- | /O(1)/ 'cons' is analogous to '(Prelude.:)' for lists.
 --
 cons :: Word8 -> ByteString -> ByteString
-cons c cs = Chunk (S.singleton c) cs
+cons c = Chunk (S.singleton c)
 {-# INLINE cons #-}
 
 -- | /O(1)/ Unlike 'cons', 'cons'' is
@@ -441,7 +395,7 @@ append = mappend
 -- | /O(n)/ 'map' @f xs@ is the ByteString obtained by applying @f@ to each
 -- element of @xs@.
 map :: (Word8 -> Word8) -> ByteString -> ByteString
-map f s = go s
+map f = go
     where
         go Empty        = Empty
         go (Chunk x xs) = Chunk y ys
@@ -452,7 +406,7 @@ map f s = go s
 
 -- | /O(n)/ 'reverse' @xs@ returns the elements of @xs@ in reverse order.
 reverse :: ByteString -> ByteString
-reverse cs0 = rev Empty cs0
+reverse = rev Empty
   where rev a Empty        = a
         rev a (Chunk c cs) = rev (Chunk (S.reverse c) a) cs
 {-# INLINE reverse #-}
@@ -465,10 +419,10 @@ intersperse _ Empty        = Empty
 intersperse w (Chunk c cs) = Chunk (S.intersperse w c)
                                    (foldrChunks (Chunk . intersperse') Empty cs)
   where intersperse' :: P.ByteString -> P.ByteString
-        intersperse' (S.PS fp o l) =
-          S.unsafeCreate (2*l) $ \p' -> withForeignPtr fp $ \p -> do
+        intersperse' (S.BS fp l) =
+          S.unsafeCreate (2*l) $ \p' -> S.unsafeWithForeignPtr fp $ \p -> do
             poke p' w
-            S.c_intersperse (p' `plusPtr` 1) (p `plusPtr` o) (fromIntegral l) w
+            S.c_intersperse (p' `plusPtr` 1) p (fromIntegral l) w
 
 -- | The 'transpose' function transposes the rows and columns of its
 -- 'ByteString' argument.
@@ -484,14 +438,14 @@ transpose css = L.map (\ss -> Chunk (S.pack ss) Empty)
 -- the left-identity of the operator), and a ByteString, reduces the
 -- ByteString using the binary operator, from left to right.
 foldl :: (a -> Word8 -> a) -> a -> ByteString -> a
-foldl f z = go z
+foldl f = go
   where go a Empty        = a
         go a (Chunk c cs) = go (S.foldl f a c) cs
 {-# INLINE foldl #-}
 
 -- | 'foldl'' is like 'foldl', but strict in the accumulator.
 foldl' :: (a -> Word8 -> a) -> a -> ByteString -> a
-foldl' f z = go z
+foldl' f = go
   where go !a Empty        = a
         go !a (Chunk c cs) = go (S.foldl' f a c) cs
 {-# INLINE foldl' #-}
@@ -500,7 +454,7 @@ foldl' f z = go z
 -- (typically the right-identity of the operator), and a ByteString,
 -- reduces the ByteString using the binary operator, from right to left.
 foldr :: (Word8 -> a -> a) -> a -> ByteString -> a
-foldr k z = foldrChunks (flip (S.foldr k)) z
+foldr k = foldrChunks (flip (S.foldr k))
 {-# INLINE foldr #-}
 
 -- | 'foldl1' is a variant of 'foldl' that has no starting value
@@ -547,14 +501,14 @@ concatMap f (Chunk c0 cs0) = to c0 cs0
 -- | /O(n)/ Applied to a predicate and a ByteString, 'any' determines if
 -- any element of the 'ByteString' satisfies the predicate.
 any :: (Word8 -> Bool) -> ByteString -> Bool
-any f cs = foldrChunks (\c rest -> S.any f c || rest) False cs
+any f = foldrChunks (\c rest -> S.any f c || rest) False
 {-# INLINE any #-}
 -- todo fuse
 
 -- | /O(n)/ Applied to a predicate and a 'ByteString', 'all' determines
 -- if all elements of the 'ByteString' satisfy the predicate.
 all :: (Word8 -> Bool) -> ByteString -> Bool
-all f cs = foldrChunks (\c rest -> S.all f c && rest) True cs
+all f = foldrChunks (\c rest -> S.all f c && rest) True
 {-# INLINE all #-}
 -- todo fuse
 
@@ -572,12 +526,54 @@ minimum (Chunk c cs) = foldlChunks (\n c' -> n `min` S.minimum c')
                                      (S.minimum c) cs
 {-# INLINE minimum #-}
 
+-- | /O(c)/ 'compareLength' compares the length of a 'ByteString'
+-- to an 'Int64'
+--
+-- @since 0.11.1.0
+compareLength :: ByteString -> Int64 -> Ordering
+compareLength _ toCmp | toCmp < 0 = GT
+compareLength Empty toCmp         = compare 0 toCmp
+compareLength (Chunk c cs) toCmp  = compareLength cs (toCmp - fromIntegral (S.length c))
+{-# INLINE compareLength #-}
+
+{-# RULES
+"ByteString.Lazy length/compareN -> compareLength" [~1] forall t n.
+  compare (length t) n = compareLength t n
+"ByteString.Lazy compareN/length -> compareLength" [~1] forall t n.
+  -- compare EQ LT = GT and vice versa
+  compare n (length t) = compare EQ $ compareLength t n
+"ByteString.Lazy length/==N -> compareLength/==EQ" [~1] forall t n.
+   length t == n = compareLength t n == EQ
+"ByteString.Lazy N==/length -> compareLength/==EQ" [~1] forall t n.
+   n == length t = compareLength t n == EQ
+"ByteString.Lazy length//=N -> compareLength//=EQ" [~1] forall t n.
+   length t /= n = compareLength t n /= EQ
+"ByteString.Lazy N/=/length -> compareLength//=EQ" [~1] forall t n.
+   n /= length t = compareLength t n /= EQ
+"ByteString.Lazy length/<N -> compareLength/==LT" [~1] forall t n.
+   length t < n = compareLength t n == LT
+"ByteString.Lazy >N/length -> compareLength/==LT" [~1] forall t n.
+   n > length t = compareLength t n == LT
+"ByteString.Lazy length/<=N -> compareLength//=GT" [~1] forall t n.
+   length t <= n = compareLength t n /= GT
+"ByteString.Lazy <=N/length -> compareLength//=GT" [~1] forall t n.
+   n >= length t = compareLength t n /= GT
+"ByteString.Lazy length/>N -> compareLength/==GT" [~1] forall t n.
+   length t > n = compareLength t n == GT
+"ByteString.Lazy <N/length -> compareLength/==GT" [~1] forall t n.
+   n < length t = compareLength t n == GT
+"ByteString.Lazy length/>=N -> compareLength//=LT" [~1] forall t n.
+   length t >= n = compareLength t n /= LT
+"ByteString.Lazy >=N/length -> compareLength//=LT" [~1] forall t n.
+   n <= length t = compareLength t n /= LT
+  #-}
+
 -- | The 'mapAccumL' function behaves like a combination of 'map' and
 -- 'foldl'; it applies a function to each element of a ByteString,
 -- passing an accumulating parameter from left to right, and returning a
 -- final value of this accumulator together with the new ByteString.
 mapAccumL :: (acc -> Word8 -> (acc, Word8)) -> acc -> ByteString -> (acc, ByteString)
-mapAccumL f s0 = go s0
+mapAccumL f = go
   where
     go s Empty        = (s, Empty)
     go s (Chunk c cs) = (s'', Chunk c' cs')
@@ -589,7 +585,7 @@ mapAccumL f s0 = go s0
 -- passing an accumulating parameter from right to left, and returning a
 -- final value of this accumulator together with the new ByteString.
 mapAccumR :: (acc -> Word8 -> (acc, Word8)) -> acc -> ByteString -> (acc, ByteString)
-mapAccumR f s0 = go s0
+mapAccumR f = go
   where
     go s Empty        = (s, Empty)
     go s (Chunk c cs) = (s'', Chunk c' cs')
@@ -606,8 +602,18 @@ mapAccumR f s0 = go s0
 --
 -- Note that
 --
--- > last (scanl f z xs) == foldl f z xs.
-scanl :: (Word8 -> Word8 -> Word8) -> Word8 -> ByteString -> ByteString
+-- > head (scanl f z xs) == z
+-- > last (scanl f z xs) == foldl f z xs
+--
+scanl
+    :: (Word8 -> Word8 -> Word8)
+    -- ^ accumulator -> element -> new accumulator
+    -> Word8
+    -- ^ starting value of accumulator
+    -> ByteString
+    -- ^ input of length n
+    -> ByteString
+    -- ^ output of length n+1
 scanl f z = snd . foldl k (z,singleton z)
  where
     k (c,acc) a = let n = f c a in (n, acc `snoc` n)
@@ -660,7 +666,7 @@ cycle cs    = cs' where cs' = foldrChunks Chunk cs' cs
 -- prepending to the ByteString and @b@ is used as the next element in a
 -- recursive call.
 unfoldr :: (a -> Maybe (Word8, a)) -> a -> ByteString
-unfoldr f z = unfoldChunk 32 z
+unfoldr f = unfoldChunk 32
   where unfoldChunk n x =
           case S.unfoldrN n f x of
             (c, Nothing)
@@ -709,33 +715,40 @@ splitAt i cs0 = splitAt' i cs0
                    in (Chunk c cs', cs'')
 
 
--- | 'takeWhile', applied to a predicate @p@ and a ByteString @xs@,
--- returns the longest prefix (possibly empty) of @xs@ of elements that
--- satisfy @p@.
+-- | Similar to 'P.takeWhile',
+-- returns the longest (possibly empty) prefix of elements
+-- satisfying the predicate.
 takeWhile :: (Word8 -> Bool) -> ByteString -> ByteString
-takeWhile f cs0 = takeWhile' cs0
+takeWhile f = takeWhile'
   where takeWhile' Empty        = Empty
         takeWhile' (Chunk c cs) =
-          case findIndexOrEnd (not . f) c of
+          case S.findIndexOrLength (not . f) c of
             0                  -> Empty
             n | n < S.length c -> Chunk (S.take n c) Empty
               | otherwise      -> Chunk c (takeWhile' cs)
 
--- | 'dropWhile' @p xs@ returns the suffix remaining after 'takeWhile' @p xs@.
+-- | Similar to 'P.dropWhile',
+-- drops the longest (possibly empty) prefix of elements
+-- satisfying the predicate and returns the remainder.
 dropWhile :: (Word8 -> Bool) -> ByteString -> ByteString
-dropWhile f cs0 = dropWhile' cs0
+dropWhile f = dropWhile'
   where dropWhile' Empty        = Empty
         dropWhile' (Chunk c cs) =
-          case findIndexOrEnd (not . f) c of
+          case S.findIndexOrLength (not . f) c of
             n | n < S.length c -> Chunk (S.drop n c) cs
               | otherwise      -> dropWhile' cs
 
--- | 'break' @p@ is equivalent to @'span' ('not' . p)@.
+-- | Similar to 'P.break',
+-- returns the longest (possibly empty) prefix of elements which __do not__
+-- satisfy the predicate and the remainder of the string.
+--
+-- 'break' @p@ is equivalent to @'span' (not . p)@ and to @('takeWhile' (not . p) &&& 'dropWhile' (not . p))@.
+--
 break :: (Word8 -> Bool) -> ByteString -> (ByteString, ByteString)
-break f cs0 = break' cs0
+break f = break'
   where break' Empty        = (Empty, Empty)
         break' (Chunk c cs) =
-          case findIndexOrEnd f c of
+          case S.findIndexOrLength f c of
             0                  -> (Empty, Chunk c cs)
             n | n < S.length c -> (Chunk (S.take n c) Empty
                                   ,Chunk (S.drop n c) cs)
@@ -782,8 +795,12 @@ spanByte c (LPS ps) = case (spanByte' ps) of (a,b) -> (LPS a, LPS b)
                       | otherwise  -> (x' : [], x'' : xs)
 -}
 
--- | 'span' @p xs@ breaks the ByteString into two segments. It is
--- equivalent to @('takeWhile' p xs, 'dropWhile' p xs)@
+-- | Similar to 'P.span',
+-- returns the longest (possibly empty) prefix of elements
+-- satisfying the predicate and the remainder of the string.
+--
+-- 'span' @p@ is equivalent to @'break' (not . p)@ and to @('takeWhile' p &&& 'dropWhile' p)@.
+--
 span :: (Word8 -> Bool) -> ByteString -> (ByteString, ByteString)
 span p = break (not . p)
 
@@ -793,16 +810,16 @@ span p = break (not . p)
 -- separators result in an empty component in the output.  eg.
 --
 -- > splitWith (==97) "aabbaca" == ["","","bb","c",""] -- fromEnum 'a' == 97
--- > splitWith (==97) []        == []
+-- > splitWith undefined ""     == []                  -- and not [""]
 --
 splitWith :: (Word8 -> Bool) -> ByteString -> [ByteString]
 splitWith _ Empty          = []
 splitWith p (Chunk c0 cs0) = comb [] (S.splitWith p c0) cs0
 
   where comb :: [P.ByteString] -> [P.ByteString] -> ByteString -> [ByteString]
-        comb acc (s:[]) Empty        = revChunks (s:acc) : []
-        comb acc (s:[]) (Chunk c cs) = comb (s:acc) (S.splitWith p c) cs
-        comb acc (s:ss) cs           = revChunks (s:acc) : comb [] ss cs
+        comb acc [s] Empty        = [revChunks (s:acc)]
+        comb acc [s] (Chunk c cs) = comb (s:acc) (S.splitWith p c) cs
+        comb acc (s:ss) cs        = revChunks (s:acc) : comb [] ss cs
 {-# INLINE splitWith #-}
 
 -- | /O(n)/ Break a 'ByteString' into pieces separated by the byte
@@ -811,6 +828,7 @@ splitWith p (Chunk c0 cs0) = comb [] (S.splitWith p c0) cs0
 -- > split 10  "a\nb\nd\ne" == ["a","b","d","e"]   -- fromEnum '\n' == 10
 -- > split 97  "aXaXaXa"    == ["","X","X","X",""] -- fromEnum 'a' == 97
 -- > split 120 "x"          == ["",""]             -- fromEnum 'x' == 120
+-- > split undefined ""     == []                  -- and not [""]
 --
 -- and
 --
@@ -826,8 +844,8 @@ split _ Empty     = []
 split w (Chunk c0 cs0) = comb [] (S.split w c0) cs0
 
   where comb :: [P.ByteString] -> [P.ByteString] -> ByteString -> [ByteString]
-        comb acc (s:[]) Empty        = revChunks (s:acc) : []
-        comb acc (s:[]) (Chunk c cs) = comb (s:acc) (S.split w c) cs
+        comb acc [s] Empty        = [revChunks (s:acc)]
+        comb acc [s] (Chunk c cs) = comb (s:acc) (S.split w c) cs
         comb acc (s:ss) cs        = revChunks (s:acc) : comb [] ss cs
 {-# INLINE split #-}
 
@@ -848,9 +866,9 @@ group = go
       | S.length c == 1  = to [c] (S.unsafeHead c) cs
       | otherwise        = to [S.unsafeTake 1 c] (S.unsafeHead c) (Chunk (S.unsafeTail c) cs)
 
-    to acc !_ Empty        = revNonEmptyChunks acc : []
+    to acc !_ Empty        = [revNonEmptyChunks acc]
     to acc !w (Chunk c cs) =
-      case findIndexOrEnd (/= w) c of
+      case S.findIndexOrLength (/= w) c of
         0                    -> revNonEmptyChunks acc
                               : go (Chunk c cs)
         n | n == S.length c  -> to (S.unsafeTake n c : acc) w cs
@@ -867,9 +885,9 @@ groupBy k = go
       | S.length c == 1  = to [c] (S.unsafeHead c) cs
       | otherwise        = to [S.unsafeTake 1 c] (S.unsafeHead c) (Chunk (S.unsafeTail c) cs)
 
-    to acc !_ Empty        = revNonEmptyChunks acc : []
+    to acc !_ Empty        = [revNonEmptyChunks acc]
     to acc !w (Chunk c cs) =
-      case findIndexOrEnd (not . k w) c of
+      case S.findIndexOrLength (not . k w) c of
         0                    -> revNonEmptyChunks acc
                               : go (Chunk c cs)
         n | n == S.length c  -> to (S.unsafeTake n c : acc) w cs
@@ -895,12 +913,35 @@ index cs0 i         = index' cs0 i
               index' cs (n - fromIntegral (S.length c))
           | otherwise       = S.unsafeIndex c (fromIntegral n)
 
+-- | /O(c)/ 'ByteString' index, starting from 0, that returns 'Just' if:
+--
+-- > 0 <= n < length bs
+--
+-- @since 0.11.0.0
+indexMaybe :: ByteString -> Int64 -> Maybe Word8
+indexMaybe _ i | i < 0 = Nothing
+indexMaybe cs0 i       = index' cs0 i
+  where index' Empty _ = Nothing
+        index' (Chunk c cs) n
+          | n >= fromIntegral (S.length c) =
+              index' cs (n - fromIntegral (S.length c))
+          | otherwise       = Just $! S.unsafeIndex c (fromIntegral n)
+
+-- | /O(1)/ 'ByteString' index, starting from 0, that returns 'Just' if:
+--
+-- > 0 <= n < length bs
+--
+-- @since 0.11.0.0
+(!?) :: ByteString -> Int64 -> Maybe Word8
+(!?) = indexMaybe
+{-# INLINE (!?) #-}
+
 -- | /O(n)/ The 'elemIndex' function returns the index of the first
 -- element in the given 'ByteString' which is equal to the query
 -- element, or 'Nothing' if there is no such element.
 -- This implementation uses memchr(3).
 elemIndex :: Word8 -> ByteString -> Maybe Int64
-elemIndex w cs0 = elemIndex' 0 cs0
+elemIndex w = elemIndex' 0
   where elemIndex' _ Empty        = Nothing
         elemIndex' n (Chunk c cs) =
           case S.elemIndex w c of
@@ -912,8 +953,9 @@ elemIndex w cs0 = elemIndex' 0 cs0
 -- element, or 'Nothing' if there is no such element. The following
 -- holds:
 --
--- > elemIndexEnd c xs ==
--- > (-) (length xs - 1) `fmap` elemIndex c (reverse xs)
+-- > elemIndexEnd c xs = case elemIndex c (reverse xs) of
+-- >   Nothing -> Nothing
+-- >   Just i  -> Just (length xs - 1 - i)
 --
 -- @since 0.10.6.0
 elemIndexEnd :: Word8 -> ByteString -> Maybe Int64
@@ -924,7 +966,7 @@ elemIndexEnd = findIndexEnd . (==)
 -- the indices of all elements equal to the query element, in ascending order.
 -- This implementation uses memchr(3).
 elemIndices :: Word8 -> ByteString -> [Int64]
-elemIndices w cs0 = elemIndices' 0 cs0
+elemIndices w = elemIndices' 0
   where elemIndices' _ Empty        = []
         elemIndices' n (Chunk c cs) = L.map ((+n).fromIntegral) (S.elemIndices w c)
                              ++ elemIndices' (n + fromIntegral (S.length c)) cs
@@ -935,13 +977,13 @@ elemIndices w cs0 = elemIndices' 0 cs0
 --
 -- But more efficiently than using length on the intermediate list.
 count :: Word8 -> ByteString -> Int64
-count w cs = foldlChunks (\n c -> n + fromIntegral (S.count w c)) 0 cs
+count w = foldlChunks (\n c -> n + fromIntegral (S.count w c)) 0
 
 -- | The 'findIndex' function takes a predicate and a 'ByteString' and
 -- returns the index of the first element in the ByteString
 -- satisfying the predicate.
 findIndex :: (Word8 -> Bool) -> ByteString -> Maybe Int64
-findIndex k cs0 = findIndex' 0 cs0
+findIndex k = findIndex' 0
   where findIndex' _ Empty        = Nothing
         findIndex' n (Chunk c cs) =
           case S.findIndex k c of
@@ -960,7 +1002,7 @@ findIndexEnd k = findIndexEnd' 0
     findIndexEnd' _ Empty = Nothing
     findIndexEnd' n (Chunk c cs) =
       let !n' = n + S.length c
-          !i  = fmap (fromIntegral . (n +)) $ S.findIndexEnd k c
+          !i  = fromIntegral . (n +) <$> S.findIndexEnd k c
       in findIndexEnd' n' cs `mplus` i
 {-# INLINE findIndexEnd #-}
 
@@ -971,7 +1013,7 @@ findIndexEnd k = findIndexEnd' 0
 -- > find f p = case findIndex f p of Just n -> Just (p ! n) ; _ -> Nothing
 --
 find :: (Word8 -> Bool) -> ByteString -> Maybe Word8
-find f cs0 = find' cs0
+find f = find'
   where find' Empty        = Nothing
         find' (Chunk c cs) = case S.find f c of
             Nothing -> find' cs
@@ -981,10 +1023,11 @@ find f cs0 = find' cs0
 -- | The 'findIndices' function extends 'findIndex', by returning the
 -- indices of all elements satisfying the predicate, in ascending order.
 findIndices :: (Word8 -> Bool) -> ByteString -> [Int64]
-findIndices k cs0 = findIndices' 0 cs0
+findIndices k = findIndices' 0
   where findIndices' _ Empty        = []
         findIndices' n (Chunk c cs) = L.map ((+n).fromIntegral) (S.findIndices k c)
                              ++ findIndices' (n + fromIntegral (S.length c)) cs
+{-# INLINE findIndices #-}
 
 -- ---------------------------------------------------------------------
 -- Searching ByteStrings
@@ -995,13 +1038,13 @@ elem w cs = case elemIndex w cs of Nothing -> False ; _ -> True
 
 -- | /O(n)/ 'notElem' is the inverse of 'elem'
 notElem :: Word8 -> ByteString -> Bool
-notElem w cs = not (elem w cs)
+notElem w cs = not (w `elem` cs)
 
 -- | /O(n)/ 'filter', applied to a predicate and a ByteString,
 -- returns a ByteString containing those characters that satisfy the
 -- predicate.
 filter :: (Word8 -> Bool) -> ByteString -> ByteString
-filter p s = go s
+filter p = go
     where
         go Empty        = Empty
         go (Chunk x xs) = chunk (S.filter p x) (go xs)
@@ -1132,6 +1175,20 @@ zipWith f (Chunk a as) (Chunk b bs) = go a as b bs
     to _ (Chunk x' xs) y ys            | not (S.null y) = go x' xs y  ys
     to _ (Chunk x' xs) _ (Chunk y' ys)                  = go x' xs y' ys
 
+-- | A specialised version of `zipWith` for the common case of a
+-- simultaneous map over two ByteStrings, to build a 3rd.
+--
+-- @since 0.11.1.0
+packZipWith :: (Word8 -> Word8 -> Word8) -> ByteString -> ByteString -> ByteString
+packZipWith _ Empty _ = Empty
+packZipWith _ _ Empty = Empty
+packZipWith f (Chunk a@(S.BS _ al) as) (Chunk b@(S.BS _ bl) bs) = Chunk (S.packZipWith f a b) $
+    case compare al bl of
+        LT -> packZipWith f as $ Chunk (S.drop al b) bs
+        EQ -> packZipWith f as bs
+        GT -> packZipWith f (Chunk (S.drop bl a) as) bs
+{-# INLINE packZipWith #-}
+
 -- | /O(n)/ 'unzip' transforms a list of pairs of bytes into a pair of
 -- ByteStrings. Note that this performs two 'pack' operations.
 unzip :: [(Word8,Word8)] -> (ByteString,ByteString)
@@ -1150,7 +1207,7 @@ inits = (Empty :) . inits'
 
 -- | /O(n)/ Return all final segments of the given 'ByteString', longest first.
 tails :: ByteString -> [ByteString]
-tails Empty         = Empty : []
+tails Empty         = [Empty]
 tails cs@(Chunk c cs')
   | S.length c == 1 = cs : tails cs'
   | otherwise       = cs : tails (Chunk (S.unsafeTail c) cs')
@@ -1164,7 +1221,7 @@ tails cs@(Chunk c cs')
 --   if a large string has been read in, and only a small part of it
 --   is needed in the rest of the program.
 copy :: ByteString -> ByteString
-copy cs = foldrChunks (Chunk . S.copy) Empty cs
+copy = foldrChunks (Chunk . S.copy) Empty
 --TODO, we could coalese small blocks here
 --FIXME: probably not strict enough, if we're doing this to avoid retaining
 -- the parent blocks then we'd better copy strictly.
@@ -1188,10 +1245,6 @@ copy cs = foldrChunks (Chunk . S.copy) Empty cs
 --
 -- The handle is closed on EOF.
 --
--- Note: the 'Handle' should be placed in binary mode with
--- 'System.IO.hSetBinaryMode' for 'hGetContentsN' to
--- work correctly.
---
 hGetContentsN :: Int -> Handle -> IO ByteString
 hGetContentsN k h = lazyRead -- TODO close on exceptions
   where
@@ -1201,8 +1254,7 @@ hGetContentsN k h = lazyRead -- TODO close on exceptions
         c <- S.hGetSome h k -- only blocks if there is no data available
         if S.null c
           then hClose h >> return Empty
-          else do cs <- lazyRead
-                  return (Chunk c cs)
+          else Chunk c <$> lazyRead
 
 -- | Read @n@ bytes into a 'ByteString', directly from the
 -- specified 'Handle', in chunks of size @k@.
@@ -1247,11 +1299,8 @@ illegalBufferSize handle fn sz =
 -- | Read entire handle contents /lazily/ into a 'ByteString'. Chunks
 -- are read on demand, using the default chunk size.
 --
--- Once EOF is encountered, the Handle is closed.
---
--- Note: the 'Handle' should be placed in binary mode with
--- 'System.IO.hSetBinaryMode' for 'hGetContents' to
--- work correctly.
+-- File handles are closed on EOF if all the file is read, or through
+-- garbage collection otherwise.
 --
 hGetContents :: Handle -> IO ByteString
 hGetContents = hGetContentsN defaultChunkSize
@@ -1273,7 +1322,11 @@ hGetNonBlocking :: Handle -> Int -> IO ByteString
 hGetNonBlocking = hGetNonBlockingN defaultChunkSize
 
 -- | Read an entire file /lazily/ into a 'ByteString'.
--- The Handle will be held open until EOF is encountered.
+--
+-- The 'Handle' will be held open until EOF is encountered.
+--
+-- Note that this function's implementation relies on 'hGetContents'.
+-- The reader is advised to read its documentation.
 --
 readFile :: FilePath -> IO ByteString
 readFile f = openBinaryFile f ReadMode >>= hGetContents
@@ -1301,7 +1354,7 @@ getContents = hGetContents stdin
 -- writes, and hence 'hPut' alone might not be suitable for concurrent writes.
 --
 hPut :: Handle -> ByteString -> IO ()
-hPut h cs = foldrChunks (\c rest -> S.hPut h c >> rest) (return ()) cs
+hPut h = foldrChunks (\c rest -> S.hPut h c >> rest) (return ())
 
 -- | Similar to 'hPut' except that it will never block. Instead it returns
 -- any tail that did not get written. This tail may be 'empty' in the case that
@@ -1329,15 +1382,6 @@ hPutStr = hPut
 putStr :: ByteString -> IO ()
 putStr = hPut stdout
 
--- | Write a ByteString to stdout, appending a newline byte
---
-putStrLn :: ByteString -> IO ()
-putStrLn ps = hPut stdout ps >> hPut stdout (singleton 0x0a)
-
-{-# DEPRECATED putStrLn
-    "Use Data.ByteString.Lazy.Char8.putStrLn instead. (Functions that rely on ASCII encodings belong in Data.ByteString.Lazy.Char8)"
-  #-}
-
 -- | The interact function takes a function of type @ByteString -> ByteString@
 -- as its argument. The entire input from the standard input device is passed
 -- to this function as its argument, and the resulting string is output on the
@@ -1362,22 +1406,74 @@ moduleError fun msg = error ("Data.ByteString.Lazy." ++ fun ++ ':':' ':msg)
 
 -- reverse a list of non-empty chunks into a lazy ByteString
 revNonEmptyChunks :: [P.ByteString] -> ByteString
-revNonEmptyChunks cs = L.foldl' (flip Chunk) Empty cs
+revNonEmptyChunks = L.foldl' (flip Chunk) Empty
 
 -- reverse a list of possibly-empty chunks into a lazy ByteString
 revChunks :: [P.ByteString] -> ByteString
-revChunks cs = L.foldl' (flip chunk) Empty cs
+revChunks = L.foldl' (flip chunk) Empty
 
--- | 'findIndexOrEnd' is a variant of findIndex, that returns the length
--- of the string if no element is found, rather than Nothing.
-findIndexOrEnd :: (Word8 -> Bool) -> P.ByteString -> Int
-findIndexOrEnd k (S.PS x s l) =
-    S.accursedUnutterablePerformIO $
-      withForeignPtr x $ \f -> go (f `plusPtr` s) 0
-  where
-    go !ptr !n | n >= l    = return l
-               | otherwise = do w <- peek ptr
-                                if k w
-                                  then return n
-                                  else go (ptr `plusPtr` 1) (n+1)
-{-# INLINE findIndexOrEnd #-}
+-- $IOChunk
+--
+-- ⚠ Using lazy I\/O functions like 'readFile' or 'hGetContents'
+-- means that the order of operations such as closing the file handle
+-- is left at the discretion of the RTS.
+-- Hence, the developer can face some issues when:
+--
+-- * The program reads a file and writes the same file. This means that the file
+--   may be locked because the handler has not been released when 'writeFile' is executed.
+-- * The program reads thousands of files, but due to lazy evaluation, the OS's file descriptor
+--   limit is reached before the handlers can be released.
+--
+-- === Why?
+--
+-- Consider the following program:
+--
+-- > import qualified Data.ByteString.Lazy as BL
+-- > main = do
+-- >   _ <- BL.readFile "foo.txt"
+-- >   BL.writeFile "foo.txt" mempty
+--
+-- Generally, in the 'IO' monad side effects happen
+-- sequentially and in full. Therefore, one might reasonably expect that
+-- reading the whole file via 'readFile' executes all three actions
+-- (open the file handle, read its content, close the file handle) before
+-- control moves to the following 'writeFile' action. This expectation holds
+-- for the strict "Data.ByteString" API. However, the above lazy 'ByteString' variant
+-- of the program fails with @openBinaryFile: resource busy (file is locked)@.
+--
+-- The reason for this is that "Data.ByteString.Lazy" is specifically designed
+-- to handle large or unbounded streams of data incrementally, without requiring all the data
+-- to be resident in memory at the same time. Incremental processing would not be possible
+-- if 'readFile' were to follow the usual rules of 'IO': evaluating all side effects
+-- would require reading the file in full and closing its handle before returning from 'readFile'. This is why
+-- 'readFile' (and 'hGetContents' in general) is implemented
+-- via 'unsafeInterleaveIO', which allows 'IO' side effects to be delayed and
+-- interleaved with subsequent processing of the return value.
+-- That's exactly what happens
+-- in the example above: 'readFile' opens a file handle, but since the content
+-- is not fully consumed, the file handle remains open, allowing the content to
+-- read __on demand__ (never in this case, since the return value is ignored).
+-- So when 'writeFile' is executed next, @foo.txt@ is still open for reading and
+-- the RTS takes care to avoid simultaneously opening it for writing, instead
+-- returning the error shown above.
+--
+-- === How to enforce the order of effects?
+--
+-- If the content is small enough to fit in memory,
+-- consider using strict 'Data.ByteString.readFile',
+-- potentially applying 'fromStrict' afterwards. E. g.,
+--
+-- > import qualified Data.ByteString as BS
+-- > import qualified Data.ByteString.Lazy as BL
+-- > main = do
+-- >   _ <- BS.readFile "foo.txt"
+-- >   BL.writeFile "foo.txt" mempty
+--
+-- If you are dealing with large or unbounded data streams,
+-- consider reaching out for a specialised package, such as
+-- <http://hackage.haskell.org/package/conduit conduit>,
+-- <http://hackage.haskell.org/package/machines-bytestring machines-bytestring>,
+-- <http://hackage.haskell.org/package/pipes-bytestring pipes-bytestring>,
+-- <http://hackage.haskell.org/package/streaming-bytestring streaming-bytestring>,
+-- <http://hackage.haskell.org/package/streamly-bytestring streamly-bytestring>,
+-- etc.
