@@ -1,6 +1,3 @@
-{-# LANGUAGE BangPatterns        #-}
-{-# LANGUAGE PackageImports      #-}
-{-# LANGUAGE ScopedTypeVariables #-}
 -- |
 -- Copyright   : (c) 2011 Simon Meier
 -- License     : BSD3-style (see LICENSE)
@@ -9,19 +6,25 @@
 -- Stability   : experimental
 -- Portability : tested on GHC only
 --
--- Benchmark all 'Builder' functions.
+
+{-# LANGUAGE BangPatterns        #-}
+{-# LANGUAGE PackageImports      #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE MagicHash           #-}
+
 module Main (main) where
 
 import           Data.Foldable                         (foldMap)
 import           Data.Monoid
-import           Gauge
+import           Data.String
+import           Test.Tasty.Bench
 import           Prelude                               hiding (words)
 
 import qualified Data.ByteString                       as S
+import qualified Data.ByteString.Char8                 as S8
 import qualified Data.ByteString.Lazy                  as L
 
 import           Data.ByteString.Builder
-import           Data.ByteString.Builder.ASCII
 import           Data.ByteString.Builder.Extra         (byteStringCopy,
                                                         byteStringInsert,
                                                         intHost)
@@ -31,16 +34,13 @@ import           Data.ByteString.Builder.Prim          (BoundedPrim, FixedPrim,
 import qualified Data.ByteString.Builder.Prim          as P
 import qualified Data.ByteString.Builder.Prim.Internal as PI
 
-import qualified Blaze.ByteString.Builder          as Blaze
-import qualified Blaze.Text                        as Blaze
-import qualified "bytestring" Data.ByteString      as OldS
-import qualified "bytestring" Data.ByteString.Lazy as OldL
-
 import           Foreign
 
 import System.Random
-import Paths_bench_bytestring
 
+import BenchBoundsCheckFusion
+import BenchCSV
+import BenchIndices
 
 ------------------------------------------------------------------------------
 -- Benchmark support
@@ -96,21 +96,20 @@ lazyByteStringData = case S.splitAt (nRepl `div` 2) byteStringData of
 byteStringChunksData :: [S.ByteString]
 byteStringChunksData = map (S.pack . replicate (4 ) . fromIntegral) intData
 
-{-# NOINLINE oldByteStringChunksData #-}
-oldByteStringChunksData :: [OldS.ByteString]
-oldByteStringChunksData = map (OldS.pack . replicate (4 ) . fromIntegral) intData
-
+{-# NOINLINE loremIpsum #-}
+loremIpsum :: S.ByteString
+loremIpsum = S8.unlines $ map S8.pack
+  [ "  Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor"
+  , "incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis"
+  , "nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."
+  , "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu"
+  , "fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in"
+  , "culpa qui officia deserunt mollit anim id est laborum."
+  ]
 
 -- benchmark wrappers
 ---------------------
 
-{-# INLINE benchBlaze #-}
-benchBlaze :: String -> a -> (a -> Blaze.Builder) -> Benchmark
-benchBlaze name x b =
-    bench (name ++" (" ++ show nRepl ++ ")") $
-        whnf (OldL.length . Blaze.toLazyByteString . b) x
-
-
 {-# INLINE benchB #-}
 benchB :: String -> a -> (a -> Builder) -> Benchmark
 benchB name x b =
@@ -225,17 +224,38 @@ sanityCheckInfo =
       ]
   ]
 
+sortInputs :: [S.ByteString]
+sortInputs = map (`S.take` S.pack [122, 121 .. 32]) [10..25]
+
+foldInputs :: [S.ByteString]
+foldInputs = map (\k -> S.pack $ if k <= 6 then take (2 ^ k) [32..95] else concat (replicate (2 ^ (k - 6)) [32..95])) [0..16]
+
+zeroes :: L.ByteString
+zeroes = L.replicate 10000 0
+
+zeroOneRepeating :: L.ByteString
+zeroOneRepeating = L.take 10000 (L.cycle (L.pack [0,1]))
+
+
+largeTraversalInput :: S.ByteString
+largeTraversalInput = S.concat (replicate 10 byteStringData)
+
+smallTraversalInput :: S.ByteString
+smallTraversalInput = S8.pack "The quick brown fox"
+
 main :: IO ()
 main = do
   mapM_ putStrLn sanityCheckInfo
-  putStrLn ""
-  wikiPage <- getDataFileName "wiki-haskell.html" >>= S.readFile
-  Gauge.defaultMain
+  defaultMain
     [ bgroup "Data.ByteString.Builder"
       [ bgroup "Small payload"
         [ benchB' "mempty"        ()  (const mempty)
         , benchB' "ensureFree 8"  ()  (const (ensureFree 8))
         , benchB' "intHost 1"     1   intHost
+        , benchB' "UTF-8 String (naive)" "hello world\0" fromString
+        , benchB' "UTF-8 String"  () $ \() -> P.cstringUtf8 "hello world\0"#
+        , benchB' "String (naive)" "hello world!" fromString
+        , benchB' "String"        () $ \() -> P.cstring "hello world!"#
         ]
 
       , bgroup "Encoding wrappers"
@@ -260,10 +280,6 @@ main = do
                 (foldMap byteString)
             , benchB ("foldMap byteStringCopy" ++ dataName) byteStringChunksData
                 (foldMap byteStringCopy)
-            , benchBlaze ("foldMap Blaze.insertByteString" ++ dataName) oldByteStringChunksData
-                (foldMap Blaze.insertByteString)
-            , benchBlaze ("foldMap Blaze.fromByteString" ++ dataName) oldByteStringChunksData
-                (foldMap Blaze.fromByteString)
             ]
 
       , bgroup "Non-bounded encodings"
@@ -275,82 +291,6 @@ main = do
           -- to Integer.
         , benchB "foldMap integerDec (small)"                     smallIntegerData        $ foldMap integerDec
         , benchB "foldMap integerDec (large)"                     largeIntegerData        $ foldMap integerDec
-        , benchBlaze "foldMap integerDec (small) (blaze-textual)" smallIntegerData        $ foldMap Blaze.integral
-        , benchBlaze "foldMap integerDec (large) (blaze-textual)" largeIntegerData        $ foldMap Blaze.integral
-        ]
-      ]
-
-    , bgroup "substrings"
-      [ bgroup "easy"
-        [ bench "easy1"    . nf (uncurry S.findSubstrings)
-                          $ easySubstrings 1 1000000
-        , bench "easy4"    . nf (uncurry S.findSubstrings)
-                          $ easySubstrings 4 1000000
-        , bench "easy16"   . nf (uncurry S.findSubstrings)
-                          $ easySubstrings 16 1000000
-        , bench "easy64"   . nf (uncurry S.findSubstrings)
-                          $ easySubstrings 64 1000000
-        , bench "easy128"  . nf (uncurry S.findSubstrings)
-                          $ easySubstrings 128 1000000
-        , bench "easy1024" . nf (uncurry S.findSubstrings)
-                          $ easySubstrings 1024 1000000
-        ]
-      , bgroup "random"
-        [ bench "random1"    . nf (uncurry S.findSubstrings)
-                          $ randomSubstrings 1 1000000
-        , bench "random4"    . nf (uncurry S.findSubstrings)
-                          $ randomSubstrings 4 1000000
-        , bench "random16"   . nf (uncurry S.findSubstrings)
-                          $ randomSubstrings 16 1000000
-        , bench "random64"   . nf (uncurry S.findSubstrings)
-                          $ randomSubstrings 64 1000000
-        , bench "random128"  . nf (uncurry S.findSubstrings)
-                          $ randomSubstrings 128 1000000
-        , bench "random1024" . nf (uncurry S.findSubstrings)
-                          $ randomSubstrings 1024 1000000
-
-        ]
-      , bgroup "hard"
-        [ bench "hard1"    . nf (uncurry S.findSubstrings)
-                          $ hardSubstrings 1 1000000
-        , bench "hard4"    . nf (uncurry S.findSubstrings)
-                          $ hardSubstrings 4 1000000
-        , bench "hard16"   . nf (uncurry S.findSubstrings)
-                          $ hardSubstrings 16 1000000
-        , bench "hard64"   . nf (uncurry S.findSubstrings)
-                          $ hardSubstrings 64 1000000
-        , bench "hard128"  . nf (uncurry S.findSubstrings)
-                          $ hardSubstrings 128 1000000
-        , bench "hard1024" . nf (uncurry S.findSubstrings)
-                          $ hardSubstrings 1024 1000000
-        ]
-      , bgroup "pathological"
-        [ bench "pathological1"    . nf (uncurry S.findSubstrings)
-                          $ pathologicalSubstrings 1 1000000
-        , bench "pathological4"    . nf (uncurry S.findSubstrings)
-                          $ pathologicalSubstrings 4 1000000
-        , bench "pathological16"   . nf (uncurry S.findSubstrings)
-                          $ pathologicalSubstrings 16 1000000
-        , bench "pathological64"   . nf (uncurry S.findSubstrings)
-                          $ pathologicalSubstrings 64 1000000
-        , bench "pathological128"  . nf (uncurry S.findSubstrings)
-                          $ pathologicalSubstrings 128 1000000
-        , bench "pathological1024" . nf (uncurry S.findSubstrings)
-                          $ pathologicalSubstrings 1024 1000000
-        ]
-      , bgroup "html"
-        [ bench "html1"    . nfIO . fmap (uncurry S.findSubstrings)
-                          $ htmlSubstrings wikiPage 1 1000000
-        , bench "html4"    . nfIO . fmap (uncurry S.findSubstrings)
-                          $ htmlSubstrings wikiPage 4 1000000
-        , bench "html16"   . nfIO . fmap (uncurry S.findSubstrings)
-                          $ htmlSubstrings wikiPage 16 1000000
-        , bench "html64"   . nfIO . fmap (uncurry S.findSubstrings)
-                          $ htmlSubstrings wikiPage 64 1000000
-        , bench "html128"  . nfIO . fmap (uncurry S.findSubstrings)
-                          $ htmlSubstrings wikiPage 128 1000000
-        , bench "html1024" . nfIO . fmap (uncurry S.findSubstrings)
-                          $ htmlSubstrings wikiPage 1024 1000000
         ]
       ]
 
@@ -438,6 +378,10 @@ main = do
       , benchFE "floatHexFixed"    $ fromIntegral >$< P.floatHexFixed
       , benchFE "doubleHexFixed"   $ fromIntegral >$< P.doubleHexFixed
       ]
+    , bgroup "intersperse"
+      [ bench "intersperse" $ whnf (S.intersperse 32) byteStringData
+      , bench "intersperse (unaligned)" $ whnf (S.intersperse 32) (S.drop 1 byteStringData)
+      ]
     , bgroup "partition"
       [
         bgroup "strict"
@@ -461,4 +405,49 @@ main = do
         , bench "balancedSlow"    $ partitionLazy (\x -> hashWord8 x < w 128)
         ]
       ]
+    , bgroup "sort" $ map (\s -> bench (S8.unpack s) $ nf S.sort s) sortInputs
+    , bgroup "words"
+      [ bench "lorem ipsum" $ nf S8.words loremIpsum
+      , bench "one huge word" $ nf S8.words byteStringData
+      ]
+    , bgroup "folds"
+      [ bgroup "foldl'" $ map (\s -> bench (show $ S.length s) $
+          nf (S.foldl' (\acc x -> acc + fromIntegral x) (0 :: Int)) s) foldInputs
+      , bgroup "foldr'" $ map (\s -> bench (show $ S.length s) $
+          nf (S.foldr' (\x acc -> fromIntegral x + acc) (0 :: Int)) s) foldInputs
+      , bgroup "mapAccumL" $ map (\s -> bench (show $ S.length s) $
+          nf (S.mapAccumL (\acc x -> (acc + fromIntegral x, succ x)) (0 :: Int)) s) foldInputs
+      , bgroup "mapAccumR" $ map (\s -> bench (show $ S.length s) $
+          nf (S.mapAccumR (\acc x -> (fromIntegral x + acc, succ x)) (0 :: Int)) s) foldInputs
+      , bgroup "scanl" $ map (\s -> bench (show $ S.length s) $
+          nf (S.scanl (+) 0) s) foldInputs
+      , bgroup "scanr" $ map (\s -> bench (show $ S.length s) $
+          nf (S.scanr (+) 0) s) foldInputs
+      , bgroup "filter" $ map (\s -> bench (show $ S.length s) $
+          nf (S.filter odd) s) foldInputs
+      ]
+    , bgroup "findIndexOrLength"
+      [ bench "takeWhile"      $ nf (L.takeWhile even) zeroes
+      , bench "dropWhile"      $ nf (L.dropWhile even) zeroes
+      , bench "break"          $ nf (L.break odd) zeroes
+      , bench "group zeroes"   $ nf L.group zeroes
+      , bench "group zero-one" $ nf L.group zeroOneRepeating
+      , bench "groupBy (>=)"   $ nf (L.groupBy (>=)) zeroes
+      , bench "groupBy (>)"    $ nf (L.groupBy (>)) zeroes
+      ]
+    , bgroup "findIndex_"
+      [ bench "findIndices"    $ nf (sum . S.findIndices (\x -> x ==  129 || x == 72)) byteStringData
+      , bench "find"           $ nf (S.find (>= 198)) byteStringData
+      ]
+    , bgroup "findIndexEnd"
+      [ bench "findIndexEnd"   $ nf (S.findIndexEnd (<= 57)) byteStringData
+      , bench "elemIndexInd"   $ nf (S.elemIndexEnd 42) byteStringData
+      ]
+    , bgroup "traversals"
+      [ bench "map (+1)"   $ nf (S.map (+ 1)) largeTraversalInput
+      , bench "map (+1)"   $ nf (S.map (+ 1)) smallTraversalInput
+      ]
+    , benchBoundsCheckFusion
+    , benchCSV
+    , benchIndices
     ]
