@@ -23,15 +23,13 @@
 #include "Schedule.h"
 #include "Sparks.h"
 #include "Trace.h"
+#include "eventlog/EventLog.h" // for flushLocalEventsBuf
 #include "sm/GC.h" // for gcWorkerThread()
 #include "STM.h"
 #include "RtsUtils.h"
 #include "sm/OSMem.h"
 #include "sm/BlockAlloc.h" // for countBlocks()
-
-#if !defined(mingw32_HOST_OS)
-#include "rts/IOManager.h" // for setIOManagerControlFd()
-#endif
+#include "IOManager.h"
 
 #include <string.h>
 
@@ -69,10 +67,8 @@ uint32_t numa_map[MAX_NUMA_NODES];
 
 /* Let foreign code get the current Capability -- assuming there is one!
  * This is useful for unsafe foreign calls because they are called with
- * the current Capability held, but they are not passed it. For example,
- * see see the integer-gmp package which calls allocate() in its
- * stgAllocForGMP() function (which gets called by gmp functions).
- * */
+ * the current Capability held, but they are not passed it.
+ */
 Capability * rts_unsafeGetMyCapability (void)
 {
 #if defined(THREADED_RTS)
@@ -100,7 +96,8 @@ findSpark (Capability *cap)
   bool retry;
   uint32_t i = 0;
 
-  if (!emptyRunQueue(cap) || cap->n_returning_tasks != 0) {
+  // This is an approximate check so relaxed load is acceptable here.
+  if (!emptyRunQueue(cap) || RELAXED_LOAD(&cap->n_returning_tasks) != 0) {
       // If there are other threads, don't try to run any new
       // sparks: sparks might be speculative, we don't want to take
       // resources away from the main computation.
@@ -317,6 +314,7 @@ initCapability (Capability *cap, uint32_t i)
     cap->interrupt = 0;
     cap->pinned_object_block = NULL;
     cap->pinned_object_blocks = NULL;
+    cap->pinned_object_empty = NULL;
 
 #if defined(PROFILING)
     cap->r.rCCCS = CCS_SYSTEM;
@@ -888,7 +886,7 @@ void waitForCapability (Capability **pCap, Task *task)
     if (cap == NULL) {
         cap = find_capability_for_task(task);
 
-        // record the Capability as the one this Task is now assocated with.
+        // record the Capability as the one this Task is now associated with.
         task->cap = cap;
     } else {
         ASSERT(task->cap == cap);
@@ -949,7 +947,15 @@ void waitForCapability (Capability **pCap, Task *task)
 /* See Note [GC livelock] in Schedule.c for why we have gcAllowed
    and return the bool */
 bool /* Did we GC? */
-yieldCapability (Capability** pCap, Task *task, bool gcAllowed)
+yieldCapability
+    ( Capability** pCap     // [in/out] Task's owned capability. Set to the
+                            //          newly owned capability on return.
+                            //          Precondition:
+                            //              pCap != NULL
+                            //              && *pCap != NULL
+    , Task *task            // [in] This thread's task.
+    , bool gcAllowed
+    )
 {
     Capability *cap = *pCap;
 
@@ -976,6 +982,10 @@ yieldCapability (Capability** pCap, Task *task, bool gcAllowed)
                 debugTrace(DEBUG_nonmoving_gc, "Flushing update remembered set blocks...");
                 break;
 
+            case SYNC_FLUSH_EVENT_LOG:
+                /* N.B. the actual flushing is performed by flushEventLog */
+                break;
+
             default:
                 break;
             }
@@ -1212,7 +1222,15 @@ shutdownCapability (Capability *cap USED_IF_THREADS,
             //
             // To reproduce this deadlock: run ffi002(threaded1)
             // repeatedly on a loaded machine.
-            ioManagerDie();
+            //
+            // FIXME: stopIOManager is not a per-capability action. It shuts
+            // down the I/O subsystem for all capabilities, but here we call
+            // it once per cap, so this is accidentally quadratic, but mainly
+            // it is confusing. Replace this with a per-capability stop, and
+            // perhaps make it synchronous so it works the first time and we
+            // don't have to come back and try again here.
+            //
+            stopIOManager();
             yieldThread();
             continue;
         }
@@ -1351,16 +1369,3 @@ bool checkSparkCountInvariant (void)
 
 }
 #endif
-
-#if !defined(mingw32_HOST_OS)
-void
-setIOManagerControlFd(uint32_t cap_no USED_IF_THREADS, int fd USED_IF_THREADS) {
-#if defined(THREADED_RTS)
-    if (cap_no < n_capabilities) {
-        RELAXED_STORE(&capabilities[cap_no]->io_manager_control_wr_fd, fd);
-    } else {
-        errorBelch("warning: setIOManagerControlFd called with illegal capability number.");
-    }
-#endif
-}
-#endif
