@@ -27,6 +27,7 @@
 
 #include "BeginPrivate.h"
 
+/* N.B. This must be consistent with CapabilityPublic in RtsAPI.h */
 struct Capability_ {
     // State required by the STG virtual machine when running Haskell
     // code.  During STG execution, the BaseReg register always points
@@ -92,6 +93,8 @@ struct Capability_ {
     bdescr *pinned_object_block;
     // full pinned object blocks allocated since the last GC
     bdescr *pinned_object_blocks;
+    // empty pinned object blocks, to be allocated into
+    bdescr *pinned_object_empty;
 
     // per-capability weak pointer list associated with nursery (older
     // lists stored in generation object)
@@ -267,7 +270,8 @@ typedef enum {
     SYNC_OTHER,
     SYNC_GC_SEQ,
     SYNC_GC_PAR,
-    SYNC_FLUSH_UPD_REM_SET
+    SYNC_FLUSH_UPD_REM_SET,
+    SYNC_FLUSH_EVENT_LOG
 } SyncType;
 
 //
