@@ -13,11 +13,6 @@
 #include "linker/M32Alloc.h"
 
 #if RTS_LINKER_USE_MMAP
-#if defined(ios_HOST_OS) || defined(darwin_HOST_OS)
-/* Inclusion of system headers usually requires _DARWIN_C_SOURCE on Mac OS X
- * because of some specific defines like MMAP_ANON, MMAP_ANONYMOUS. */
-#define _DARWIN_C_SOURCE 1
-#endif
 #include <sys/mman.h>
 #endif
 
@@ -174,6 +169,14 @@ typedef struct _Segment {
 #define NEED_SYMBOL_EXTRAS 1
 #endif
 
+/*
+ * We use the m32 allocator for symbol extras on Windows and other mmap-using
+ * platforms.
+ */
+#if RTS_LINKER_USE_MMAP || defined(mingw32_HOST_ARCH)
+#define NEED_M32 1
+#endif
+
 /* Jump Islands are sniplets of machine code required for relative
  * address relocations on the PowerPC, x86_64 and ARM.
  */
@@ -187,7 +190,8 @@ typedef struct {
     } jumpIsland;
 #elif defined(x86_64_HOST_ARCH)
     uint64_t    addr;
-    uint8_t     jumpIsland[6];
+    // See Note [TLSGD relocation] in elf_tlsgd.c
+    uint8_t     jumpIsland[8];
 #elif defined(arm_HOST_ARCH)
     uint8_t     jumpIsland[16];
 #endif
@@ -305,7 +309,7 @@ struct _ObjectCode {
        require extra information.*/
     StrHashTable *extraInfos;
 
-#if RTS_LINKER_USE_MMAP == 1
+#if defined(NEED_M32)
     /* The m32 allocators used for allocating small sections and symbol extras
      * during loading. We have two: one for (writeable) data and one for
      * (read-only/executable) code. */
@@ -341,6 +345,12 @@ extern Mutex dl_mutex;
 /* Type of the initializer */
 typedef void (*init_t) (int argc, char **argv, char **env);
 
+typedef enum _SymStrength {
+    STRENGTH_NORMAL,
+    STRENGTH_WEAK,
+    STRENGTH_STRONG,
+} SymStrength;
+
 /* SymbolInfo tracks a symbol's address, the object code from which
    it originated, and whether or not it's weak.
 
@@ -356,7 +366,7 @@ typedef void (*init_t) (int argc, char **argv, char **env);
 typedef struct _RtsSymbolInfo {
     SymbolAddr* value;
     ObjectCode *owner;
-    HsBool weak;
+    SymStrength strength;
 } RtsSymbolInfo;
 
 void exitLinker( void );
@@ -385,13 +395,17 @@ int ghciInsertSymbolTable(
     StrHashTable *table,
     const SymbolName* key,
     SymbolAddr* data,
-    HsBool weak,
+    SymStrength weak,
     ObjectCode *owner);
 
 /* Lock-free version of lookupSymbol. When 'dependent' is not NULL, adds it as a
  * dependent to the owner of the symbol. */
 SymbolAddr* lookupDependentSymbol (SymbolName* lbl, ObjectCode *dependent);
 
+/* Perform TLSGD symbol lookup returning the address of the resulting GOT entry,
+ * which in this case holds the module id and the symbol offset. */
+StgInt64 lookupTlsgdSymbol(const char *, unsigned long, ObjectCode *);
+
 extern StrHashTable *symhash;
 
 pathchar*
