@@ -20,7 +20,7 @@
 #include "Profiling.h"
 #include "Messages.h"
 #if defined(mingw32_HOST_OS)
-#include "win32/IOManager.h"
+#include "win32/MIOManager.h"
 #endif
 
 static void blockedThrowTo (Capability *cap,
@@ -174,11 +174,11 @@ throwToSelf (Capability *cap, StgTSO *tso, StgClosure *exception)
 
      - or it is masking exceptions (TSO_BLOCKEX)
 
-   Currently, if the target is BlockedOnMVar, BlockedOnSTM, or
-   BlockedOnBlackHole then we acquire ownership of the TSO by locking
-   its parent container (e.g. the MVar) and then raise the exception.
-   We might change these cases to be more message-passing-like in the
-   future.
+   Currently, if the target is BlockedOnMVar, BlockedOnSTM,
+   BlockedOnIOCompletion or BlockedOnBlackHole then we acquire ownership of the
+   TSO by locking its parent container (e.g. the MVar) and then raise the
+   exception.  We might change these cases to be more message-passing-like in
+   the future.
 
    Returns:
 
@@ -187,7 +187,7 @@ throwToSelf (Capability *cap, StgTSO *tso, StgClosure *exception)
    MessageThrowTo *   exception was not raised; the source TSO
                       should now put itself in the state
                       BlockedOnMsgThrowTo, and when it is ready
-                      it should unlock the mssage using
+                      it should unlock the message using
                       unlockClosure(msg, &stg_MSG_THROWTO_info);
                       If it decides not to raise the exception after
                       all, it can revoke it safely with
@@ -344,6 +344,7 @@ check_target:
 
     case BlockedOnMVar:
     case BlockedOnMVarRead:
+    case BlockedOnIOCompletion:
     {
         /*
           To establish ownership of this TSO, we need to acquire a
@@ -368,7 +369,9 @@ check_target:
 
         // we have the MVar, let's check whether the thread
         // is still blocked on the same MVar.
-        if ((target->why_blocked != BlockedOnMVar && target->why_blocked != BlockedOnMVarRead)
+        if ((target->why_blocked != BlockedOnMVar
+             && target->why_blocked != BlockedOnMVarRead
+             && target->why_blocked != BlockedOnIOCompletion)
             || (StgMVar *)target->block_info.closure != mvar) {
             unlockClosure((StgClosure *)mvar, info);
             goto retry;
@@ -456,7 +459,7 @@ check_target:
         blockedThrowTo(cap,target,msg);
         return THROWTO_BLOCKED;
 
-#if !defined(THREADEDED_RTS)
+#if !defined(THREADED_RTS)
     case BlockedOnRead:
     case BlockedOnWrite:
     case BlockedOnDelay:
@@ -680,6 +683,7 @@ removeFromQueues(Capability *cap, StgTSO *tso)
 
   case BlockedOnMVar:
   case BlockedOnMVarRead:
+  case BlockedOnIOCompletion:
       removeFromMVarBlockedQueue(tso);
       goto done;
 
