@@ -21,7 +21,10 @@
 #include <ctype.h>
 #endif
 
+#include <errno.h>
+
 #include <string.h>
+#include <stdlib.h>
 
 #if defined(HAVE_UNISTD_H)
 #include <unistd.h>
@@ -32,6 +35,7 @@
 #endif
 
 #include <fs_rts.h>
+#include <stdbool.h>
 
 // Flag Structure
 RTS_FLAGS RtsFlags;
@@ -98,6 +102,9 @@ static int  openStatsFile (
 static StgWord64 decodeSize (
     const char *flag, uint32_t offset, StgWord64 min, StgWord64 max);
 
+static double parseDouble (
+    const char *arg, bool *error);
+
 static void bad_option (const char *s);
 
 #if defined(DEBUG)
@@ -146,18 +153,19 @@ void initRtsFlagsDefaults(void)
     RtsFlags.GcFlags.stkChunkSize       = (32 * 1024) / sizeof(W_);
     RtsFlags.GcFlags.stkChunkBufferSize = (1 * 1024) / sizeof(W_);
 
-    RtsFlags.GcFlags.minAllocAreaSize   = (1024 * 1024)       / BLOCK_SIZE;
+    /* -A default. See #16499 for a discussion about the tradeoffs */
+    RtsFlags.GcFlags.minAllocAreaSize   = (4 * 1024 * 1024)       / BLOCK_SIZE;
     RtsFlags.GcFlags.largeAllocLim      = 0; /* defaults to minAllocAreasize */
     RtsFlags.GcFlags.nurseryChunkSize   = 0;
-    RtsFlags.GcFlags.minOldGenSize      = (1024 * 1024)       / BLOCK_SIZE;
+    RtsFlags.GcFlags.minOldGenSize      = (1024 * 1024)       / BLOCK_SIZE; /* -O default */
     RtsFlags.GcFlags.maxHeapSize        = 0;    /* off by default */
     RtsFlags.GcFlags.heapLimitGrace     = (1024 * 1024);
     RtsFlags.GcFlags.heapSizeSuggestion = 0;    /* none */
     RtsFlags.GcFlags.heapSizeSuggestionAuto = false;
     RtsFlags.GcFlags.pcFreeHeap         = 3;    /* 3% */
     RtsFlags.GcFlags.oldGenFactor       = 2;
+    RtsFlags.GcFlags.returnDecayFactor  = 4;
     RtsFlags.GcFlags.useNonmoving       = false;
-    RtsFlags.GcFlags.nonmovingSelectorOpt = false;
     RtsFlags.GcFlags.generations        = 2;
     RtsFlags.GcFlags.squeezeUpdFrames   = true;
     RtsFlags.GcFlags.compact            = false;
@@ -204,9 +212,9 @@ void initRtsFlagsDefaults(void)
 
     RtsFlags.ProfFlags.doHeapProfile      = false;
     RtsFlags.ProfFlags.heapProfileInterval = USToTime(100000); // 100ms
+    RtsFlags.ProfFlags.startHeapProfileAtStartup = true;
 
 #if defined(PROFILING)
-    RtsFlags.ProfFlags.includeTSOs        = false;
     RtsFlags.ProfFlags.showCCSOnException = false;
     RtsFlags.ProfFlags.maxRetainerSetSize = 8;
     RtsFlags.ProfFlags.ccsLength          = 25;
@@ -228,7 +236,9 @@ void initRtsFlagsDefaults(void)
     RtsFlags.TraceFlags.sparks_sampled= false;
     RtsFlags.TraceFlags.sparks_full   = false;
     RtsFlags.TraceFlags.user          = false;
+    RtsFlags.TraceFlags.ticky         = false;
     RtsFlags.TraceFlags.trace_output  = NULL;
+    RtsFlags.TraceFlags.eventlogFlushTime = 0;
 #endif
 
 #if defined(PROFILING)
@@ -248,6 +258,16 @@ void initRtsFlagsDefaults(void)
     RtsFlags.MiscFlags.internalCounters        = false;
     RtsFlags.MiscFlags.linkerAlwaysPic         = DEFAULT_LINKER_ALWAYS_PIC;
     RtsFlags.MiscFlags.linkerMemBase           = 0;
+#if defined(DEFAULT_NATIVE_IO_MANAGER)
+    RtsFlags.MiscFlags.ioManager               = IO_MNGR_NATIVE;
+#else
+    RtsFlags.MiscFlags.ioManager               = IO_MNGR_POSIX;
+#endif
+#if defined(THREADED_RTS) && defined(mingw32_HOST_OS)
+    RtsFlags.MiscFlags.numIoWorkerThreads      = getNumberOfProcessors();
+#else
+    RtsFlags.MiscFlags.numIoWorkerThreads      = 1;
+#endif
 
 #if defined(THREADED_RTS)
     RtsFlags.ParFlags.nCapabilities     = 1;
@@ -298,12 +318,18 @@ usage_text[] = {
 "  -kc<size> Sets the stack chunk size (default 32k)",
 "  -kb<size> Sets the stack chunk buffer size (default 1k)",
 "",
-"  -A<size>  Sets the minimum allocation area size (default 1m) Egs: -A20m -A10k",
+"  -A<size>  Sets the minimum allocation area size (default 4m) Egs: -A20m -A10k",
 "  -AL<size> Sets the amount of large-object memory that can be allocated",
 "            before a GC is triggered (default: the value of -A)",
 "  -F<n>     Sets the collecting threshold for old generations as a factor of",
 "            the live data in that generation the last time it was collected",
 "            (default: 2.0)",
+"  -Fd<n>    Sets the inverse rate which memory is returned to the OS after being",
+"            optimistically retained after being allocated. Subsequent major",
+"            collections not caused by heap overflow will return an amount of",
+"            memory controlled by this factor (higher is slower). Setting the factor",
+"            to 0 means memory is not returned.",
+"            (default 4.0)",
 "  -n<size>  Allocation area chunk size (0 = disabled, default: 0)",
 "  -O<size>  Sets the minimum size of the old generation (default 1M)",
 "  -M<size>  Sets the maximum heap size (default unlimited)  Egs: -M256k -M1G",
@@ -330,7 +356,7 @@ usage_text[] = {
 "  -S[<file>] Detailed GC statistics (if <file> omitted, uses stderr)",
 "",
 "",
-"  -Z         Don't squeeze out update frames on stack overflow",
+"  -Z         Don't squeeze out update frames on context switch",
 "  -B         Sound the bell at the start of each garbage collection",
 #if defined(PROFILING)
 "",
@@ -372,6 +398,11 @@ usage_text[] = {
 "  -hT      Produce a heap profile grouped by closure type",
 #endif /* PROFILING */
 
+"  -i<sec>  Time between heap profile samples (seconds, default: 0.1)",
+"  --no-automatic-heap-samples  Do not start the heap profile interval timer on start-up,",
+"                               Rather, the application will be responsible for triggering",
+"                               heap profiler samples."
+
 #if defined(TRACING)
 "",
 "  -ol<file>  Send binary eventlog to <file> (default: <program>.eventlog)",
@@ -386,15 +417,18 @@ usage_text[] = {
 "                p    par spark events (sampled)",
 "                f    par spark events (full detail)",
 "                u    user events (emitted from Haskell code)",
+#if defined(TICKY_TICKY)
+"                T    ticky-ticky counter samples",
+#endif
 "                a    all event classes above",
 #  if defined(DEBUG)
 "                t    add time stamps (only useful with -v)",
 #  endif
 "               -x    disable an event class, for any flag above",
 "             the initial enabled event classes are 'sgpu'",
+" --eventlog-flush-interval=<secs> Flush the eventlog at the specified interval.",
 #endif
 
-"  -i<sec>  Time between heap profile samples (seconds, default: 0.1)",
 "",
 #if defined(TICKY_TICKY)
 "  -r<file>  Produce ticky-ticky statistics (with -rstderr for stderr)",
@@ -474,7 +508,14 @@ usage_text[] = {
 "            fatal error. When symbols are available an attempt will be",
 "            made to resolve addresses to names. (default: yes)",
 #endif
+"  --io-manager=<native|posix>",
+"            The I/O manager subsystem to use. (default: posix)",
 #if defined(THREADED_RTS)
+#if defined(mingw32_HOST_OS)
+"  --io-manager-threads=<num>",
+"            The number of worker threads to use in the native I/O manager to",
+"            handle completion events. (default: num cores)",
+#endif
 "  -e<n>     Maximum number of outstanding local sparks (default: 4096)",
 #endif
 #if defined(x86_64_HOST_ARCH)
@@ -542,7 +583,7 @@ char** getUTF8Args(int* argc)
 
     // We create two argument arrays, one which is later permutated by the RTS
     // instead of the main argv.
-    // The other one is used to free the allocted memory later.
+    // The other one is used to free the allocated memory later.
     char** argv = (char**) stgMallocBytes(sizeof(char*) * (*argc + 1),
                                           "getUTF8Args 1");
     win32_full_utf8_argv = (char**) stgMallocBytes(sizeof(char*) * (*argc + 1),
@@ -933,12 +974,32 @@ error = true;
                       OPTION_SAFE;
                       RtsFlags.MiscFlags.internalCounters = true;
                   }
+                  else if (strequal("io-manager=native",
+                               &rts_argv[arg][2])) {
+                      OPTION_UNSAFE;
+                      RtsFlags.MiscFlags.ioManager = IO_MNGR_NATIVE;
+                  }
+                  else if (strequal("io-manager=posix",
+                               &rts_argv[arg][2])) {
+                      OPTION_UNSAFE;
+                      RtsFlags.MiscFlags.ioManager = IO_MNGR_POSIX;
+                  }
                   else if (strequal("info",
                                &rts_argv[arg][2])) {
                       OPTION_SAFE;
                       printRtsInfo(rtsConfig);
                       stg_exit(0);
                   }
+                  else if (!strncmp("eventlog-flush-interval=",
+                               &rts_argv[arg][2], 24)) {
+                      OPTION_SAFE;
+                      double intervalSeconds = parseDouble(rts_argv[arg]+26, &error);
+                      if (error) {
+                          errorBelch("bad value for --eventlog-flush-interval");
+                      }
+                      RtsFlags.TraceFlags.eventlogFlushTime =
+                          fsecondsToTime(intervalSeconds);
+                  }
                   else if (strequal("copying-gc",
                                &rts_argv[arg][2])) {
                       OPTION_SAFE;
@@ -950,6 +1011,31 @@ error = true;
                       RtsFlags.GcFlags.useNonmoving = true;
                   }
 #if defined(THREADED_RTS)
+#if defined(mingw32_HOST_OS)
+                  else if (!strncmp("io-manager-threads",
+                                &rts_argv[arg][2], 18)) {
+                      OPTION_SAFE;
+                      uint32_t num;
+                      if (rts_argv[arg][20] == '=') {
+                          num = (StgWord)strtol(rts_argv[arg]+21,
+                                                 (char **) NULL, 10);
+                      } else {
+                          errorBelch("%s: Expected number of threads to use.",
+                                     rts_argv[arg]);
+                          error = true;
+                          break;
+                      }
+
+                      if (num < 1) {
+                          errorBelch("%s: Expected number of threads to be at least 1.",
+                                     rts_argv[arg]);
+                          error = true;
+                          break;
+                      }
+
+                      RtsFlags.MiscFlags.numIoWorkerThreads = num;
+                  }
+#endif
                   else if (!strncmp("numa", &rts_argv[arg][2], 4)) {
                       if (!osBuiltWithNumaSupport()) {
                           errorBelch("%s: This GHC build was compiled without NUMA support.",
@@ -1017,6 +1103,12 @@ error = true;
                       }
                       break;
                   }
+                  else if (strequal("no-automatic-heap-samples",
+                               &rts_argv[arg][2])) {
+                      OPTION_SAFE;
+                      RtsFlags.ProfFlags.startHeapProfileAtStartup = false;
+                      break;
+                  }
                   else {
                       OPTION_SAFE;
                       errorBelch("unknown RTS option: %s",rts_argv[arg]);
@@ -1068,10 +1160,19 @@ error = true;
 
               case 'F':
                 OPTION_UNSAFE;
-                RtsFlags.GcFlags.oldGenFactor = atof(rts_argv[arg]+2);
+                switch(rts_argv[arg][2]) {
+                case 'd':
+                  RtsFlags.GcFlags.returnDecayFactor = atof(rts_argv[arg]+3);
+                  if (RtsFlags.GcFlags.returnDecayFactor < 0)
+                    bad_option( rts_argv[arg] );
+                  break;
+                default:
+                  RtsFlags.GcFlags.oldGenFactor = atof(rts_argv[arg]+2);
 
-                if (RtsFlags.GcFlags.oldGenFactor < 0)
-                  bad_option( rts_argv[arg] );
+                  if (RtsFlags.GcFlags.oldGenFactor < 0)
+                    bad_option( rts_argv[arg] );
+                  break;
+                };
                 break;
 
               case 'D':
@@ -1317,10 +1418,17 @@ error = true;
 #if !defined(PROFILING)
                 switch (rts_argv[arg][2]) {
                   case '\0':
+                    errorBelch("-h is deprecated and will be removed in the next release (9.4),\nuse -hT instead.");
+
+                    FALLTHROUGH;
                   case 'T':
                     OPTION_UNSAFE;
                     RtsFlags.ProfFlags.doHeapProfile = HEAP_BY_CLOSURE_TYPE;
                     break;
+                  case 'i':
+                    OPTION_UNSAFE;
+                    RtsFlags.ProfFlags.doHeapProfile = HEAP_BY_INFO_TABLE;
+                    break;
                   default:
                     OPTION_SAFE;
                     PROFILING_BUILD_ONLY();
@@ -1338,8 +1446,13 @@ error = true;
                 if (rts_argv[arg][2] == '\0') {
                   /* use default */
                 } else {
+                    double intervalSeconds = parseDouble(rts_argv[arg]+2, &error);
+
+                    if (error) {
+                        errorBelch("bad value for -i");
+                    }
                     RtsFlags.ProfFlags.heapProfileInterval =
-                        fsecondsToTime(atof(rts_argv[arg]+2));
+                        fsecondsToTime(intervalSeconds);
                 }
                 break;
 
@@ -1349,8 +1462,13 @@ error = true;
                 if (rts_argv[arg][2] == '\0')
                     RtsFlags.ConcFlags.ctxtSwitchTime = 0;
                 else {
+                    double intervalSeconds = parseDouble(rts_argv[arg]+2, &error);
+
+                    if (error) {
+                        errorBelch("bad value for -C");
+                    }
                     RtsFlags.ConcFlags.ctxtSwitchTime =
-                        fsecondsToTime(atof(rts_argv[arg]+2));
+                        fsecondsToTime(intervalSeconds);
                 }
                 break;
 
@@ -1360,8 +1478,13 @@ error = true;
                     // turns off ticks completely
                     RtsFlags.MiscFlags.tickInterval = 0;
                 } else {
+                    double intervalSeconds = parseDouble(rts_argv[arg]+2, &error);
+
+                    if (error) {
+                        errorBelch("bad value for -V");
+                    }
                     RtsFlags.MiscFlags.tickInterval =
-                        fsecondsToTime(atof(rts_argv[arg]+2));
+                        fsecondsToTime(intervalSeconds);
                 }
                 break;
 
@@ -1587,10 +1710,6 @@ error = true;
                     OPTION_SAFE;
                     RtsFlags.GcFlags.useNonmoving = true;
                     unchecked_arg_start++;
-                    if (rts_argv[arg][3] == 's') {
-                        RtsFlags.GcFlags.nonmovingSelectorOpt = true;
-                        unchecked_arg_start++;
-                    }
                     break;
 
                 case 'c': /* Debugging tool: show current cost centre on
@@ -1604,11 +1723,9 @@ error = true;
 
                 case 't':  /* Include memory used by TSOs in a heap profile */
                     OPTION_SAFE;
-                    PROFILING_BUILD_ONLY(
-                        RtsFlags.ProfFlags.includeTSOs = true;
-                        );
-                    unchecked_arg_start++;
-                    goto check_rest;
+                    errorBelch("The -xt option has been removed (#16795)");
+                    error = true;
+                    break;
 
                   /*
                    * The option prefix '-xx' is reserved for future
@@ -1721,6 +1838,14 @@ static void normaliseRtsOpts (void)
         RtsFlags.ProfFlags.heapProfileIntervalTicks = 0;
     }
 
+    if (RtsFlags.TraceFlags.eventlogFlushTime > 0) {
+        RtsFlags.TraceFlags.eventlogFlushTicks =
+            RtsFlags.TraceFlags.eventlogFlushTime /
+            RtsFlags.MiscFlags.tickInterval;
+    } else {
+        RtsFlags.TraceFlags.eventlogFlushTicks = 0;
+    }
+
     if (RtsFlags.GcFlags.stkChunkBufferSize >
         RtsFlags.GcFlags.stkChunkSize / 2) {
         errorBelch("stack chunk buffer size (-kb) must be less than 50%%\n"
@@ -1771,6 +1896,16 @@ static void normaliseRtsOpts (void)
         barf("The non-moving collector doesn't support -G1");
     }
 
+#if !defined(PROFILING) && !defined(DEBUG)
+    // The mark-region collector is incompatible with heap census unless
+    // we zero slop of blackhole'd thunks, which doesn't happen in the
+    // vanilla way. See #9666.
+    if (RtsFlags.ProfFlags.doHeapProfile && RtsFlags.GcFlags.sweep) {
+        barf("The mark-region collector can only be used with profiling\n"
+             "when linked against the profiled RTS.");
+    }
+#endif
+
     if (RtsFlags.ProfFlags.doHeapProfile != NO_HEAP_PROFILING &&
             RtsFlags.GcFlags.useNonmoving) {
         barf("The non-moving collector doesn't support profiling");
@@ -1781,6 +1916,11 @@ static void normaliseRtsOpts (void)
                    "the compacting collector.");
         errorUsage();
     }
+
+    if (RtsFlags.TraceFlags.ticky && RtsFlags.TickyFlags.showTickyStats) {
+        barf("The ticky-ticky eventlog output cannot be used in conjunction with\n"
+             "+RTS -r<file>.");
+    }
 }
 
 static void errorUsage (void)
@@ -1924,6 +2064,35 @@ decodeSize(const char *flag, uint32_t offset, StgWord64 min, StgWord64 max)
     return val;
 }
 
+/* ----------------------------------------------------------------------------------
+ * parseDouble: parse a double from a string, setting a flag in case of an error
+-------------------------------------------------------------------------------- */
+
+static double
+parseDouble(const char *arg, bool *error)
+{
+    char *endptr;
+    double out;
+    errno = 0;
+
+    out = strtod(arg, &endptr);
+
+    if (errno != 0 || endptr == arg) {
+        *error = true;
+        return out;
+    }
+
+    while (isspace((unsigned char)*endptr)) {
+        ++endptr;
+    }
+
+    if (*endptr != 0) {
+        *error = true;
+    }
+
+    return out;
+}
+
 #if defined(DEBUG)
 static void read_debug_flags(const char* arg)
 {
@@ -2010,6 +2179,8 @@ static bool read_heap_profiling_flag(const char *arg)
     bool error = false;
     switch (arg[2]) {
     case '\0':
+      errorBelch("-h is deprecated and will be removed in the next release (9.4),\nuse -hc instead.");
+      FALLTHROUGH;
     case 'C':
     case 'c':
     case 'M':
@@ -2018,6 +2189,7 @@ static bool read_heap_profiling_flag(const char *arg)
     case 'd':
     case 'Y':
     case 'y':
+    case 'i':
     case 'R':
     case 'r':
     case 'B':
@@ -2098,6 +2270,9 @@ static bool read_heap_profiling_flag(const char *arg)
         case 'y':
             RtsFlags.ProfFlags.doHeapProfile = HEAP_BY_TYPE;
             break;
+        case 'i':
+            RtsFlags.ProfFlags.doHeapProfile = HEAP_BY_INFO_TABLE;
+            break;
         case 'R':
         case 'r':
             RtsFlags.ProfFlags.doHeapProfile = HEAP_BY_RETAINER;
@@ -2194,6 +2369,15 @@ static void read_trace_flags(const char *arg)
             RtsFlags.TraceFlags.user      = enabled;
             enabled = true;
             break;
+        case 'T':
+#if defined(TICKY_TICKY)
+            RtsFlags.TraceFlags.ticky     = enabled;
+            enabled = true;
+            break;
+#else
+            errorBelch("Program not compiled with ticky-ticky support");
+            break;
+#endif
         default:
             errorBelch("unknown trace option: %c",*c);
             break;
@@ -2421,3 +2605,16 @@ built in the -debug, -eventlog, -prof ways. And even if they do, the
 damage should be limited to DOS, information disclosure and writing
 files like <progname>.eventlog, not arbitrary files.
 */
+
+/* ----------------------------------------------------------------------------
+   Helper utilities to query state.
+   ------------------------------------------------------------------------- */
+
+bool is_io_mng_native_p (void)
+{
+#if defined(mingw32_HOST_OS)
+  return RtsFlags.MiscFlags.ioManager == IO_MNGR_NATIVE;
+#else
+  return false;
+#endif
+}
