@@ -32,6 +32,7 @@
 #include "StaticPtrTable.h"
 #include "Hash.h"
 #include "Profiling.h"
+#include "IPE.h"
 #include "ProfHeap.h"
 #include "Timer.h"
 #include "Globals.h"
@@ -46,12 +47,15 @@
 # include "RetainerProfile.h"
 #endif
 
+#include "IOManager.h"
 #if defined(mingw32_HOST_OS) && !defined(THREADED_RTS)
-#include "win32/AsyncIO.h"
+#include "win32/AsyncMIO.h"
+#include "win32/AsyncWinIO.h"
 #endif
 
 #if defined(mingw32_HOST_OS)
 #include <fenv.h>
+#include <windows.h>
 #else
 #include "posix/TTY.h"
 #endif
@@ -67,11 +71,16 @@
 static int hs_init_count = 0;
 static bool rts_shutdown = false;
 
+#if defined(mingw32_HOST_OS)
+/* Indicates CodePage to set program to after exit.  */
+static int64_t __codePage = -1;
+#endif
+
 static void flushStdHandles(void);
 
 /* -----------------------------------------------------------------------------
    Initialise floating point unit on x86 (currently disabled; See Note
-   [x86 Floating point precision] in compiler/nativeGen/X86/Instr.hs)
+   [x86 Floating point precision] in compiler/GHC/CmmToAsm/X86/Instr.hs)
    -------------------------------------------------------------------------- */
 
 #define X86_INIT_FPU 0
@@ -121,6 +130,46 @@ void fpreset(void) {
     _fpreset();
 }
 #endif
+
+/* Set the console's CodePage to UTF-8 if using the new I/O manager and the CP
+   is still the default one.  */
+static void
+initConsoleCP (void)
+{
+    /* Set the initial codepage to automatic.  */
+    __codePage = -1;
+
+    /* Check if the codepage is still the system default ANSI codepage.  */
+    if (GetConsoleCP () == GetOEMCP ()
+        && GetConsoleOutputCP () == GetOEMCP ()) {
+      if (!SetConsoleCP (CP_UTF8) || !SetConsoleOutputCP (CP_UTF8))
+        errorBelch ("Unable to set console CodePage, Unicode output may be "
+                    "garbled.\n");
+      else
+        IF_DEBUG (scheduler, debugBelch ("Codepage set to UTF-8.\n"));
+
+      /* Assign the codepage so we can restore it on exit.  */
+      __codePage = (int64_t)GetOEMCP ();
+    }
+}
+
+/* Restore the CodePage to what it was before we started.  If the CodePage was
+   already set then this call is a no-op.  */
+void
+hs_restoreConsoleCP (void)
+{
+    /* If we set the CP at startup, we should set it on exit.  */
+    if (__codePage == -1)
+      return;
+
+    UINT cp = (UINT)__codePage;
+    __codePage = -1;
+    if (SetConsoleCP (cp) && SetConsoleOutputCP (cp)) {
+      IF_DEBUG (scheduler, debugBelch ("Codepage restored to OEM.\n"));
+    } else {
+      IF_DEBUG (scheduler, debugBelch ("Unable to restore CodePage to OEM.\n"));
+    }
+}
 #endif
 
 /* -----------------------------------------------------------------------------
@@ -221,6 +270,12 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
 #endif /* DEBUG */
     }
 
+    /* Initialize console Codepage.  */
+#if defined(mingw32_HOST_OS)
+   if (is_io_mng_native_p())
+      initConsoleCP();
+#endif
+
     /* Initialise the stats department, phase 1 */
     initStats1();
 
@@ -245,8 +300,8 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
     initScheduler();
 
     /* Trace some basic information about the process */
-    traceWallClockTime();
-    traceOSProcessInfo();
+    traceInitEvent(traceWallClockTime);
+    traceInitEvent(traceOSProcessInfo);
     flushTrace();
 
     /* initialize the storage manager */
@@ -283,16 +338,15 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
     getStablePtr((StgPtr)cannotCompactPinned_closure);
     getStablePtr((StgPtr)cannotCompactMutable_closure);
     getStablePtr((StgPtr)nestedAtomically_closure);
-    getStablePtr((StgPtr)absentSumFieldError_closure);
-        // `Id` for this closure is marked as non-CAFFY,
-        // see Note [aBSENT_SUM_FIELD_ERROR_ID] in MkCore.
-
     getStablePtr((StgPtr)runSparks_closure);
     getStablePtr((StgPtr)ensureIOManagerIsRunning_closure);
+    getStablePtr((StgPtr)interruptIOManager_closure);
     getStablePtr((StgPtr)ioManagerCapabilitiesChanged_closure);
 #if !defined(mingw32_HOST_OS)
     getStablePtr((StgPtr)blockedOnBadFD_closure);
     getStablePtr((StgPtr)runHandlersPtr_closure);
+#else
+    getStablePtr((StgPtr)processRemoteCompletion_closure);
 #endif
 
     /*
@@ -310,14 +364,13 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
     /* initialise file locking, if necessary */
     initFileLocking();
 
-#if defined(DEBUG)
     /* initialise thread label table (tso->char*) */
     initThreadLabelTable();
-#endif
 
 #if defined(PROFILING)
     initProfiling();
 #endif
+    traceInitEvent(dumpIPEToEventLog);
     initHeapProfiling();
 
     /* start the virtual timer 'subsystem'. */
@@ -332,19 +385,12 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
     }
 #endif
 
-#if defined(mingw32_HOST_OS) && !defined(THREADED_RTS)
-    startupAsyncIO();
-#endif
+    initIOManager();
 
     x86_init_fpu();
 
     startupHpc();
 
-    // ditto.
-#if defined(THREADED_RTS)
-    ioManagerStart();
-#endif
-
     /* Record initialization times */
     stat_endInit();
 }
@@ -401,9 +447,7 @@ hs_exit_(bool wait_foreign)
     checkFPUStack();
 #endif
 
-#if defined(THREADED_RTS)
-    ioManagerDie();
-#endif
+    stopIOManager();
 
     /* stop all running tasks. This is also where we stop concurrent non-moving
      * collection if it's running */
@@ -432,6 +476,17 @@ hs_exit_(bool wait_foreign)
      */
     exitTimer(true);
 
+    /*
+     * Dump the ticky counter definitions
+     * We do this at the end of execution since tickers are registered in the
+     * course of program execution.
+     */
+#if defined(TICKY_TICKY) && defined(TRACING)
+    if (RtsFlags.TraceFlags.ticky) {
+        emitTickyCounterDefs();
+    }
+#endif
+
     // set the terminal settings back to what they were
 #if !defined(mingw32_HOST_OS)
     resetTerminalSettings();
@@ -454,6 +509,10 @@ hs_exit_(bool wait_foreign)
     // also outputs the stats (+RTS -s) info.
     exitStorage();
 
+    /* flush and clean up capabilities' eventlog buffers before cleaning up
+     * scheduler */
+    finishCapEventLogging();
+
     /* free the tasks */
     freeScheduler();
 
@@ -478,10 +537,8 @@ hs_exit_(bool wait_foreign)
     /* free the stable name table */
     exitStableNameTable();
 
-#if defined(DEBUG)
     /* free the thread label table */
     freeThreadLabelTable();
-#endif
 
 #if defined(PROFILING)
     reportCCSProfiling();
@@ -514,8 +571,12 @@ hs_exit_(bool wait_foreign)
     if (tf != NULL) fclose(tf);
 #endif
 
-#if defined(mingw32_HOST_OS) && !defined(THREADED_RTS)
-    shutdownAsyncIO(wait_foreign);
+    exitIOManager(wait_foreign);
+
+    /* Restore the console Codepage.  */
+#if defined(mingw32_HOST_OS)
+   if (is_io_mng_native_p())
+      hs_restoreConsoleCP();
 #endif
 
     /* tear down statistics subsystem */
