@@ -86,27 +86,6 @@ Time stat_getElapsedTime(void)
    Measure the current MUT time, for profiling
    ------------------------------------------------------------------------ */
 
-static double
-mut_user_time_until( Time t )
-{
-    ACQUIRE_LOCK(&stats_mutex);
-    double ret = TimeToSecondsDbl(t - stats.gc_cpu_ns - stats.nonmoving_gc_cpu_ns);
-    RELEASE_LOCK(&stats_mutex);
-    return ret;
-    // heapCensus() time is included in GC_tot_cpu, so we don't need
-    // to subtract it here.
-
-    // TODO: This seems wrong to me. Surely we should be subtracting
-    // (at least) start_init_cpu?
-}
-
-double
-mut_user_time( void )
-{
-    Time cpu = getProcessCPUTime();
-    return mut_user_time_until(cpu);
-}
-
 #if defined(PROFILING)
 /*
   mut_user_time_during_RP() returns the MUT time during retainer profiling.
@@ -175,13 +154,9 @@ initStats0(void)
         .par_copied_bytes = 0,
         .cumulative_par_max_copied_bytes = 0,
         .cumulative_par_balanced_copied_bytes = 0,
-        .gc_spin_spin = 0,
-        .gc_spin_yield = 0,
-        .mut_spin_spin = 0,
-        .mut_spin_yield = 0,
         .any_work = 0,
-        .no_work = 0,
         .scav_find_work = 0,
+        .max_n_todo_overflow = 0,
         .init_cpu_ns = 0,
         .init_elapsed_ns = 0,
         .mutator_cpu_ns = 0,
@@ -365,7 +340,12 @@ stat_endNonmovingGcSync ()
     stats.nonmoving_gc_sync_max_elapsed_ns =
       stg_max(stats.gc.nonmoving_gc_sync_elapsed_ns,
               stats.nonmoving_gc_sync_max_elapsed_ns);
+    Time sync_elapsed = stats.gc.nonmoving_gc_sync_elapsed_ns;
     RELEASE_LOCK(&stats_mutex);
+
+    if (RtsFlags.GcFlags.giveStats == VERBOSE_GC_STATS) {
+      statsPrintf("# sync %6.3f\n", TimeToSecondsDbl(sync_elapsed));
+    }
     traceConcSyncEnd();
 }
 
@@ -476,9 +456,8 @@ stat_startGC (Capability *cap, gc_thread *gct)
 void
 stat_endGC (Capability *cap, gc_thread *initiating_gct, W_ live, W_ copied, W_ slop,
             uint32_t gen, uint32_t par_n_threads, gc_thread **gc_threads,
-            W_ par_max_copied, W_ par_balanced_copied, W_ gc_spin_spin, W_ gc_spin_yield,
-            W_ mut_spin_spin, W_ mut_spin_yield, W_ any_work, W_ no_work,
-            W_ scav_find_work)
+            W_ par_max_copied, W_ par_balanced_copied, W_ any_work,
+            W_ scav_find_work, W_ max_n_todo_overflow)
 {
     ACQUIRE_LOCK(&stats_mutex);
 
@@ -558,12 +537,8 @@ stat_endGC (Capability *cap, gc_thread *initiating_gct, W_ live, W_ copied, W_ s
         stats.cumulative_par_balanced_copied_bytes +=
             stats.gc.par_balanced_copied_bytes;
         stats.any_work += any_work;
-        stats.no_work += no_work;
         stats.scav_find_work += scav_find_work;
-        stats.gc_spin_spin += gc_spin_spin;
-        stats.gc_spin_yield += gc_spin_yield;
-        stats.mut_spin_spin += mut_spin_spin;
-        stats.mut_spin_yield += mut_spin_yield;
+        stats.max_n_todo_overflow += stg_max(max_n_todo_overflow, stats.max_n_todo_overflow);
     }
     stats.gc_cpu_ns += stats.gc.cpu_ns;
     stats.gc_elapsed_ns += stats.gc.elapsed_ns;
@@ -594,7 +569,7 @@ stat_endGC (Capability *cap, gc_thread *initiating_gct, W_ live, W_ copied, W_ s
         // Emit events to the event log
 
         // Has to be emitted while all caps stopped for GC, but before GC_END.
-        // See trac.haskell.org/ThreadScope/wiki/RTSsummaryEvents
+        // See https://gitlab.haskell.org/ghc/ghc/-/wikis/RTSsummaryEvents
         // for a detailed design rationale of the current setup
         // of GC eventlog events.
         traceEventGcGlobalSync(cap);
@@ -658,6 +633,9 @@ stat_endGC (Capability *cap, gc_thread *initiating_gct, W_ live, W_ copied, W_ s
         traceEventHeapSize(cap,
                            CAPSET_HEAP_DEFAULT,
                            mblocks_allocated * MBLOCK_SIZE);
+        traceEventBlocksSize(cap,
+                           CAPSET_HEAP_DEFAULT,
+                           n_alloc_blocks * BLOCK_SIZE);
     }
     RELEASE_LOCK(&stats_mutex);
 }
@@ -984,16 +962,6 @@ static void report_summary(const RTSSummaryStats* sum)
                     , col_width[1], "gc_alloc_block_sync"
                     , col_width[2], gc_alloc_block_sync.spin
                     , col_width[3], gc_alloc_block_sync.yield);
-        statsPrintf("%*s" "%*s" "%*" FMT_Word64 "%*" FMT_Word64 "\n"
-                    , col_width[0], ""
-                    , col_width[1], "gc_spin"
-                    , col_width[2], stats.gc_spin_spin
-                    , col_width[3], stats.gc_spin_yield);
-        statsPrintf("%*s" "%*s" "%*" FMT_Word64 "%*" FMT_Word64 "\n"
-                    , col_width[0], ""
-                    , col_width[1], "mut_spin"
-                    , col_width[2], stats.mut_spin_spin
-                    , col_width[3], stats.mut_spin_yield);
         statsPrintf("%*s" "%*s" "%*" FMT_Word64 "%*s\n"
                     , col_width[0], ""
                     , col_width[1], "whitehole_gc"
@@ -1041,14 +1009,14 @@ static void report_summary(const RTSSummaryStats* sum)
                     , col_width[0], ""
                     , col_width[1], "any_work"
                     , col_width[2], stats.any_work);
-        statsPrintf("%*s" "%*s" "%*" FMT_Word64 "\n"
-                    , col_width[0], ""
-                    , col_width[1], "no_work"
-                    , col_width[2], stats.no_work);
         statsPrintf("%*s" "%*s" "%*" FMT_Word64 "\n"
                     , col_width[0], ""
                     , col_width[1], "scav_find_work"
                     , col_width[2], stats.scav_find_work);
+        statsPrintf("%*s" "%*s" "%*" FMT_Word64 "\n"
+                    , col_width[0], ""
+                    , col_width[1], "max_n_todo_overflow"
+                    , col_width[2], stats.max_n_todo_overflow);
 #elif defined(THREADED_RTS) // THREADED_RTS && PROF_SPIN
         statsPrintf("Internal Counters require the RTS to be built "
                 "with PROF_SPIN"); // PROF_SPIN is not #defined here
@@ -1170,10 +1138,6 @@ static void report_machine_readable (const RTSSummaryStats * sum)
     MR_STAT("gc_alloc_block_sync_yield", FMT_Word64,
             gc_alloc_block_sync.yield);
     MR_STAT("gc_alloc_block_sync_spin", FMT_Word64, gc_alloc_block_sync.spin);
-    MR_STAT("gc_spin_spin", FMT_Word64, stats.gc_spin_spin);
-    MR_STAT("gc_spin_yield", FMT_Word64, stats.gc_spin_yield);
-    MR_STAT("mut_spin_spin", FMT_Word64, stats.mut_spin_spin);
-    MR_STAT("mut_spin_yield", FMT_Word64, stats.mut_spin_yield);
     MR_STAT("waitForGcThreads_spin", FMT_Word64, waitForGcThreads_spin);
     MR_STAT("waitForGcThreads_yield", FMT_Word64,
             waitForGcThreads_yield);
@@ -1188,10 +1152,10 @@ static void report_machine_readable (const RTSSummaryStats * sum)
             whitehole_threadPaused_spin);
     MR_STAT("any_work", FMT_Word64,
             stats.any_work);
-    MR_STAT("no_work", FMT_Word64,
-            stats.no_work);
     MR_STAT("scav_find_work", FMT_Word64,
             stats.scav_find_work);
+    MR_STAT("max_n_todo_overflow", FMT_Word64,
+            stats.max_n_todo_overflow);
 #endif // PROF_SPIN
 #endif // THREADED_RTS
 
@@ -1213,6 +1177,26 @@ static void report_machine_readable (const RTSSummaryStats * sum)
         MR_STAT_GEN(g, "sync_yield", FMT_Word64, gc_sum->sync_yield);
 #endif
     }
+    // non-moving collector statistics
+    if (RtsFlags.GcFlags.useNonmoving) {
+        const int n_major_colls = sum->gc_summary_stats[RtsFlags.GcFlags.generations-1].collections;
+        MR_STAT("nonmoving_sync_wall_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_sync_elapsed_ns));
+        MR_STAT("nonmoving_sync_max_pause_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_sync_max_elapsed_ns));
+        MR_STAT("nonmoving_sync_avg_pause_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_sync_elapsed_ns) / n_major_colls);
+
+        MR_STAT("nonmoving_concurrent_cpu_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_cpu_ns));
+        MR_STAT("nonmoving_concurrent_wall_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_elapsed_ns));
+        MR_STAT("nonmoving_concurrent_max_pause_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_max_elapsed_ns));
+        MR_STAT("nonmoving_concurrent_avg_pause_seconds", "f",
+                TimeToSecondsDbl(stats.nonmoving_gc_elapsed_ns) / n_major_colls);
+    }
+
 
     statsPrintf(" ]\n");
 }
@@ -1563,7 +1547,7 @@ See #13830
 */
 
 /*
-Note [Internal Counter Stats]
+Note [Internal Counters Stats]
 -----------------------------
 What do the counts at the end of a '+RTS -s --internal-counters' report mean?
 They are detailed below. Most of these counters are used by multiple threads
@@ -1583,9 +1567,6 @@ Actual SpinLocks:
 * gc_alloc_block:
     This SpinLock protects the block allocator and free list manager. See
     BlockAlloc.c.
-* gc_spin and mut_spin:
-    These SpinLocks are used to herd gc worker threads during parallel garbage
-    collection. See gcWorkerThread, wakeup_gc_threads and releaseGCThreads.
 * gen[g].sync:
     These SpinLocks, one per generation, protect the generations[g] data
     structure during garbage collection.
@@ -1601,7 +1582,6 @@ don't. We count these white-hole spins and include them in the SpinLocks table.
 If a particular loop does not yield, we put "n/a" in the table. They are named
 for the function that has the spinning loop except that several loops in the
 garbage collector accumulate into whitehole_gc.
-TODO: Should these counters be more or less granular?
 
 white-hole spin counters:
 * whitehole_gc
@@ -1609,19 +1589,17 @@ white-hole spin counters:
 * whitehole_executeMessage
 * whitehole_threadPaused
 
-
-We count the number of calls of several functions in the parallel garbage
-collector.
+We have several stats allowing us to observe the internals of the parallel
+garbage collector:
 
 Parallel garbage collector counters:
 * any_work:
-    A cheap function called whenever a gc_thread is ready for work. Does
-    not do any work.
-* no_work:
-    Incremented whenever any_work finds no work.
+    Incremented whenever a parallel GC looks for work to steal.
 * scav_find_work:
-    Called to do work when any_work return true.
-
+    Counts iterations of scavenge loop
+* max_n_todo_overflow:
+    Tracks the maximum length of todo_overflow lists in the gc_thread structure.
+    See comment in grab_local_todo_block.
 */
 
 /* -----------------------------------------------------------------------------
