@@ -22,6 +22,7 @@
 #include "Threads.h"
 #include "Printer.h"
 #include "RtsFlags.h"
+#include "ThreadLabels.h"
 
 #if defined(HAVE_UNISTD_H)
 #include <unistd.h>
@@ -40,6 +41,10 @@ int TRACE_cap;
 static Mutex trace_utx;
 #endif
 
+#if defined(DEBUG)
+static void traceCap_stderr(Capability *cap, char *msg, ...);
+#endif
+
 /* ---------------------------------------------------------------------------
    Starting up / shutting down the tracing facilities
  --------------------------------------------------------------------------- */
@@ -117,10 +122,10 @@ void resetTracing (void)
     restartEventLogging();
 }
 
-void flushTrace (void)
+void flushTrace ()
 {
     if (eventlog_enabled) {
-        flushEventLog();
+        flushEventLog(NULL);
     }
 }
 
@@ -164,12 +169,14 @@ static char *thread_stop_reasons[] = {
     [6 + BlockedOnSTM]          = "blocked on STM",
     [6 + BlockedOnDoProc]       = "blocked on asyncDoProc",
     [6 + BlockedOnCCall]        = "blocked on a foreign call",
+    [6 + BlockedOnIOCompletion] = "blocked on I/O Completion port",
     [6 + BlockedOnCCall_Interruptible] = "blocked on a foreign call (interruptible)",
     [6 + BlockedOnMsgThrowTo]   =  "blocked on throwTo",
     [6 + ThreadMigrating]       =  "migrating"
 };
 #endif
 
+
 #if defined(DEBUG)
 static void traceSchedEvent_stderr (Capability *cap, EventTypeNum tag,
                                     StgTSO *tso,
@@ -179,45 +186,50 @@ static void traceSchedEvent_stderr (Capability *cap, EventTypeNum tag,
     ACQUIRE_LOCK(&trace_utx);
 
     tracePreface();
+    char *threadLabel = (char *)lookupThreadLabel(tso->id);
+    if(!threadLabel)
+    {
+        threadLabel = "";
+    }
     switch (tag) {
     case EVENT_CREATE_THREAD:   // (cap, thread)
-        debugBelch("cap %d: created thread %" FMT_Word "\n",
-                   cap->no, (W_)tso->id);
+        debugBelch("cap %d: created thread %" FMT_Word "[\"%s\"]" "\n",
+                   cap->no, (W_)tso->id, threadLabel);
         break;
     case EVENT_RUN_THREAD:      //  (cap, thread)
-        debugBelch("cap %d: running thread %" FMT_Word " (%s)\n",
-                   cap->no, (W_)tso->id, what_next_strs[tso->what_next]);
+        debugBelch("cap %d: running thread %" FMT_Word "[\"%s\"]"" (%s)\n",
+                   cap->no, (W_)tso->id, threadLabel, what_next_strs[tso->what_next]);
         break;
     case EVENT_THREAD_RUNNABLE: // (cap, thread)
-        debugBelch("cap %d: thread %" FMT_Word " appended to run queue\n",
-                   cap->no, (W_)tso->id);
+        debugBelch("cap %d: thread %" FMT_Word "[\"%s\"]"" appended to run queue\n",
+                   cap->no, (W_)tso->id, threadLabel);
         break;
     case EVENT_MIGRATE_THREAD:  // (cap, thread, new_cap)
-        debugBelch("cap %d: thread %" FMT_Word " migrating to cap %d\n",
-                   cap->no, (W_)tso->id, (int)info1);
+        debugBelch("cap %d: thread %" FMT_Word "[\"%s\"]" " migrating to cap %d\n",
+                   cap->no, (W_)tso->id, threadLabel, (int)info1);
         break;
     case EVENT_THREAD_WAKEUP:   // (cap, thread, info1_cap)
-        debugBelch("cap %d: waking up thread %" FMT_Word " on cap %d\n",
-                   cap->no, (W_)tso->id, (int)info1);
+        debugBelch("cap %d: waking up thread %" FMT_Word "[\"%s\"]" " on cap %d\n",
+                   cap->no, (W_)tso->id, threadLabel, (int)info1);
         break;
 
     case EVENT_STOP_THREAD:     // (cap, thread, status)
         if (info1 == 6 + BlockedOnBlackHole) {
-            debugBelch("cap %d: thread %" FMT_Word " stopped (blocked on black hole owned by thread %lu)\n",
-                       cap->no, (W_)tso->id, (long)info2);
+            debugBelch("cap %d: thread %" FMT_Word "[\"%s\"]" " stopped (blocked on black hole owned by thread %lu)\n",
+                       cap->no, (W_)tso->id, threadLabel, (long)info2);
         } else if (info1 == StackOverflow) {
-            debugBelch("cap %d: thead %" FMT_Word
+            debugBelch("cap %d: thread %" FMT_Word "[\"%s\"]"
                        " stopped (stack overflow, size %lu)\n",
-                      cap->no, (W_)tso->id, (long)info2);
+                      cap->no, (W_)tso->id, threadLabel, (long)info2);
 
         } else {
-            debugBelch("cap %d: thread %" FMT_Word " stopped (%s)\n",
-                       cap->no, (W_)tso->id, thread_stop_reasons[info1]);
+            debugBelch("cap %d: thread %" FMT_Word "[\"%s\"]" " stopped (%s)\n",
+                       cap->no, (W_)tso->id, threadLabel, thread_stop_reasons[info1]);
         }
         break;
     default:
-        debugBelch("cap %d: thread %" FMT_Word ": event %d\n\n",
-                   cap->no, (W_)tso->id, tag);
+        debugBelch("cap %d: thread %" FMT_Word "[\"%s\"]" ": event %d\n\n",
+                   cap->no, (W_)tso->id, threadLabel, tag);
         break;
     }
 
@@ -362,6 +374,23 @@ void traceEventGcStats_  (Capability *cap,
     }
 }
 
+void traceEventMemReturn_ (Capability *cap,
+                          uint32_t    current_mblocks,
+                          uint32_t    needed_mblocks,
+                          uint32_t    returned_mblocks)
+{
+#if defined(DEBUG)
+    if (RtsFlags.TraceFlags.tracing == TRACE_STDERR) {
+        traceCap_stderr(cap, "Memory Return (Current: %u) (Needed: %u) (Returned: %u)"
+                       , current_mblocks, needed_mblocks, returned_mblocks);
+    } else
+#endif
+    {
+        postEventMemReturn( cap, CAPSET_HEAP_DEFAULT
+                          , current_mblocks, needed_mblocks, returned_mblocks);
+    }
+}
+
 void traceCapEvent_ (Capability   *cap,
                      EventTypeNum  tag)
 {
@@ -631,6 +660,19 @@ void traceHeapProfSampleString(StgWord8 profile_id,
     }
 }
 
+void traceIPE(StgInfoTable * info,
+              const char *table_name,
+              const char *closure_desc,
+              const char *ty_desc,
+              const char *label,
+              const char *module,
+              const char *srcloc )
+{
+    if (eventlog_enabled) {
+        postIPE((W_) INFO_PTR_TO_STRUCT(info), table_name, closure_desc, ty_desc, label, module, srcloc);
+    }
+}
+
 #if defined(PROFILING)
 void traceHeapProfCostCentre(StgWord32 ccID,
                              const char *label,
