@@ -87,6 +87,12 @@
 #include <numa.h>
 #endif
 
+// For gettimeofday()
+#include <sys/time.h>
+
+// TODO does this need configure magic?
+#include <time.h>
+
 /*
  * This (allegedly) OS threads independent layer was initially
  * abstracted away from code that used Pthreads, so the functions
@@ -97,33 +103,70 @@
 void
 initCondition( Condition* pCond )
 {
-  pthread_cond_init(pCond, NULL);
-  return;
+  pthread_condattr_t attr;
+  CHECK(pthread_condattr_init(&attr) == 0);
+#if defined(HAVE_CLOCK_GETTIME) && defined(HAVE_PTHREAD_CONDATTR_SETCLOCK)
+  pCond->timeout_clk = CLOCK_REALTIME;
+  if (pthread_condattr_setclock(&attr, CLOCK_MONOTONIC) == 0) {
+      pCond->timeout_clk = CLOCK_MONOTONIC;
+  }
+#endif
+  CHECK(pthread_cond_init(&pCond->cond, &attr) == 0);
+  CHECK(pthread_condattr_destroy(&attr) == 0);
 }
 
 void
 closeCondition( Condition* pCond )
 {
-  pthread_cond_destroy(pCond);
-  return;
+  CHECK(pthread_cond_destroy(&pCond->cond) == 0);
 }
 
-bool
+void
 broadcastCondition ( Condition* pCond )
 {
-  return (pthread_cond_broadcast(pCond) == 0);
+  CHECK(pthread_cond_broadcast(&pCond->cond) == 0);
 }
 
-bool
+void
 signalCondition ( Condition* pCond )
 {
-  return (pthread_cond_signal(pCond) == 0);
+  CHECK(pthread_cond_signal(&pCond->cond) == 0);
+}
+
+void
+waitCondition ( Condition* pCond, Mutex* pMut )
+{
+  CHECK(pthread_cond_wait(&pCond->cond, pMut) == 0);
 }
 
 bool
-waitCondition ( Condition* pCond, Mutex* pMut )
-{
-  return (pthread_cond_wait(pCond,pMut) == 0);
+timedWaitCondition ( Condition* pCond, Mutex* pMut, Time timeout) {
+    struct timespec ts;
+
+#if defined(HAVE_CLOCK_GETTIME) && defined(HAVE_PTHREAD_CONDATTR_SETCLOCK)
+    CHECK(clock_gettime(pCond->timeout_clk, &ts) == 0);
+#else
+    struct timeval tv;
+    CHECK(gettimeofday(&tv, NULL) == 0);
+    ts.tv_sec = tv.tv_sec;
+    ts.tv_nsec = 1000 * tv.tv_usec;
+#endif
+
+    uint64_t sec = TimeToSeconds(timeout);
+    ts.tv_sec += sec;
+    ts.tv_nsec += TimeToNS(timeout - SecondsToTime(sec));
+    ts.tv_sec += ts.tv_nsec / 1000000000;
+    ts.tv_nsec %= 1000000000;
+
+    int ret = pthread_cond_timedwait(&pCond->cond, pMut, &ts);
+    switch (ret) {
+    case ETIMEDOUT:
+        return false;
+    case 0:
+        return true;
+    default:
+        barf("pthread_cond_timedwait failed");
+    }
 }
 
 void
@@ -261,22 +304,43 @@ uint32_t
 getNumberOfProcessors (void)
 {
     uint32_t nproc = RELAXED_LOAD(&nproc_cache);
-
     if (nproc == 0) {
-#if defined(HAVE_SYSCONF) && defined(_SC_NPROCESSORS_ONLN)
-        nproc = sysconf(_SC_NPROCESSORS_ONLN);
-#elif defined(HAVE_SYSCONF) && defined(_SC_NPROCESSORS_CONF)
-        nproc = sysconf(_SC_NPROCESSORS_CONF);
-#elif defined(darwin_HOST_OS)
+#if defined(HAVE_SCHED_GETAFFINITY)
+        cpu_set_t mask;
+        CPU_ZERO(&mask);
+        if (sched_getaffinity(0, sizeof(mask), &mask) == 0) {
+            for (int i = 0; i < CPU_SETSIZE; i++) {
+                if (CPU_ISSET(i, &mask))
+                    nproc++;
+            }
+            return nproc;
+        }
+#endif
+
+#if defined(darwin_HOST_OS)
         size_t size = sizeof(uint32_t);
-        if(sysctlbyname("hw.logicalcpu",&nproc,&size,NULL,0) != 0) {
+        if (sysctlbyname("machdep.cpu.thread_count",&nproc,&size,NULL,0) != 0) {
+            if (sysctlbyname("hw.logicalcpu",&nproc,&size,NULL,0) != 0) {
+                if (sysctlbyname("hw.ncpu",&nproc,&size,NULL,0) != 0)
+                    nproc = 1;
+            }
+        }
+#elif defined(freebsd_HOST_OS)
+        cpuset_t mask;
+        CPU_ZERO(&mask);
+        if(cpuset_getaffinity(CPU_LEVEL_CPUSET, CPU_WHICH_PID, -1, sizeof(mask), &mask) == 0) {
+            return CPU_COUNT(&mask);
+        } else {
+            size_t size = sizeof(uint32_t);
             if(sysctlbyname("hw.ncpu",&nproc,&size,NULL,0) != 0)
                 nproc = 1;
         }
-#elif defined(freebsd_HOST_OS)
-        size_t size = sizeof(uint32_t);
-        if(sysctlbyname("hw.ncpu",&nproc,&size,NULL,0) != 0)
-            nproc = 1;
+#elif defined(HAVE_SYSCONF) && defined(_SC_NPROCESSORS_ONLN)
+        // N.B. This is the number of physical processors.
+        nproc = sysconf(_SC_NPROCESSORS_ONLN);
+#elif defined(HAVE_SYSCONF) && defined(_SC_NPROCESSORS_CONF)
+        // N.B. This is the number of physical processors.
+        nproc = sysconf(_SC_NPROCESSORS_CONF);
 #else
         nproc = 1;
 #endif
