@@ -12,6 +12,7 @@
 #include "Schedule.h"
 #include "RtsSignals.h"
 #include "Signals.h"
+#include "IOManager.h"
 #include "RtsUtils.h"
 #include "Prelude.h"
 #include "Ticker.h"
@@ -632,7 +633,7 @@ set_sigtstp_action (bool handle)
 void
 install_vtalrm_handler(int sig, TickProc handle_tick)
 {
-    struct sigaction action;
+    struct sigaction action = {};
 
     action.sa_handler = handle_tick;
 
@@ -674,20 +675,17 @@ install_vtalrm_handler(int sig, TickProc handle_tick)
 void
 initDefaultHandlers(void)
 {
-    struct sigaction action,oact;
+    struct sigaction action = {};
+    struct sigaction oact = {};
 
     // install the SIGINT handler
     action.sa_handler = shutdown_handler;
     sigemptyset(&action.sa_mask);
-    action.sa_flags = 0;
+    action.sa_flags = 0; // disable SA_RESTART
     if (sigaction(SIGINT, &action, &oact) != 0) {
         sysErrorBelch("warning: failed to install SIGINT handler");
     }
 
-#if defined(HAVE_SIGINTERRUPT)
-    siginterrupt(SIGINT, 1);    // isn't this the default? --SDM
-#endif
-
     // install the SIGFPE handler
 
     // In addition to handling SIGINT, also handle SIGFPE by ignoring it.
