@@ -202,7 +202,7 @@ Mutex concurrent_coll_finished_lock;
  *    generation.
  *
  *  - Note [Aging under the non-moving collector] (NonMoving.c) describes how
- *    we accomodate aging
+ *    we accommodate aging
  *
  *  - Note [Non-moving GC: Marking evacuated objects] (Evac.c) describes how
  *    non-moving objects reached by evacuate() are marked, which is necessary
@@ -229,6 +229,10 @@ Mutex concurrent_coll_finished_lock;
  *  - Note [StgStack dirtiness flags and concurrent marking] (TSO.h) describes
  *    the protocol for concurrent marking of stacks.
  *
+ *  - Note [Nonmoving write barrier in Perform{Take,Put}] (PrimOps.cmm) describes
+ *    a tricky barrier necessary when resuming threads blocked on MVar
+ *    operations.
+ *
  *  - Note [Static objects under the nonmoving collector] (Storage.c) describes
  *    treatment of static objects.
  *
@@ -407,7 +411,7 @@ Mutex concurrent_coll_finished_lock;
  * Note [Dirty flags in the non-moving collector]
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  * Some mutable object types (e.g. MUT_VARs, TVARs) have a one-bit dirty flag
- * encoded in their info table pointer. The moving collector's uses this flag
+ * encoded in their info table pointer. The moving collector uses this flag
  * to minimize redundant mut_list entries. The flag is preserves the following
  * simple invariant:
  *
@@ -430,7 +434,7 @@ Mutex concurrent_coll_finished_lock;
  *     An object being marked as dirty implies that all of its fields are on
  *     the mark queue (or, equivalently, update remembered set).
  *
- * With this guarantee we can safely make the the write barriers dirty objects
+ * With this guarantee we can safely make the write barriers dirty objects
  * no-ops. We perform this optimisation for the following object types:
  *
  *  - MVAR
@@ -982,7 +986,7 @@ void nonmovingCollect(StgWeak **dead_weaks, StgTSO **resurrected_threads)
     // If we're interrupting or shutting down, do not let this capability go and
     // run a STW collection. Reason: we won't be able to acquire this capability
     // again for the sync if we let it go, because it'll immediately start doing
-    // a major GC, becuase that's what we do when exiting scheduler (see
+    // a major GC, because that's what we do when exiting scheduler (see
     // exitScheduler()).
     if (sched_state == SCHED_RUNNING) {
         concurrent_coll_running = true;
@@ -1002,7 +1006,7 @@ void nonmovingCollect(StgWeak **dead_weaks, StgTSO **resurrected_threads)
 #endif
 }
 
-/* Mark mark queue, threads, and weak pointers until no more weaks have been
+/* Mark queue, threads, and weak pointers until no more weaks have been
  * resuscitated
  */
 static void nonmovingMarkThreadsWeaks(MarkQueue *mark_queue)
@@ -1226,7 +1230,7 @@ static void nonmovingMark_(MarkQueue *mark_queue, StgWeak **dead_weaks, StgTSO *
 
 #if defined(THREADED_RTS)
 finish:
-    boundTaskExiting(task);
+    exitMyTask();
 
     // We are done...
     mark_thread = 0;
