@@ -29,12 +29,12 @@
 
 static bool check_in_nonmoving_heap(StgClosure *p);
 static void mark_closure (MarkQueue *queue, const StgClosure *p, StgClosure **origin);
-static void mark_tso (MarkQueue *queue, StgTSO *tso);
-static void mark_stack (MarkQueue *queue, StgStack *stack);
-static void mark_PAP_payload (MarkQueue *queue,
-                              StgClosure *fun,
-                              StgClosure **payload,
-                              StgWord size);
+static void trace_tso (MarkQueue *queue, StgTSO *tso);
+static void trace_stack (MarkQueue *queue, StgStack *stack);
+static void trace_PAP_payload (MarkQueue *queue,
+                               StgClosure *fun,
+                               StgClosure **payload,
+                               StgWord size);
 
 // How many Array# entries to add to the mark queue at once?
 #define MARK_ARRAY_CHUNK_LENGTH 128
@@ -90,7 +90,7 @@ memcount n_nonmoving_marked_compact_blocks = 0;
  * move the same large object to nonmoving_marked_large_objects more than once.
  */
 static Mutex nonmoving_large_objects_mutex;
-// Note that we don't need a similar lock for compact objects becuase we never
+// Note that we don't need a similar lock for compact objects because we never
 // mark a compact object eagerly in a write barrier; all compact objects are
 // marked by the mark thread, so there can't be any races here.
 #endif
@@ -190,8 +190,8 @@ StgIndStatic *debug_caf_list_snapshot = (StgIndStatic*)END_OF_CAF_LIST;
  *  4. The mark thread marks everything it was sent; runs out of things to mark
  *  5. Mark thread initiates a sync
  *  6. Capabilities send their final update remembered sets and suspend execution
- *  7. Mark thread marks everything is was sent
- *  8. Mark thead allows capabilities to resume.
+ *  7. Mark thread marks everything it was sent
+ *  8. Mark thread allows capabilities to resume.
  *
  * However, this is obviously a fair amount of complexity and so far the
  * periodic eager flushing approach has been sufficient.
@@ -629,7 +629,7 @@ void updateRemembSetPushThunkEager(Capability *cap,
         if (check_in_nonmoving_heap(ap->fun)) {
             push_closure(queue, ap->fun, NULL);
         }
-        mark_PAP_payload(queue, ap->fun, ap->payload, ap->n_args);
+        trace_PAP_payload(queue, ap->fun, ap->payload, ap->n_args);
         break;
     }
     case THUNK_SELECTOR:
@@ -715,7 +715,7 @@ void updateRemembSetPushTSO(Capability *cap, StgTSO *tso)
 {
     if (needs_upd_rem_set_mark((StgClosure *) tso)) {
         debugTrace(DEBUG_nonmoving_gc, "upd_rem_set: TSO %p", tso);
-        mark_tso(&cap->upd_rem_set.queue, tso);
+        trace_tso(&cap->upd_rem_set.queue, tso);
         finish_upd_rem_set_mark((StgClosure *) tso);
     }
 }
@@ -730,7 +730,7 @@ void updateRemembSetPushStack(Capability *cap, StgStack *stack)
               != nonmovingMarkEpoch) {
             // We have claimed the right to mark the stack.
             debugTrace(DEBUG_nonmoving_gc, "upd_rem_set: STACK %p", stack->sp);
-            mark_stack(&cap->upd_rem_set.queue, stack);
+            trace_stack(&cap->upd_rem_set.queue, stack);
             finish_upd_rem_set_mark((StgClosure *) stack);
             return;
         } else {
@@ -790,8 +790,8 @@ void markQueuePushThunkSrt (MarkQueue *q, const StgInfoTable *info)
 }
 
 void markQueuePushArray (MarkQueue *q,
-                            const StgMutArrPtrs *array,
-                            StgWord start_index)
+                         const StgMutArrPtrs *array,
+                         StgWord start_index)
 {
     push_array(q, array, start_index);
 }
@@ -941,19 +941,25 @@ markQueueLength (MarkQueue *q)
  * any outstanding transactions.
  */
 static void
-mark_trec_header (MarkQueue *queue, StgTRecHeader *trec)
+trace_trec_chunk (MarkQueue *queue, StgTRecChunk *chunk)
+{
+    markQueuePushClosure_(queue, (StgClosure *) chunk);
+    for (StgWord i=0; i < chunk->next_entry_idx; i++) {
+        TRecEntry *ent = &chunk->entries[i];
+        markQueuePushClosure_(queue, (StgClosure *) ent->tvar);
+        markQueuePushClosure_(queue, ent->expected_value);
+        markQueuePushClosure_(queue, ent->new_value);
+    }
+}
+
+static void
+trace_trec_header (MarkQueue *queue, StgTRecHeader *trec)
 {
     while (trec != NO_TREC) {
         StgTRecChunk *chunk = trec->current_chunk;
         markQueuePushClosure_(queue, (StgClosure *) trec);
-        markQueuePushClosure_(queue, (StgClosure *) chunk);
         while (chunk != END_STM_CHUNK_LIST) {
-            for (StgWord i=0; i < chunk->next_entry_idx; i++) {
-                TRecEntry *ent = &chunk->entries[i];
-                markQueuePushClosure_(queue, (StgClosure *) ent->tvar);
-                markQueuePushClosure_(queue, ent->expected_value);
-                markQueuePushClosure_(queue, ent->new_value);
-            }
+            trace_trec_chunk(queue, chunk);
             chunk = chunk->prev_chunk;
         }
         trec = trec->enclosing_trec;
@@ -961,7 +967,7 @@ mark_trec_header (MarkQueue *queue, StgTRecHeader *trec)
 }
 
 static void
-mark_tso (MarkQueue *queue, StgTSO *tso)
+trace_tso (MarkQueue *queue, StgTSO *tso)
 {
     // TODO: Clear dirty if contains only old gen objects
 
@@ -971,7 +977,7 @@ mark_tso (MarkQueue *queue, StgTSO *tso)
 
     markQueuePushClosure_(queue, (StgClosure *) tso->blocked_exceptions);
     markQueuePushClosure_(queue, (StgClosure *) tso->bq);
-    mark_trec_header(queue, tso->trec);
+    trace_trec_header(queue, tso->trec);
     markQueuePushClosure_(queue, (StgClosure *) tso->stackobj);
     markQueuePushClosure_(queue, (StgClosure *) tso->_link);
     if (   tso->why_blocked == BlockedOnMVar
@@ -993,16 +999,16 @@ do_push_closure (StgClosure **p, void *user)
 }
 
 static void
-mark_large_bitmap (MarkQueue *queue,
-                   StgClosure **p,
-                   StgLargeBitmap *large_bitmap,
-                   StgWord size)
+trace_large_bitmap (MarkQueue *queue,
+                    StgClosure **p,
+                    StgLargeBitmap *large_bitmap,
+                    StgWord size)
 {
     walk_large_bitmap(do_push_closure, p, large_bitmap, size, queue);
 }
 
 static void
-mark_small_bitmap (MarkQueue *queue, StgClosure **p, StgWord size, StgWord bitmap)
+trace_small_bitmap (MarkQueue *queue, StgClosure **p, StgWord size, StgWord bitmap)
 {
     while (size > 0) {
         if ((bitmap & 1) == 0) {
@@ -1016,10 +1022,10 @@ mark_small_bitmap (MarkQueue *queue, StgClosure **p, StgWord size, StgWord bitma
 }
 
 static GNUC_ATTR_HOT
-void mark_PAP_payload (MarkQueue *queue,
-                       StgClosure *fun,
-                       StgClosure **payload,
-                       StgWord size)
+void trace_PAP_payload (MarkQueue *queue,
+                        StgClosure *fun,
+                        StgClosure **payload,
+                        StgWord size)
 {
     const StgFunInfoTable *fun_info = get_fun_itbl(UNTAG_CONST_CLOSURE(fun));
     ASSERT(fun_info->i.type != PAP);
@@ -1031,20 +1037,20 @@ void mark_PAP_payload (MarkQueue *queue,
         bitmap = BITMAP_BITS(fun_info->f.b.bitmap);
         goto small_bitmap;
     case ARG_GEN_BIG:
-        mark_large_bitmap(queue, payload, GET_FUN_LARGE_BITMAP(fun_info), size);
+        trace_large_bitmap(queue, payload, GET_FUN_LARGE_BITMAP(fun_info), size);
         break;
     case ARG_BCO:
-        mark_large_bitmap(queue, payload, BCO_BITMAP(fun), size);
+        trace_large_bitmap(queue, payload, BCO_BITMAP(fun), size);
         break;
     default:
         bitmap = BITMAP_BITS(stg_arg_bitmaps[fun_info->f.fun_type]);
     small_bitmap:
-        mark_small_bitmap(queue, (StgClosure **) p, size, bitmap);
+        trace_small_bitmap(queue, (StgClosure **) p, size, bitmap);
         break;
     }
 }
 
-/* Helper for mark_stack; returns next stack frame. */
+/* Helper for trace_stack; returns next stack frame. */
 static StgPtr
 mark_arg_block (MarkQueue *queue, const StgFunInfoTable *fun_info, StgClosure **args)
 {
@@ -1058,14 +1064,14 @@ mark_arg_block (MarkQueue *queue, const StgFunInfoTable *fun_info, StgClosure **
         goto small_bitmap;
     case ARG_GEN_BIG:
         size = GET_FUN_LARGE_BITMAP(fun_info)->size;
-        mark_large_bitmap(queue, (StgClosure**)p, GET_FUN_LARGE_BITMAP(fun_info), size);
+        trace_large_bitmap(queue, (StgClosure**)p, GET_FUN_LARGE_BITMAP(fun_info), size);
         p += size;
         break;
     default:
         bitmap = BITMAP_BITS(stg_arg_bitmaps[fun_info->f.fun_type]);
         size = BITMAP_SIZE(stg_arg_bitmaps[fun_info->f.fun_type]);
     small_bitmap:
-        mark_small_bitmap(queue, (StgClosure**)p, size, bitmap);
+        trace_small_bitmap(queue, (StgClosure**)p, size, bitmap);
         p += size;
         break;
     }
@@ -1073,7 +1079,7 @@ mark_arg_block (MarkQueue *queue, const StgFunInfoTable *fun_info, StgClosure **
 }
 
 static GNUC_ATTR_HOT void
-mark_stack_ (MarkQueue *queue, StgPtr sp, StgPtr spBottom)
+trace_stack_ (MarkQueue *queue, StgPtr sp, StgPtr spBottom)
 {
     ASSERT(sp <= spBottom);
 
@@ -1103,7 +1109,7 @@ mark_stack_ (MarkQueue *queue, StgPtr sp, StgPtr spBottom)
             // NOTE: the payload starts immediately after the info-ptr, we
             // don't have an StgHeader in the same sense as a heap closure.
             sp++;
-            mark_small_bitmap(queue, (StgClosure **) sp, size, bitmap);
+            trace_small_bitmap(queue, (StgClosure **) sp, size, bitmap);
             sp += size;
         }
         follow_srt:
@@ -1118,7 +1124,7 @@ mark_stack_ (MarkQueue *queue, StgPtr sp, StgPtr spBottom)
             StgBCO *bco = (StgBCO *)*sp;
             sp++;
             StgWord size = BCO_BITMAP_SIZE(bco);
-            mark_large_bitmap(queue, (StgClosure **) sp, BCO_BITMAP(bco), size);
+            trace_large_bitmap(queue, (StgClosure **) sp, BCO_BITMAP(bco), size);
             sp += size;
             continue;
         }
@@ -1130,7 +1136,7 @@ mark_stack_ (MarkQueue *queue, StgPtr sp, StgPtr spBottom)
 
             size = GET_LARGE_BITMAP(&info->i)->size;
             sp++;
-            mark_large_bitmap(queue, (StgClosure **) sp, GET_LARGE_BITMAP(&info->i), size);
+            trace_large_bitmap(queue, (StgClosure **) sp, GET_LARGE_BITMAP(&info->i), size);
             sp += size;
             // and don't forget to follow the SRT
             goto follow_srt;
@@ -1148,17 +1154,17 @@ mark_stack_ (MarkQueue *queue, StgPtr sp, StgPtr spBottom)
         }
 
         default:
-            barf("mark_stack: weird activation record found on stack: %d", (int)(info->i.type));
+            barf("trace_stack: weird activation record found on stack: %d", (int)(info->i.type));
         }
     }
 }
 
 static GNUC_ATTR_HOT void
-mark_stack (MarkQueue *queue, StgStack *stack)
+trace_stack (MarkQueue *queue, StgStack *stack)
 {
     // TODO: Clear dirty if contains only old gen objects
 
-    mark_stack_(queue, stack->sp, stack->stack + stack->stack_size);
+    trace_stack_(queue, stack->sp, stack->stack + stack->stack_size);
 }
 
 /* See Note [Static objects under the nonmoving collector].
@@ -1278,7 +1284,7 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     ASSERT(!IS_FORWARDING_PTR(p->header.info));
 
     // N.B. only the first block of a compact region is guaranteed to carry
-    // BF_NONMOVING; conseqently we must separately check for BF_COMPACT.
+    // BF_NONMOVING; consequently we must separately check for BF_COMPACT.
     if (bd->flags & (BF_COMPACT | BF_NONMOVING)) {
 
         if (bd->flags & BF_COMPACT) {
@@ -1338,7 +1344,7 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     }
 
     // A pinned object that is still attached to a capability (because it's not
-    // filled yet). No need to trace it pinned objects can't contain poiners.
+    // filled yet). No need to trace it pinned objects can't contain pointers.
     else if (bd->flags & BF_PINNED) {
 #if defined(DEBUG)
         bool found_it = false;
@@ -1516,31 +1522,27 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     }
 
     case THUNK_SELECTOR:
-        if (RtsFlags.GcFlags.nonmovingSelectorOpt) {
-            nonmoving_eval_thunk_selector(queue, (StgSelector*)p, origin);
-        } else {
-            PUSH_FIELD((StgSelector *) p, selectee);
-        }
+        nonmoving_eval_thunk_selector(queue, (StgSelector*)p, origin);
         break;
 
     case AP_STACK: {
         StgAP_STACK *ap = (StgAP_STACK *)p;
         PUSH_FIELD(ap, fun);
-        mark_stack_(queue, (StgPtr) ap->payload, (StgPtr) ap->payload + ap->size);
+        trace_stack_(queue, (StgPtr) ap->payload, (StgPtr) ap->payload + ap->size);
         break;
     }
 
     case PAP: {
         StgPAP *pap = (StgPAP *) p;
         PUSH_FIELD(pap, fun);
-        mark_PAP_payload(queue, pap->fun, pap->payload, pap->n_args);
+        trace_PAP_payload(queue, pap->fun, pap->payload, pap->n_args);
         break;
     }
 
     case AP: {
         StgAP *ap = (StgAP *) p;
         PUSH_FIELD(ap, fun);
-        mark_PAP_payload(queue, ap->fun, ap->payload, ap->n_args);
+        trace_PAP_payload(queue, ap->fun, ap->payload, ap->n_args);
         break;
     }
 
@@ -1552,7 +1554,6 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     case MUT_ARR_PTRS_DIRTY:
     case MUT_ARR_PTRS_FROZEN_CLEAN:
     case MUT_ARR_PTRS_FROZEN_DIRTY:
-        // TODO: Check this against Scav.c
         markQueuePushArray(queue, (StgMutArrPtrs *) p, 0);
         break;
 
@@ -1569,7 +1570,7 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     }
 
     case TSO:
-        mark_tso(queue, (StgTSO *) p);
+        trace_tso(queue, (StgTSO *) p);
         break;
 
     case STACK: {
@@ -1582,7 +1583,7 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
         if (cas_word8(&stack->marking, marking, nonmovingMarkEpoch)
               != nonmovingMarkEpoch) {
             // We have claimed the right to mark the stack.
-            mark_stack(queue, stack);
+            trace_stack(queue, stack);
         } else {
             // A mutator has already started marking the stack; we just let it
             // do its thing and move on. There's no reason to wait; we know that
@@ -1602,19 +1603,9 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
         break;
     }
 
-    case TREC_CHUNK: {
-        // TODO: Should we abort here? This should have already been marked
-        // when we dirtied the TSO
-        StgTRecChunk *tc = ((StgTRecChunk *) p);
-        PUSH_FIELD(tc, prev_chunk);
-        TRecEntry *end = &tc->entries[tc->next_entry_idx];
-        for (TRecEntry *e = &tc->entries[0]; e < end; e++) {
-            markQueuePushClosure_(queue, (StgClosure *) e->tvar);
-            markQueuePushClosure_(queue, (StgClosure *) e->expected_value);
-            markQueuePushClosure_(queue, (StgClosure *) e->new_value);
-        }
+    case TREC_CHUNK:
+        // N.B. chunk contents are deeply marked by trace_trec_header
         break;
-    }
 
     case WHITEHOLE:
         while (*(StgInfoTable* volatile*) &p->header.info == &stg_WHITEHOLE_info);
@@ -1903,7 +1894,7 @@ void nonmovingMarkDeadWeaks (struct MarkQueue_ *queue, StgWeak **dead_weaks)
     }
 }
 
-// Non-moving heap variant of of `tidyThreadList`
+// Non-moving heap variant of `tidyThreadList`
 void nonmovingTidyThreads ()
 {
     StgTSO *next;
