@@ -61,6 +61,7 @@ StgIndStatic  *dyn_caf_list        = NULL;
 StgIndStatic  *debug_caf_list      = NULL;
 StgIndStatic  *revertible_caf_list = NULL;
 bool           keepCAFs;
+bool           highMemDynamic;
 
 W_ large_alloc_lim;    /* GC if n_large_blocks in any nursery
                         * reaches this. */
@@ -82,6 +83,16 @@ generation *oldest_gen  = NULL; /* oldest generation, for convenience */
 nursery *nurseries = NULL;
 uint32_t n_nurseries;
 
+/* Pinned Nursery Size, the number of blocks that we reserve for
+ * pinned data. The number chosen here decides whether pinned objects
+ * are allocated from the free_list (if n < BLOCKS_PER_MBLOCK) or whether
+ * a fresh mblock is allocated each time.
+ * See Note [Sources of Block Level Fragmentation]
+ * */
+
+#define PINNED_EMPTY_SIZE BLOCKS_PER_MBLOCK
+
+
 /*
  * When we are using nursery chunks, we need a separate next_nursery
  * pointer for each NUMA node.
@@ -139,6 +150,22 @@ initGeneration (generation *gen, int g)
     gen->old_weak_ptr_list = NULL;
 }
 
+
+#if defined(TRACING)
+// Defined as it's own top-level function so it can be passed to traceInitEvent
+static void
+traceHeapInfo (void){
+  traceEventHeapInfo(CAPSET_HEAP_DEFAULT,
+                     RtsFlags.GcFlags.generations,
+                     RtsFlags.GcFlags.maxHeapSize * BLOCK_SIZE,
+                     RtsFlags.GcFlags.minAllocAreaSize * BLOCK_SIZE,
+                     MBLOCK_SIZE,
+                     BLOCK_SIZE);
+}
+#else
+#define traceHeapInfo
+#endif
+
 void
 initStorage (void)
 {
@@ -235,12 +262,8 @@ initStorage (void)
 
   RELEASE_SM_LOCK;
 
-  traceEventHeapInfo(CAPSET_HEAP_DEFAULT,
-                     RtsFlags.GcFlags.generations,
-                     RtsFlags.GcFlags.maxHeapSize * BLOCK_SIZE,
-                     RtsFlags.GcFlags.minAllocAreaSize * BLOCK_SIZE,
-                     MBLOCK_SIZE,
-                     BLOCK_SIZE);
+  traceInitEvent(traceHeapInfo);
+
 }
 
 void storageAddCapabilities (uint32_t from, uint32_t to)
@@ -336,6 +359,43 @@ freeStorage (bool free_heap)
     freeGcThreads();
 }
 
+static void
+listGenBlocks (ListBlocksCb cb, void *user, generation* gen)
+{
+    cb(user, gen->blocks);
+    cb(user, gen->large_objects);
+    cb(user, gen->compact_objects);
+    cb(user, gen->compact_blocks_in_import);
+}
+
+// Traverse all the different places that the rts stores blocks
+// and call a callback on each of them.
+void listAllBlocks (ListBlocksCb cb, void *user)
+{
+  uint32_t g, i;
+  for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
+      for (i = 0; i < n_capabilities; i++) {
+          cb(user, capabilities[i]->mut_lists[g]);
+          cb(user, gc_threads[i]->gens[g].part_list);
+          cb(user, gc_threads[i]->gens[g].scavd_list);
+          cb(user, gc_threads[i]->gens[g].todo_bd);
+      }
+      listGenBlocks(cb, user, &generations[g]);
+  }
+
+  for (i = 0; i < n_nurseries; i++) {
+      cb(user, nurseries[i].blocks);
+  }
+  for (i = 0; i < n_capabilities; i++) {
+      if (capabilities[i]->pinned_object_block != NULL) {
+          cb(user, capabilities[i]->pinned_object_block);
+      }
+      cb(user, capabilities[i]->pinned_object_blocks);
+      cb(user, capabilities[i]->pinned_object_empty);
+  }
+}
+
+
 /* -----------------------------------------------------------------------------
    Note [CAF management]
    ~~~~~~~~~~~~~~~~~~~~~
@@ -422,7 +482,7 @@ freeStorage (bool free_heap)
    but can assume only two bits are available in STATIC_LINK (due to 32-bit
    systems).
 
-   To accomodate this we move handling of static objects entirely to the
+   To accommodate this we move handling of static objects entirely to the
    oldest generation when the nonmoving collector is in use. To do this safely
    and efficiently we allocate the blackhole created by lockCAF() directly in
    the non-moving heap. This means that the moving collector can completely
@@ -493,7 +553,7 @@ lockCAF (StgRegTable *reg, StgIndStatic *caf)
     // Secondly I think static thunks can't have payload: anything that they
     // reference should be in SRTs
     ASSERT(orig_info_tbl->layout.payload.ptrs == 0);
-    // Becuase the payload is empty we just push the SRT
+    // Because the payload is empty we just push the SRT
     IF_NONMOVING_WRITE_BARRIER_ENABLED {
         StgThunkInfoTable *thunk_info = itbl_to_thunk_itbl(orig_info_tbl);
         if (thunk_info->i.srt) {
@@ -534,7 +594,7 @@ newCAF(StgRegTable *reg, StgIndStatic *caf)
     bh = lockCAF(reg, caf);
     if (!bh) return NULL;
 
-    if(keepCAFs)
+    if(keepCAFs && !(highMemDynamic && (void*) caf > (void*) 0x80000000))
     {
         // Note [dyn_caf_list]
         // If we are in GHCi _and_ we are using dynamic libraries,
@@ -588,6 +648,12 @@ setKeepCAFs (void)
     keepCAFs = 1;
 }
 
+void
+setHighMemDynamic (void)
+{
+    highMemDynamic = 1;
+}
+
 // An alternate version of newCAF which is used for dynamically loaded
 // object code in GHCi.  In this case we want to retain *all* CAFs in
 // the object code, because they might be demanded at any time from an
@@ -922,6 +988,64 @@ accountAllocation(Capability *cap, W_ n)
 
 }
 
+/* Note [slop on the heap]
+ * ~~~~~~~~~~~~~~~~~~~~~~~
+ *
+ * We use the term "slop" to refer to allocated memory on the heap which isn't
+ * occupied by any closure. Usually closures are packet tightly into the heap
+ * blocks, storage for one immediately following another. However there are
+ * situations where slop is left behind:
+ *
+ * - Allocating large objects (BF_LARGE)
+ *
+ *   These are given an entire block, but if they don't fill the entire block
+ *   the rest is slop. See allocateMightFail in Storage.c.
+ *
+ * - Allocating pinned objects with alignment (BF_PINNED)
+ *
+ *   These are packet into blocks like normal closures, however they
+ *   can have alignment constraints and any memory that needed to be skipped for
+ *   alignment becomes slop. See allocatePinned in Storage.c.
+ *
+ * - Shrinking (Small)Mutable(Byte)Array#
+ *
+ *    The size of these closures can be decreased after allocation, leaving any,
+ *    now unused memory, behind as slop. See stg_resizzeMutableByteArrayzh,
+ *    stg_shrinkSmallMutableArrayzh, and stg_shrinkMutableByteArrayzh in
+ *    PrimOps.cmm.
+ *
+ *    This type of slop is extra tricky because it can also be pinned and
+ *    large.
+ *
+ * - Overwriting closures
+ *
+ *   During GC the RTS overwrites closures with forwarding pointers, this can
+ *   leave slop behind depending on the size of the closure being
+ *   overwritten. See Note [zeroing slop when overwriting closures].
+ *
+ * Under various ways we actually zero slop so we can linearly scan over blocks
+ * of closures. This trick is used by the sanity checking code and the heap
+ * profiler, see Note [skipping slop in the heap profiler].
+ *
+ * In general we zero:
+ *
+ *  - Pinned object alignment slop, see MEMSET_SLOP_W in allocatePinned.
+ *  - Large object alignment slop, see MEMSET_SLOP_W in allocatePinned.
+ *  - Shrunk array slop, see OVERWRITING_CLOSURE_MUTABLE.
+ *
+ * Note that this is necessary even in the vanilla (e.g. non-profiling) RTS
+ * since the user may trigger a heap census via +RTS -hT, which can be used
+ * even when not linking against the profiled RTS. Failing to zero slop
+ * due to array shrinking has resulted in a few nasty bugs (#17572, #9666).
+ * However, since array shrink may result in large amounts of slop (unlike
+ * alignment), we take care to only zero such slop when heap profiling or DEBUG
+ * are enabled.
+ *
+ * When performing LDV profiling or using a (single threaded) debug RTS we zero
+ * slop even when overwriting immutable closures, see Note [zeroing slop when
+ * overwriting closures].
+ */
+
 /* -----------------------------------------------------------------------------
    StgPtr allocate (Capability *cap, W_ n)
 
@@ -1074,6 +1198,21 @@ allocateMightFail (Capability *cap, W_ n)
     return p;
 }
 
+/**
+ * Calculate the number of words we need to add to 'p' so it satisfies the
+ * alignment constraint '(p + off) & (align-1) == 0'.
+ */
+#define ALIGN_WITH_OFF_W(p, align, off) \
+    (((-((uintptr_t)p) - off) & (align-1)) / sizeof(W_))
+
+/**
+ * When profiling we zero the space used for alignment. This allows us to
+ * traverse pinned blocks in the heap profiler.
+ *
+ * See Note [skipping slop in the heap profiler]
+ */
+#define MEMSET_SLOP_W(p, val, len_w) memset(p, val, (len_w) * sizeof(W_))
+
 /* ---------------------------------------------------------------------------
    Allocate a fixed/pinned object.
 
@@ -1099,29 +1238,49 @@ allocateMightFail (Capability *cap, W_ n)
    ------------------------------------------------------------------------- */
 
 StgPtr
-allocatePinned (Capability *cap, W_ n)
+allocatePinned (Capability *cap, W_ n /*words*/, W_ alignment /*bytes*/, W_ align_off /*bytes*/)
 {
     StgPtr p;
     bdescr *bd;
 
+    // Alignment and offset have to be a power of two
+    ASSERT(alignment && !(alignment & (alignment - 1)));
+    ASSERT(alignment >= sizeof(W_));
+
+    ASSERT(!(align_off & (align_off - 1)));
+
+    const StgWord alignment_w = alignment / sizeof(W_);
+
     // If the request is for a large object, then allocate()
     // will give us a pinned object anyway.
     if (n >= LARGE_OBJECT_THRESHOLD/sizeof(W_)) {
-        p = allocateMightFail(cap, n);
+        // For large objects we don't bother optimizing the number of words
+        // allocated for alignment reasons. Here we just allocate the maximum
+        // number of extra words we could possibly need to satisfy the alignment
+        // constraint.
+        p = allocateMightFail(cap, n + alignment_w - 1);
         if (p == NULL) {
             return NULL;
         } else {
             Bdescr(p)->flags |= BF_PINNED;
+            W_ off_w = ALIGN_WITH_OFF_W(p, alignment, align_off);
+            MEMSET_SLOP_W(p, 0, off_w);
+            p += off_w;
+            MEMSET_SLOP_W(p + n, 0, alignment_w - off_w - 1);
             return p;
         }
     }
 
-    accountAllocation(cap, n);
     bd = cap->pinned_object_block;
 
+    W_ off_w = 0;
+
+    if(bd)
+        off_w = ALIGN_WITH_OFF_W(bd->free, alignment, align_off);
+
     // If we don't have a block of pinned objects yet, or the current
     // one isn't large enough to hold the new object, get a new one.
-    if (bd == NULL || (bd->free + n) > (bd->start + BLOCK_SIZE_W)) {
+    if (bd == NULL || (bd->free + off_w + n) > (bd->start + BLOCK_SIZE_W)) {
 
         // stash the old block on cap->pinned_object_blocks.  On the
         // next GC cycle these objects will be moved to
@@ -1137,26 +1296,48 @@ allocatePinned (Capability *cap, W_ n)
         // avoid that (benchmarks that allocate a lot of pinned
         // objects scale really badly if we do this).
         //
-        // So first, we try taking the next block from the nursery, in
-        // the same way as allocate().
-        bd = cap->r.rCurrentNursery->link;
+        // See Note [Sources of Block Level Fragmentation]
+        // for a more complete history of this section.
+        bd = cap->pinned_object_empty;
         if (bd == NULL) {
-            // The nursery is empty: allocate a fresh block (we can't fail
+            // The pinned block list is empty: allocate a fresh block (we can't fail
             // here).
             ACQUIRE_SM_LOCK;
-            bd = allocBlockOnNode(cap->node);
+            bd = allocNursery(cap->node, NULL, PINNED_EMPTY_SIZE);
             RELEASE_SM_LOCK;
-            initBdescr(bd, g0, g0);
-        } else {
-            newNurseryBlock(bd);
-            // we have a block in the nursery: steal it
-            cap->r.rCurrentNursery->link = bd->link;
-            if (bd->link != NULL) {
-                bd->link->u.back = cap->r.rCurrentNursery;
-            }
-            cap->r.rNursery->n_blocks -= bd->blocks;
         }
 
+        // Bump up the nursery pointer to avoid the pathological situation
+        // where a program is *only* allocating pinned objects.
+        // T4018 fails without this safety.
+        // This has the effect of counting a full pinned block in the same way
+        // as a full nursery block, so GCs will be triggered at the same interval
+        // if you are only allocating pinned data compared to normal allocations
+        // via allocate().
+        bdescr * nbd;
+        nbd = cap->r.rCurrentNursery->link;
+        if (nbd != NULL){
+          newNurseryBlock(nbd);
+          cap->r.rCurrentNursery->link = nbd->link;
+          if (nbd->link != NULL) {
+              nbd->link->u.back = cap->r.rCurrentNursery;
+            }
+          dbl_link_onto(nbd, &cap->r.rNursery->blocks);
+          // Important for accounting purposes
+          if (cap->r.rCurrentAlloc){
+            finishedNurseryBlock(cap, cap->r.rCurrentAlloc);
+          }
+          cap->r.rCurrentAlloc = nbd;
+        }
+
+
+        cap->pinned_object_empty = bd->link;
+        newNurseryBlock(bd);
+        if (bd->link != NULL) {
+          bd->link->u.back = cap->pinned_object_empty;
+        }
+        initBdescr(bd, g0, g0);
+
         cap->pinned_object_block = bd;
         bd->flags  = BF_PINNED | BF_LARGE | BF_EVACUATED;
 
@@ -1173,10 +1354,20 @@ allocatePinned (Capability *cap, W_ n)
         // the next GC the BF_EVACUATED flag will be cleared, and the
         // block will be promoted as usual (if anything in it is
         // live).
+
+        off_w = ALIGN_WITH_OFF_W(bd->free, alignment, align_off);
     }
 
     p = bd->free;
+
+    MEMSET_SLOP_W(p, 0, off_w);
+
+    n += off_w;
+    p += off_w;
     bd->free += n;
+
+    accountAllocation(cap, n);
+
     return p;
 }
 
@@ -1372,10 +1563,10 @@ dirty_MVAR(StgRegTable *reg, StgClosure *p, StgClosure *old_val)
 // program.  Also emits events reporting the per-cap allocation
 // totals.
 //
-StgWord
+uint64_t
 calcTotalAllocated (void)
 {
-    W_ tot_alloc = 0;
+    uint64_t tot_alloc = 0;
     W_ n;
 
     for (n = 0; n < n_capabilities; n++) {
@@ -1616,178 +1807,6 @@ void flushExec (W_ len, AdjustorExecutable exec_addr)
 #endif
 }
 
-#if defined(linux_HOST_OS)
-
-// On Linux we need to use libffi for allocating executable memory,
-// because it knows how to work around the restrictions put in place
-// by SELinux.
-
-AdjustorWritable allocateExec (W_ bytes, AdjustorExecutable *exec_ret)
-{
-    void **ret, **exec;
-    ACQUIRE_SM_LOCK;
-    ret = ffi_closure_alloc (sizeof(void *) + (size_t)bytes, (void**)&exec);
-    RELEASE_SM_LOCK;
-    if (ret == NULL) return ret;
-    *ret = ret; // save the address of the writable mapping, for freeExec().
-    *exec_ret = exec + 1;
-    return (ret + 1);
-}
-
-// freeExec gets passed the executable address, not the writable address.
-void freeExec (AdjustorExecutable addr)
-{
-    AdjustorWritable writable;
-    writable = *((void**)addr - 1);
-    ACQUIRE_SM_LOCK;
-    ffi_closure_free (writable);
-    RELEASE_SM_LOCK
-}
-
-#elif defined(USE_LIBFFI_FOR_ADJUSTORS) && defined(darwin_HOST_OS)
-
-static HashTable* allocatedExecs;
-
-AdjustorWritable allocateExec(W_ bytes, AdjustorExecutable *exec_ret)
-{
-    AdjustorWritable writ;
-    ffi_closure* cl;
-    // This check is necessary as we can't use allocateExec for anything *but*
-    // ffi_closures on ios/darwin on arm.  libffi does some heavy lifting to
-    // get around the X^W restrictions, and we can't just use this codepath
-    // to allocate generic executable space. For those cases we have to refer
-    // back to allocateWrite/markExec/freeWrite (see above.)
-    if (bytes != sizeof(ffi_closure)) {
-        barf("allocateExec: for ffi_closure only");
-    }
-    ACQUIRE_SM_LOCK;
-    cl = writ = ffi_closure_alloc((size_t)bytes, exec_ret);
-    if (cl != NULL) {
-        if (allocatedExecs == NULL) {
-            allocatedExecs = allocHashTable();
-        }
-        insertHashTable(allocatedExecs, (StgWord)*exec_ret, writ);
-    }
-    RELEASE_SM_LOCK;
-    return writ;
-}
-
-AdjustorWritable execToWritable(AdjustorExecutable exec)
-{
-    AdjustorWritable writ;
-    ACQUIRE_SM_LOCK;
-    if (allocatedExecs == NULL ||
-       (writ = lookupHashTable(allocatedExecs, (StgWord)exec)) == NULL) {
-        RELEASE_SM_LOCK;
-        barf("execToWritable: not found");
-    }
-    RELEASE_SM_LOCK;
-    return writ;
-}
-
-void freeExec(AdjustorExecutable exec)
-{
-    AdjustorWritable writ;
-    ffi_closure* cl;
-    cl = writ = execToWritable(exec);
-    ACQUIRE_SM_LOCK;
-    removeHashTable(allocatedExecs, (StgWord)exec, writ);
-    ffi_closure_free(cl);
-    RELEASE_SM_LOCK
-}
-
-#else
-
-AdjustorWritable allocateExec (W_ bytes, AdjustorExecutable *exec_ret)
-{
-    void *ret;
-    W_ n;
-
-    ACQUIRE_SM_LOCK;
-
-    // round up to words.
-    n  = (bytes + sizeof(W_) + 1) / sizeof(W_);
-
-    if (n+1 > BLOCK_SIZE_W) {
-        barf("allocateExec: can't handle large objects");
-    }
-
-    if (exec_block == NULL ||
-        exec_block->free + n + 1 > exec_block->start + BLOCK_SIZE_W) {
-        bdescr *bd;
-        W_ pagesize = getPageSize();
-        bd = allocGroup(stg_max(1, pagesize / BLOCK_SIZE));
-        debugTrace(DEBUG_gc, "allocate exec block %p", bd->start);
-        bd->gen_no = 0;
-        bd->flags = BF_EXEC;
-        bd->link = exec_block;
-        if (exec_block != NULL) {
-            exec_block->u.back = bd;
-        }
-        bd->u.back = NULL;
-        setExecutable(bd->start, bd->blocks * BLOCK_SIZE, true);
-        exec_block = bd;
-    }
-    *(exec_block->free) = n;  // store the size of this chunk
-    exec_block->gen_no += n;  // gen_no stores the number of words allocated
-    ret = exec_block->free + 1;
-    exec_block->free += n + 1;
-
-    RELEASE_SM_LOCK
-    *exec_ret = ret;
-    return ret;
-}
-
-void freeExec (void *addr)
-{
-    StgPtr p = (StgPtr)addr - 1;
-    bdescr *bd = Bdescr((StgPtr)p);
-
-    if ((bd->flags & BF_EXEC) == 0) {
-        barf("freeExec: not executable");
-    }
-
-    if (*(StgPtr)p == 0) {
-        barf("freeExec: already free?");
-    }
-
-    ACQUIRE_SM_LOCK;
-
-    bd->gen_no -= *(StgPtr)p;
-    *(StgPtr)p = 0;
-
-    if (bd->gen_no == 0) {
-        // Free the block if it is empty, but not if it is the block at
-        // the head of the queue.
-        if (bd != exec_block) {
-            debugTrace(DEBUG_gc, "free exec block %p", bd->start);
-            dbl_link_remove(bd, &exec_block);
-            setExecutable(bd->start, bd->blocks * BLOCK_SIZE, false);
-            freeGroup(bd);
-        } else {
-            bd->free = bd->start;
-        }
-    }
-
-    RELEASE_SM_LOCK
-}
-
-#endif /* switch(HOST_OS) */
-
-#if RTS_LINKER_USE_MMAP
-AdjustorWritable allocateWrite(W_ bytes) {
-    return mmapForLinker(bytes, PROT_READ | PROT_WRITE, MAP_ANONYMOUS, -1, 0);
-}
-
-void markExec(W_ bytes, AdjustorWritable writ) {
-    mmapForLinkerMarkExecutable(writ, bytes);
-}
-
-void freeWrite(W_ bytes, AdjustorWritable writ) {
-    munmap(writ, bytes);
-}
-#endif
-
 #if defined(DEBUG)
 
 // handy function for use in gdb, because Bdescr() is inlined.
@@ -1800,3 +1819,106 @@ _bdescr (StgPtr p)
 }
 
 #endif
+
+/*
+Note [Sources of Block Level Fragmentation]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+
+Block level fragmentation is when there is unused space in megablocks.
+The amount of fragmentation can be calculated as the difference between the
+total size of allocated blocks and the total size of allocated megablocks.
+
+The act of the copying collection naturally reduces fragmentation by moving
+data between megablocks. Over time, the effect is that most megablocks end up quite full because
+data will be copied out of fragmented megablocks. The new block is chosen from
+the free list where the aim is to choose a gap of approximately the right size for
+the copied block so the data will end up in a probably less fragmented block.
+There are two situations where we end up with block fragmentation.
+
+1. Fragmentation from pinned data
+2. Fragmentation from nursery allocated blocks
+
+# Pinned Data Fragmentation
+
+There are two sources of
+pinned data, large objects and pinned bytearrays. After one of these object
+types is allocated, it is never moved by the collector
+and therefore if all the other blocks are collected around it then you can end
+up with a megablock with one pinned block and no other blocks. No special
+effort is taken in the compiler
+to ensure that this kind of fragmentation doesn't happen in the first place and
+once the heap is fragmented in this way, there's nothing you can do about it
+beyond hoping that the pinned data is eventually freed.
+
+# Nursery Fragmentation
+
+The other reason that a block may not ever be moved or emptied is if it forms
+part of the nursery.  When the nursery is first allocated then it is made up of
+megablock sized chunks, so if the nursery is 4 megabytes then it will consist of
+blocks from about 4 megablocks.
+
+Over time, the nursery is resized (by resizeNurseries) under various conditions.
+It gets bigger when
+we are allocating more and then smaller when we are allocating less.
+When the nursery is resized
+blocks are added or removed to it at potentially smaller sizes than a complete
+megablock. For example, if the nursery size needs to increase by 1, then
+the free list is consulted for a block of size 1 (from a random block)
+and that's added to the nursery.
+
+Over time the make-up of the nursery changes from 4
+contiguous megablocks to a hodge-podge of blocks from different megablocks. In
+some programs (see #19481), the fragmentation is so bad that a program with
+only 4 MB of live data can retain over 500 megablocks because each of these
+megablocks contributed a small number of blocks to the nursery.
+
+In particular, and confusingly, this second form of fragmentation was caused
+by the act of allocating pinned objects. `allocPinned` was the primary
+reason that the nursery size decreases by small amounts.  When `allocPinned`
+needed a block then it took a block permanently out of
+the nursery which shrunk the size of the nursery by 1 block. Then next time the size
+of the nursery was checked, the `alloc_nurseries` found that the existing
+nursery was smaller than the desired size and a new blocked needed
+to be added. This allocation was serviced from an arbitrary megablock
+which had some free space. The effect over time as more allocation happened
+was the nursery became made up of blocks from many different megablocks.
+
+Instead now we maintain a separate small list of blocks in `pinned_object_empty`
+which fresh blocks are taken from when we need a new one for a pinned block rather
+than threatening the continuity of the nursery. The size of this list is controlled
+by the PINNED_EMPTY_SIZE macro.
+
+In theory, this kind of fragmentation due to the nursery could still happen
+but in practice removing the primary cause (allocatePinned) was sufficient to
+greatly improve the situation. Another way to "fix" fragmentation of the nursery
+would be to periodically reallocate it when it was fragmented across many megablocks.
+
+Ticket: #19481
+
+# When can fragmentation be observed?
+
+Fragmentation is observed when the live data in a program is low compared to
+the overall resident size of the heap. The block allocator can reuse unused
+space within a megablock and therefore as residency
+increases again, the fragmented blocks will get filled up. Having a block-level
+fragmented heap means your program will never go below a certain memory
+threshold but it doesn't "use" more memory during periods of high residency.
+To clarify, say you observe 100 MB of fragmentation when your live data is
+4 MB, if your live data rise to 200MB then you probably will not still observe 100 MB
+of fragmentation as the block allocate will use the space in fragmented megablocks.
+
+# How to observe fragmentation
+
+Your heap is probably fragmented when
+
+* Live bytes is low
+* Memory in use (number of megablocks) is comparatively high
+* The size of the free list dominates residency (this can be observed using the
+  debug RTS and the memory inventory produced by -Dg).
+
+# Compacting Collector
+
+The compacting collector does nothing to improve megablock
+level fragmentation. The role of the compacting GC is to remove object level
+fragmentation and to use less memory when collecting. - see #19248
+*/
