@@ -9,7 +9,8 @@
  * ---------------------------------------------------------------------------*/
 
 #include "Rts.h"
-#include "IOManager.h"
+#include "ThrIOManager.h"
+#include "MIOManager.h"
 #include "rts\OSThreads.h"
 #include "Prelude.h"
 #include <windows.h>
