@@ -35,6 +35,7 @@ TARGETS=(
     #########################
 
     # Linux ARM
+    "arm-unknown-linux-gnueabi"
     "arm-unknown-linux-gnueabihf"
     "arm-unknown-linux-musleabihf"
     "armv6-unknown-linux-gnueabihf"
@@ -58,6 +59,9 @@ TARGETS=(
     "i386-unknown-linux-gnu"
     "i386-unknown-linux-musl"
     "i386-unknown-linux"
+    "i686-unknown-linux-gnu"
+    "i686-unknown-linux-musl"
+    "i686-unknown-linux"
     "x86_64-unknown-linux-gnu"
     "x86_64-unknown-linux-musl"
     "x86_64-unknown-linux"
@@ -72,6 +76,9 @@ TARGETS=(
     "powerpc64le-unknown-linux"
     # Linux s390x
     "s390x-ibm-linux"
+    # Linux riscv64
+    "riscv64-unknown-linux-gnu"
+    "riscv64-unknown-linux"
 
     #########################
     # Darwin
@@ -81,7 +88,6 @@ TARGETS=(
     "i386-apple-darwin"
     "x86_64-apple-darwin"
     "arm64-apple-darwin"
-    "aarch64-apple-darwin"
     # iOS
     "armv7-apple-ios"
     "arm64-apple-ios"
